/*----------------------------------------------------------------------** 
**									**
** An example program using VPOS, written by student Plamen Kiradjiew.	**
** It is a database system for storing, retrieving and			**
** maintenance of equivalent words in several languages.		**
**									**
**----------------------------------------------------------------------*/

#include <stdio.h>
#include <ctype.h>
#include <dynmem.h>
#include <vpos.h>

#define UPPER  2   /* flag constant for first upper case letter */

    static int LMAX = {2};  /* range of language array L */
    static char *L[]={
                          "ENGLISH",
                          "GERMAN"
    };          /* array of used languages */




main(argc, argv)
    int argc;
    char *argv[];

  {
    char c;
    
    DmInit( 100000L, NULL );

/* function choice */
    do {
      printf("\n\n\n\n");
      printf("1. insert\n");
      printf("2. delete\n");
      printf("3. synonyms\n");
      printf("4. unlink words\n");
      printf("5. browsing\n");
      printf("6. exit\n:");
      c=getchar();
      if (c=='\n') c=getchar();
      if (c=='6') break;
      if (c>='1' && c<'6') {
        switch(c) {
          case '1': Insert(); break;
          case '2': Delete();
                    break;
          case '4': Unlink();
                    break;
          case '3': Show(); break;
          case '5': Browse(); break;
        }    
      } 
    }
    while (1);

/* data base display */     
    open_dictionary();
    DispWholeVPOS( stdout );
    close_dictionary();
    DmClose();
    exit(0);

  }



open_dictionary()

  { 

    PsOpen( "DictionaryDB" );

  }
    


close_dictionary()
  {

    PsClose();

  }




long insert_word(word,language)
/* insert a word from certain language
   - a word exists in the dictionary by a chain of letter objects
     with a RING named "language" at the end ;
   return the point of the inserted or founded word */

    char *word,*language;

  {
    long next,p;
    char s[2],*w=word;
    int upper_case ={0},flag={0},exist={0};

    p=AtmNext(0L);
    if (p==NULLPNTR) 
      p=AtmCreateAndInsert(SPIDER, PsNameCode("DICTIONARY"), 
                      NULLPNTR, 0L);

/* check for upper case */
    if (isupper(*word)) 
      upper_case=UPPER;
     str_to_lower(word);
/* compressing the spaces in the word */
     space_compress(word);

/* determining of the point of word */
    while (*word!='\0') {
      s[0]=*word;
      s[1]='\0';
      next=AtmFind(p,PsNameCode(s),SPIDER);
      if (next==NULLPNTR)
        next=AtmCreateAndInsert(SPIDER, PsNameCode(s), p, NULLPNTR);
      p=next;
      word++;
    }

/* check for existance of such word */
    next=AtmFind(p,PsNameCode(language),RING);
    if (next!=NULLPNTR) 
      do 
        if (exist=(flag=AtmFlag(next,UPPER)) && upper_case || 
                                               !flag && !upper_case)
          break;
      while ((next=AtmFindNext(next,PsNameCode(language),RING))!=NULLPNTR);

/* if word doesn't exist then insert it */
    if(!exist) {
     next=AtmCreateAndInsert(RING, PsNameCode(language), p, NULLPNTR);
/* set on flag UPPER if word begins with an upper case letter */
     if (upper_case>0) AtmFlagOn(next,upper_case);
    }
    if(upper_case==UPPER) *w=toupper(*w);
    return(next);

  }



delete_word(word,language)
/* delete a word from certain language - with the deletion disappear
   the connections to other words;
   return - 0, if word successful deleted or
          - (-1), if such word doesn't exist in dictionary */

    char *word,*language;

  { 
    long p,o;
    
/* does the word exist ? */
    if ((p=find_word(word,language))==NULLPNTR)
      return(-1);
    o=AtmOwner(p);
    AtmDelete(p);
/* deleting letter objects to the top, if there are no successors */
    while (AtmSubord(p=o)==NULLPNTR) {
      o=AtmOwner(p);
      AtmDelete(p);
    }
    return(0);
  }



long find_word(word,language)
/* search for a letter path corresponding to a word and find a RING
   with name "language" ;
   return - point of founded word or
          - NULLPNTR otherwise  */

    char *word,*language;

  {
    long p,next,l;
    char s[2];
    int upper_case={0},flag;

    p=AtmNext(0L);
    if (p==NULLPNTR) return (NULLPNTR);
/* check for upper case */
    if(isupper(*word)) upper_case=UPPER;
    str_to_lower(word);
/* compressing the spaces in the word */
     space_compress(word);

/* passing the letter path */
    while (*word!='\0') {
      s[0]=*word;
      s[1]='\0';
      next=AtmFind(p,PsNameCode(s),SPIDER);
      if (next==NULLPNTR) return (NULLPNTR);
      p=next;
      word++;
    }

/* looking for a RING with a flag in dependence on the value of upper_case */
    if((next=AtmFind(p,PsNameCode(language),RING))!=NULLPNTR) {
      do {
        flag=AtmFlag(next,UPPER);
        if(upper_case && flag || !upper_case && !flag) 
          return(next);
      } while ((next=AtmFindNext(next,PsNameCode(language),RING))!=NULLPNTR);
    }
    return(NULLPNTR);
  }



long find_point(word)
/* search for existance of a letter path corresponding to word without
   looking for any RING ;
   return - point of the last letter of word or
          - NULLPNTR otherwise  */
    char *word;
  
  {
    long p,next,l;
    char s[2];

    p=AtmNext(0L);
    if (p==NULLPNTR) return (NULLPNTR);
    while (*word!='\0') {
      s[0]=*word;
      s[1]='\0';
      next=AtmFind(p,PsNameCode(s),SPIDER);
      if (next==NULLPNTR) return (NULLPNTR);
      p=next;
      word++;
    }
    return(p);
  }



get_word(point,word)
/* by a given point return in word the corresponding word */

    long point;
    char *word;

  {
    char name[12],*w=(word+29);
    int upper_case={0};

    *w='\0';
    if (point==NULLPNTR) {
      *word=*w; return ;
    }
/* check for upper case flag in the pointed RING */
    if(AtmFlag(point,UPPER)!=0) upper_case=UPPER;
    point=AtmOwner(point);
    PsCodeName(AtmName(point),name);

/* saving the letters in w in reverse order */
    while (strcmp(name,"DICTIONARY")!=0) {
      w--;
      *w=*name;
      point=AtmOwner(point);
      PsCodeName(AtmName(point),name);
    } 
/* change of the first letter if necessary (upper_case set to UPPER)  */
    if(upper_case==UPPER) *w=toupper(*w);
    do *word++=*w++;
      while (*w!='\0');
    *word='\0';
  }
      


set_link(point1,point2)
/* set connection between two words by adding a LOGICAL POINTER as
   successor of the RING of the first word; a BACKWARD POINTER is 
   added automatically as successor of the second word
   return - 0 if done or
          - (-1) when one of the points is NULLPNTR or the connection
             still exists */

    long point1,point2;   /* points of the words */

  {
    long  link,l1,l2;
    unsigned short int dmo;

    if (point1==NULLPNTR || point2==NULLPNTR) return(-1);

/* check if point2 is referenced from the RING of point1 */
    if ((l1=AtmFind(point1,AtmName(point2),VPOSLINK))!=NULLPNTR ) 
      do 
        if (AtmSubord(l1)==point2 ) return(-1);
       while ((l1=AtmFindNext(l1,AtmName(point2),VPOSLINK))!=NULLPNTR);

/* check if point1 is referenced from the RING of point2 */
    if ((l2=AtmFind(point2,AtmName(point1),VPOSLINK))!=NULLPNTR ) 
      do 
        if ( AtmSubord(l2)==point1) return(-1);
      while ((l2=AtmFind(l2,AtmName(point1),VPOSLINK))!=NULLPNTR );

/* creating of a LOGICAL POINTER in the RING of point1, pointing to
   pointing to point2  */
    dmo = AtmCreate(VPOSLINK,AtmName(point2));
    AtmDetPointer(dmo,point2);
    link=AtmInsLast(dmo,point1);
    return(0);
  }



drop_link(point1,point2)
/* unlink two by LOGICAL POINTERS connected points;
   return - 0 if unlink is done or
           (-1) when one of the points is NULLPNTR or the connection
           doesn't exist    */

    long point1,point2; /* points of the words */

  {
    long l1,l2;

    if (point1==NULLPNTR || point2==NULLPNTR) return(-1);

/* delete the existing LOGICAL POINTER in the RING of point1 */
    if ((l1=AtmFind(point1,AtmName(point2),VPOSLINK))!=NULLPNTR ) 
      do 
        if (AtmSubord(l1)==point2 ) { AtmDelete(l1); return(0); }
       while ((l1=AtmFindNext(l1,AtmName(point2),VPOSLINK))!=NULLPNTR);

/* delete the existing LOGICAL POINTER in the RING of point2 */
    if ((l2=AtmFind(point2,AtmName(point1),VPOSLINK))!=NULLPNTR ) 
      do 
        if ( AtmSubord(l2)==point1) { AtmDelete(l2); return(0); }
      while ((l2=AtmFind(l2,AtmName(point1),VPOSLINK))!=NULLPNTR );

    return(-1);
  }



int synonyms(sword,slang,twords,tlang)
    char *sword,*slang,twords[20][30],*tlang;
/* return a list twords of the words in language tlang referenced
   from sword in language slang;
   function value : the number of founded synonyms or (-1) if sword doesn't 
                    exist       */

  {
    int i;
    unsigned short int table;
    long p,link,backwards[20],*b=backwards;

/* initializing the table of points to synonyms of sword */
    for (i=0;i<20;i++) backwards[i]=0;
    i=0;
/* find sword */
    p=find_word(sword,slang);
    if (p==NULLPNTR) return(-1);

/* search for the words from tlang referenced from the RING of sword */
    if ((link=AtmFind(p,PsNameCode(tlang),VPOSLINK))!=NULLPNTR) {
      get_word(AtmSubord(link),twords[i++]);
      while ((link=AtmFindNext(link,PsNameCode(tlang),VPOSLINK))!=NULLPNTR) 
        get_word(AtmSubord(link),twords[i++]);
    }

/* search for the words pointing to sword (marked in BACHWARD atoms */
    table=AtmReferences(p);
    DmGet(table,backwards);
    while (*b!=0) {
    p=AtmOwner(*b);
    if (AtmName(p)==PsNameCode(tlang))
      get_word(p,twords[i++]);
    b++;
    }
    return(i);
  }



language_choice(from,to)
/* menu of the contained languages with choice of the source and target
   language, saved as strings in from and to    */

    char *from,*to;

  {
    int i;
    char n1,n2;

    do {
      printf("\n\n\n");
      for (i=0;i<LMAX;i++) printf("%d. %s\n",i+1,L[i]);
      printf("\nfrom:");
      n1=getchar();
      if (n1=='\n') n1=getchar();
      n1-='0';
      printf("to:");
      n2=getchar();
      if (n2=='\n') n2=getchar();
      n2-='0';
    } while (n1<1 || n1>LMAX || n2<1 || n2>LMAX);
    strcpy(from,L[n1-1]);
    strcpy(to,L[n2-1]);
    getchar();

  }



Insert()
/* the INSERT activity */
  
  {
    char word[30],synonym[20][30],from[20],to[20];
    long p1,p2;
    int i,ok,n;

/* input of the source and target language */    
    language_choice(from,to);
/* input of words from the source language */
    do {
      printf("\n\n\n\n");
      printf("%s WORD : ",from);
      fgets(word,30,stdin);
      if (word[0]=='\n') break;
      ok=1;
/* check for the correctness of the word */
      for (i=0;i<30 && word[i]!='\n';i++)
        if (!isalpha(word[i]) && word[i]!=' ') { ok=0;break; }
      if (ok) {
        word[i]='\0';
        open_dictionary();
/* insert the word from the source language */
        p1=insert_word(word,from);
/* finding the currently contained synonyms of the word */
        n=synonyms(word,from,synonym,to);
        close_dictionary();
/* printing the synonyms */
        for(i=0;i<n;i++) printf("\n%45s WORD : %s\n",to,synonym[i]);
        n=i;
/* input of additional synonyms */
        do {
          printf("\n%45s WORD : ",to);
          fgets(synonym[n],30,stdin);
          if (synonym[n][0]=='\n') break;
          ok=1;
/* check for the correctness of the word */
          for (i=0;i<30 && synonym[n][i]!='\n';i++)
            if (!isalpha(synonym[n][i]) && synonym[n][i]!=' ') { ok=0;break; }
          if (ok) {
            synonym[n][i]='\0';
            open_dictionary();
/* insertion of the new synonym */
            p2=insert_word(synonym[n],to);
/* connection between the source and synonym word */
            set_link(p1,p2);
            close_dictionary();
            n++;
          }
        } while (1);
      }
    } while (1);
  }



Show()
/* the activity for LISTING the synonyms of a word */

  {
    char word[30],twords[20][30],from[20],to[20];
    int n,i;

/* determining the source and target language */
    language_choice(from,to);
/* input of the word from the source language */
    do {
      printf("\n\n\n%s WORD : ",from);
      fgets(word,30,stdin);
      if (word[0]=='\n') break;
      for(i=1;word[i]!='\n';i++);
      word[i]='\0';
      open_dictionary();
/* finding the corresponding synonyms */
      n=synonyms(word,from,twords,to);
      close_dictionary();
/* output */
      if (n==-1) printf("---THIS WORD DOESN'T EXIST IN DICTIONARY !");
      else {
        printf("%s SYNONYMS : ",to);
        for (i=0;i<n;i++) printf("%s ",twords[i]);
      }
    } while (1);
  }



Delete()
/* the DELETE activity */  

  {
    int n1,i;
    char word[30],language[20];
    
/* input of the language,which is the word to be deleted from */
    do {
      printf("\n\n\n");
      for (i=0;i<LMAX;i++) printf("%d. %s\n",i+1,L[i]);
      printf("\nlanguage : ");
      n1=getchar();
      if (n1=='\n') n1=getchar();
      n1-='0';
    } while (n1<1 || n1>LMAX);
    getchar();
    strcpy(language,L[n1-1]);
/* input of the word to be deleted */
    printf("\n%s word : ",language);
    fgets(word,30,stdin);
    if (word[0]=='\n') return;
    for(i=1;word[i]!='\n';i++);
    word[i]='\0';
    open_dictionary();
/* deleting of the word */
    if(delete_word(word,language)<0)
      printf("---THIS WORD DOESN'T EXIST IN DICTIONARY !");
    close_dictionary();
  }



Unlink()
/* DISCONNECT two existing words without deleting them */

  {
    char from[20],to[20], w1[30],w2[30];
    int i;

/* input of the languages of the two words */
    language_choice(from,to);
/* input of the two words */
    printf("\n\n\n%s WORD : ",from);
    fgets(w1,30,stdin);
    if (w1[0]=='\n') return;
    for(i=1;w1[i]!='\n';i++);
    w1[i]='\0';
    printf("\n%s WORD : ",to);
    fgets(w2,30,stdin);
    if (w2[0]=='\n') return;
    for(i=1;w2[i]!='\n';i++);
    w2[i]='\0';
    open_dictionary();
/* disconnect */
    if(drop_link(find_word(w1,from),find_word(w2,to))<0)
       printf("---SOME WORD DOESN'T EXIST IN DICTIONARY !");
    close_dictionary();
  }



Browse()
/* the BROWSE activity : listing of all contained words from 
   a certain language between some strings */

  {
    char language[20],from[30],to[30];
    int i,n1;

/* input of the needed language */
    do {
      printf("\n\n\n");
      for (i=0;i<LMAX;i++) printf("%d. %s\n",i+1,L[i]);
      printf("\nlanguage : ");
      n1=getchar();
      if (n1=='\n') n1=getchar();
      n1-='0';
    } while (n1<1 || n1>LMAX);
    getchar();
    strcpy(language,L[n1-1]);
/* begin of the interval */
    printf("\nfrom : ");
    fgets(from,30,stdin);
    if (from[0]=='\n') return;
    for(i=1;from[i]!='\n';i++);
    from[i]='\0';
/* end of the interval */
    printf("\nto : ");
    fgets(to,30,stdin);
    if (to[0]=='\n') return;
    for(i=1;to[i]!='\n';i++);
    to[i]='\0';
    printf("\n\n\n");
    open_dictionary();
/* browsing the interval from the certain language */
    browse_dictionary(from,to,language);
    close_dictionary();
  }



browse_dictionary(from,to,lang)
/* browsing in a by strings given interval from a language lang */

    char *from,*to,*lang;

  {
    char c;
    int len;
    long p;

/* transforming the borders of the interval (from and to) to lower case */
    str_to_lower(from);
    space_compress(from);
    str_to_lower(to);
/* traversing the structure below the word from */
    if((p=find_point(from))!=NULLPNTR) traverse(p,from,to,lang);
    len=strlen(from);
/* traversing the structure below every letter of from moving to the top 
   of dictionary */
    while (len>0) {
     
      for (c=++from[--len];c<='z';c++)  {
       from[len]=c;
        p=find_point(from);
        if (p!=NULLPNTR) if (traverse(p,from,to,lang)<0) return;
      }
      from[len]='\0';
    }
  }



traverse(p,word,to,lang)
/* traversing the structure below a string word from language lang
   pointed by p until word > to (end of interval) ;
   return (-1) if to (the end of interval) is reached
            0 - otherwise */

    char *word,*lang,*to;
    long p;
  {
    char c,s[2],w[30];
    long q;

/* check if word exceeds to */
    if (str_compare(word,to)>0) return(-1); 
/* if word exists then print */
    if ((q=AtmFind(p,PsNameCode(lang),RING))!=NULLPNTR) {
      if(AtmFlag(q,UPPER)!=0) printf("%c%s\n",toupper(*word),(word+1));
      else printf("%s\n",word);
    }
    while ((q=AtmFindNext(q,PsNameCode(lang),RING))!=NULLPNTR) {
      if(AtmFlag(q,UPPER)!=0) printf("%c%s\n",toupper(*word),(word+1));
      else printf("%s\n",word);
    }
/* if no letter successors (of type SPIDER) then return 0 to the
   upper level  */
    if (AtmFind(p,0xFFFF,SPIDER)==NULLPNTR) return(0);
/* check if word contains two words separated by space and pass 
   to the next letter */
    if ((q=AtmFind(p,PsNameCode(" "),SPIDER))!=NULLPNTR) {
      strcpy(w,word);
      strcat(w," ");
      if (traverse(q,w,to,lang)<0) return(-1);
    }
/* passing all successing letters in lexicographical order */
    for (c='a';c<='z';c++) {
      s[0]=c;s[1]='\0';
      if ((q=AtmFind(p,PsNameCode(s),SPIDER))!=NULLPNTR) {
        strcpy(w,word);
        strcat(w,s);
        if (traverse(q,w,to,lang)<0) return(-1);
      }
    }
    return(0);
  }



str_to_lower(s)
/* convert string s to lower case */

    char *s;

  { 
    while (*s!='\0') {
      if (isupper(*s)) *s=tolower(*s);
      s++;
    }
  }



str_compare(s,t)
/* string comparison without considering the contained spaces  */

    char *s,*t;

  {
    do {
      if(*s=='\0' && *t=='\0') return(0);
      while (*s==' ') s++;
      while (*t==' ') t++;
    } while (*s++==*t++);
   return(*--s-*--t);
  }


space_compress(s)
/* compress the spaces in a string */

    char *s;

  {
    int one={0};
    char *w=s;
 
    while (*s==' ') s++;
    do {
      while (*s==' ') {s++;one=1;}
      if(one && *s!='\0') {*w++=' ';one=0;}
      *w++=*s;
    } while (*s++!='\0');
  }

