/*-------------------------------------------------------------------------+
| Virtul persistent object store, basic functions (SUN version)            |
|                                                                          |
| Date of last updating: 18 November 1991                                  |
|                                                                          |
| Copyright (C) 1991 by Kazimierz Subieta.                                 |
| This software is subject to a licence agreement.                         |
| No restrictions when used for didactic purposes.                         |
+_________________________________________________________________________*/

#include        <fcntl.h>
#include        <limits.h>
#include        <stdio.h>
#include        <string.h>

#include        <messages.h>
#include        <dynmem.h>
#include        <vpos.h>
#include        <vpos-mis.h>

#define b0  0x0001
#define b1  0x0002
#define b2  0x0004
#define b3  0x0008
#define b4  0x0010
#define b5  0x0020
#define b6  0x0040
#define b7  0x0080
#define b8  0x0100
#define b9  0x0200
#define b10 0x0400
#define b11 0x0800
#define b12 0x1000
#define b13 0x2000
#define b14 0x4000
#define b15 0x8000

#define APPARENT        0x10
#define CANCELLED       0x02
#define HEADONLY        0x04
#define IS_UPDATED      ->FLAGS & 0x01
#define MAXATOM         16777215L
#define MAXFILE         16777215L
#define MAXOPENFILES    10
#define NEWFILE         0x04
#define TONEXT          0xFD
#define TOOWNER         0x02

unsigned char RING_B            =   0;
unsigned char RING_E            =   5;  /* 0..5  - ring atoms */
unsigned char LOGICAL_B         =   6;
unsigned char LOGICAL_E         =   8;  /* 6-8  - logical pointer atoms */
unsigned char HOLDER            =   9;  /* 9     - place holder atom */
unsigned char TOEND             =  10;  /* 10    - reference to end of a ring */
unsigned char LONGINT_B         =  11;
unsigned char LONGINT_E         =  20;  /* 11..20 - long integer atoms */
unsigned char BYTES_B           =  21;
unsigned char BYTES_E           = 240;  /* 21..240 - variable size atoms */
unsigned char SPIDER_B          = 241;
unsigned char SPIDER_E          = 254;  /* 241..254 - spider atoms */
unsigned char BACKWARD          = 255;  /* 255   - backward pointers atom */

int  PS_ERROR   =       b0;     /* Physical structure errors:
			b0 - uninitialized physical structure,
			b1 - cannot open data file,
			b2 - file reading error,
			b3 - file writing error,
			b4 - unallowed operation during atom creation,
			b5 - too large atom;
			b6 - unsuccessful assignment to atom value
			b7 - table of names overflow
			b8 - created file already exists
			b9 - too many data files
			b10 - unallowed deletion of a file
			b12 - unsuccessful closing of the file
			b13 - too large file
			b14 - unallowed application of a function
			b15 - wrong AFTER parameter in AtxInsert, AtxMove
						*/
unsigned char   CHANGES_IN_VPOS = 0;    /* 1 if anything has been changed */
unsigned short int  PERMNAMES = 0;      /* DM object containing permanent names */
unsigned char     CURR_FILE     = 0;    /* Current file for storing new atoms */
char      WORK_EXT[]    = "$$%";
char      OLD_EXT[]     = "%";
unsigned short int  FILETABLE   = 0;    /* DM object containing the data file table */
int       ACTUAL_FILE   = -1;   /* Actual file corresponding to CURR_FPOINTER*/
FILE *    CURR_FPOINTER = NULL; /* File pointer to actually opened file */
unsigned short int  AINVMAIN    = 0;    /* DM object with the main atom inventory */
unsigned  AINVSIZE      = 1301; /* Size of atom inventory, in items */
atomdesc *BEGAMAIN;             /* Beginning of atom inventory */
unsigned short int  AINVOVER    = 0;    /* DM object with overflow items */
atomdesc *BEGAOVER;
int       OPENFILES     = 0;    /* The number of opened files */
static  int  LASTCLOSED = 0;    /* Last closed file */

/*...........................................................................*/

atomdesc *  AtxAllocate(/* int, long, unsigned FLAGS, POINT, DMOBJ */ );
long        AtxBack(    /* long POINT */ );
long        AtxChangeSize( /* long, long POINT, VALSIZE */ );
void        AtxDeAllocate( /*long POINT */ );
void        AtxDelBack( /* long, long BACK, POINT */ );
int         AtxDiscNext(/* long PREVIOUS */ );
void        AtxInclude( /* long, long, long POINT, OWNER, AFTER */ );
void        AtxInsBack( /* long, int, long POINT, TYPE, POINTER */ );
long        AtxLast(    /* long POINT */ );
void        AtxLeaveHolder( /* long, long POINT, NEWPOINT */ );
atomdesc *  AtxLocate(  /* long POINT */ );
void        AtxMove(    /* long, long, long POINT, OWNER, AFTER*/ );
int         AtxPntAssign(/* long, long, int POINT, POINTER, TYPE */ );
long        AtxPriNexOwn(/* long, long *, long * POINT, NEXT, OWNER */ );
int         AtxRdPhys(  /* long, unsigned, int * POINT, DMOBJ, MODE */ );
long        AtxSize(    /* long POINT */ );
long        AtxSubord(  /* long POINT */ );
long        AtxValAssign(/* long, int, char *, long */ );
long        AtxValSize( /* long POINT */ );
void        DelSubordinates( /* long POINT */ );
long        FindAreaUpdRef( /* long, long POINT, ATOMSIZE */ );
#define         HashPoint(P) ((P)%AINVSIZE)
long        InclToSpider( /* long, long, long SPOWNER, POINT, AFTER */ );
int         IsSpiderMember( /* long, long * POINT, SPOWNER */ );
int         PsCloFile(  /* int, int FILENO, FRQ */ );
void        PsError();
void        PsExtFile(  /* FILE *, long FPOINTER, SIZE */ );
int         PsFindAfter( /* long, long POINT, SIZE */ );
long        PsFindArea( /* long, long SIZE, SUGESTION */ );
FILE *      PsFileOpen( /* long POINT */ );
void        PsGarbage(  /* int, long, long FILENO, CPOINT, CSIZE */ );
void        PsMakeGarbage( /* long POINT */ );
void        RemDmAtom(  /* atomdesc *, int AT, FRQ */ );
void        ShowActGar( /* int */ );
void        SkipParam(  /* long *, long * OWNER, AFTER */ );
unsigned short int    UpdDmObject(/* long POINT */ );
void        WrtAtom(    /* filedesc *, int, atomdesc *, int
				 FDESC, FILENO, INV, FRQ */ );

/*---------------------------------------------------------------------------*/
/* Comment on organization of SPIDER atoms. They are organized as normal
variable-size atoms. Their value contains triples <type, name, point>, "type"
is a member type, "name" is a name of a member, and "point" is a point of the
member. The list of triples may be empty. Each member of a spider contains
directly a point to the owner. Backward atoms are treated in the same way, but
it is assumed that their names are 0xFFF8, what gives a possibility of fast
recognizing if the given member is associated with a backward atom. The triple
<type,name,point> is described as a structure spider_item, having members
(char)MEMBER_TYPE, (unsigned short int)MEMBER_NAME and (long)MEMBER.
If the MEMBER is NULLPNTR it means a cancelled item. */





/*------------------------------------------------------ AtxFile -----------*/
int AtxFile( P )
    long P;
  {
    P = *((unsigned char *)&P + 3);
 /*    P = *(unsigned char *)&P;    09.06.1996 AJ */
    return((int)P);
  }


/*------------------------------------------------------ AtxPoint ----------*/
long AtxPoint( P )
    long P;
  {
    P = P & 0x00FFFFFFL;
    return(P);
  }


/*-------------------------------------------------------- AtxType ---------*/
int AtxType( P )
    long P;
  {
    P = P == NULLPNTR ? -1 : (unsigned char)((atom_head *)
			     DmPointer( AtxRead(P,1)->DMOBJ))->TYPE;
    return((int)P);
  }


/*------------------------------------------------------ ChangeTypeInSpider */
/* If POINT is a member of a spider, set type NEWTYPE within spider's dir */

static void  ChangeTypeInSpider( POINT, NEWTYPE )
    long POINT;
    int NEWTYPE;
  {
    long OWNER;
    if( IsSpiderMember( POINT, &OWNER ) )
      {
	/* Change type in the spider */
	unsigned short int DM_SPIDER = AtxRead( OWNER, 0 )->DMOBJ;
	spider_item *SPI = (spider_item *)( DmFix( DM_SPIDER ) + ATOMHEAD );
	for( ; DmInside( (char *)SPI, DM_SPIDER ); SPI++ )
	    if( SPI->MEMBER == POINT )
	      {
		SPI->MEMBER_TYPE = NEWTYPE;
		break;
	      }
	DmUnfix( DM_SPIDER );
      }
  }


/*----- External ---------------------------------------- AtmMoveAllMembers -*/
/* Move all members from old owner to the new; then remove the old owner */

long AtmMoveAllMembers( OLD_OWNER, NEW_OWNER )
    long OLD_OWNER;
    long NEW_OWNER;
  {
    if( OLD_OWNER != NEW_OWNER )
      {
	long MEM, NEXT_MEM;
	for( MEM = AtmSubord( OLD_OWNER ); MEM != NULLPNTR; MEM = NEXT_MEM )
	  {
	    NEXT_MEM = AtmNext( MEM );
	    AtmMoveLast( MEM, NEW_OWNER );
	  }
	AtmDelete( OLD_OWNER );
	SkipHolders( NEW_OWNER );
      }
    return( NEW_OWNER );
  }


/*----- External ----------------------------------------- AtmAssign --------*/
/* Assign a value from another atom; return new atom point. Copy recursively
the structure from ANPOINT to POINT. Name of POINT remains unchanged, type is
inherited from ANPOINT. The procedure is recursive. Logical pointers, backward
pointers, and TOENDs are not copied. POINT may be an owner of a ring or
a spider; in this case all its subordinates are removed */

long AtmAssign( POINT, ANPOINT )
    long POINT, ANPOINT;
  {
    int TYPE, ANTYPE;
    long DUMMY = NULLPNTR;
    SkipHolders( POINT );
    CHANGES_IN_VPOS = 1;
    CURR_FILE = AtxFile( POINT );
    if( ANPOINT != NULLPNTR )
	SkipHolders( ANPOINT );
    else
      {
	PS_ERROR |= b6;
	return( POINT );
      }
    if( POINT == ANPOINT )
	return( POINT );
    TYPE = AtxType( POINT );
    ANTYPE = AtxType( ANPOINT );
    if( RING_B <= ANTYPE && ANTYPE <= RING_E ||
				SPIDER_B <= ANTYPE && ANTYPE <= SPIDER_E )
      {
	long A;
	/* Make a copy of ANPOINT and connect it under DUMMY */
	DUMMY = AtmCreateAndInsert( ANTYPE, 0xFFFF, 0L, NULLPNTR );
	/* Copy all subordinates of a ring/spider */
	for( A = AtmSubord( ANPOINT ); A != NULLPNTR; A = AtmNext( A ) )
	  {
	    unsigned short int NN;
	    int TT = AtmTypeAndName( A, &NN );
	    PsDelHolder( AtmAssign( AtmInsLast( AtmCreate( TT, NN), DUMMY),A));
	  }
      }
    DelSubordinates( POINT );
    /* Inherit type from ANPOINT */
    if( LOGICAL_B <= TYPE && TYPE <= LOGICAL_E ) /* Cancel backward pointer */
	AtxPntAssign( POINT, NULLPNTR, ANTYPE );
    else
	AtxValAssign( POINT, ANTYPE, NULL, 0L );
    if( DUMMY != NULLPNTR )
      {
	POINT = AtmMoveAllMembers( DUMMY, POINT );
	return( POINT );
      }
    if( TYPE != ANTYPE )
	ChangeTypeInSpider( POINT, ANTYPE );
    if( LONGINT_B <= ANTYPE && ANTYPE <= LONGINT_E )
					/* For all integer types */
      {
	/* Assign elementary value from another atom; return new atom point */
	unsigned short int ANDMOBJ = AtxRead( ANPOINT, 0 )->DMOBJ;
	long LONGVALUE = ((atom_head *)DmPointer( ANDMOBJ ))->VALUE;
	POINT = AtxValAssign( POINT, ANTYPE, (char *)&LONGVALUE, 4 );
      }
    else if( BYTES_B <= ANTYPE && ANTYPE <= BYTES_E )
					/* For other atomic types */
      {
	/* Assign elementary value from another atom; return new atom point */
	unsigned short int ANDMOBJ = AtxRead( ANPOINT, 0 )->DMOBJ;
	char *ANBG = DmFix( ANDMOBJ );
	long ANSIZE = AtxValSize( ANPOINT );
	POINT = AtxValAssign( POINT, ANTYPE, ANBG + ATOMHEAD, ANSIZE );
	DmUnfix( ANDMOBJ );
      }
    else if( LOGICAL_B <= ANTYPE && ANTYPE <= LOGICAL_E )
	AtxPntAssign( POINT, AtxSubord( ANPOINT ), ANTYPE );
    if( DUMMY != NULLPNTR )
      {
	AtxDelete( NULLPNTR, DUMMY );
	PS_ERROR |= b6;
      }
    SkipHolders( POINT );
    return( POINT );
  }


/*------------------------------------------------------- DelSubordinates ---*/
/* Delete all subordinates of the point */

static void  DelSubordinates( POINT )
    long POINT;
  {
    int TYPE = AtxType( POINT );
    if( RING_B <= TYPE && TYPE <= RING_E )
      {
	long FIRST = AtxSubord( POINT );
	if( FIRST != NULLPNTR )
	  {
	    long P, P1;
	    for( P = AtxNext( FIRST ); P != NULLPNTR; P = P1 )
	      {
		P1 = AtxNext( P );
		if( AtxType( P1 ) == BACKWARD )
		    P1 = AtxNext( P1 );
		AtxDelete( FIRST, P );
	      }
	    AtxDelete( NULLPNTR, FIRST );
	  }
      }
    else if( SPIDER_B <= TYPE && TYPE <= SPIDER_E )
      { /* Making copy of the spider's value causes inconsistency! */
/* New version */
	atomdesc *AD = AtxRead( POINT, 0 );
	spider_item *SPI = (spider_item *)DmFixTop(AD->DMOBJ) - 1;
	AD->FLAGS |= UPDATED;
	for( ; DmInside( (char *)SPI - ATOMHEAD, AD->DMOBJ ); SPI-- )
	    if( SPI->MEMBER != NULLPNTR )
		AtxDelete( NULLPNTR, SPI->MEMBER );
	DmUnfix( AD->DMOBJ );

/* Old version
	unsigned short int SP_ATOM = AtxGetValue( POINT );
	spider_item *SPI = (spider_item *)DmFixTop( SP_ATOM ) - 1;
	for( ; DmInside( (char *)SPI, SP_ATOM ); SPI-- )
	    if( SPI->MEMBER != NULLPNTR )
		AtxDelete( NULLPNTR, SPI->MEMBER );
	DmUnfixDelete( SP_ATOM );
*/
      }
  }


/*----- External ----------------------------------------- AtmDelete --------*/
/* Delete an atom; external */

void AtmDelete( POINT )
    long POINT;
  {
    if( DM_COMPLAIN )
	PsAtxDel( );
    if( POINT != NULLPNTR )
      {
	SkipHolders( POINT );
	AtxDelete( NULLPNTR, POINT );
      }
  }


/*-------------------------------------------------------- AtxDelete --------*/
/* Delete an atom. All subordinated atoms and references to deleted atoms are 
   also deleted. The procedure is recursive. */

void  AtxDelete( PREVIOUS, POINT )
    long PREVIOUS;
    long POINT;
  {
    int ATYPE = AtxType( POINT );
    if( ATYPE == BACKWARD )   /* If BACKWARD atom, delete references from it */
      {
	long *H;
	unsigned short int BA;
	BA = AtxGetValue( POINT );
	for( H = (long *)DmFix( BA ); DmInside( (char *)H, BA ); H++ )
	    if( *H != NULLPNTR )
		AtxDelete( NULLPNTR, *H );
	DmUnfixDelete( BA );
      }
    else
      {
	/* Delete all references to the atom */
	long SPOWNER;
	long NEXT = AtxNext( POINT );
	if( AtxType( NEXT ) == BACKWARD )
	    AtxDelete( POINT, NEXT );
	DelSubordinates( POINT ); /* Delete all subordinates of the POINT */
	/* Delete backward pointer */
	if( LOGICAL_B <= ATYPE && ATYPE <= HOLDER )
	    AtxDelBack( AtxBack( POINT ), POINT );
	/* Disconnect the atom and return its space to garbage */
	if( IsSpiderMember( POINT, &SPOWNER ) )
	    DiscFromSpider( SPOWNER, POINT );
	else if( PREVIOUS == NULLPNTR )
	    AtxDisconnect( POINT );
	else
	    AtxDiscNext( PREVIOUS );
	PsMakeGarbage( POINT );
      }
  }


/*-------------------------------------------------------- DiscFromSpider --*/
/* Disconnect the member from the spider, together with its backward atom */

void  DiscFromSpider( SPOWNER, MEMBER )
    long SPOWNER, MEMBER;
  {
    atomdesc *AD = AtxRead( SPOWNER, 0 );
    spider_item *SPI = (spider_item *)(DmFix(AD->DMOBJ) + ATOMHEAD);
    AD->FLAGS |= UPDATED;
    for( ; DmInside( (char *)SPI, AD->DMOBJ ); SPI++ )
	if( SPI->MEMBER == MEMBER )
	  {
	    SPI->MEMBER = NULLPNTR;
	    SPI++;
	    if( DmInside( (char *)SPI, AD->DMOBJ ) && SPI->MEMBER != NULLPNTR
			&& SPI->MEMBER_NAME == 0xFFF8 )
		SPI->MEMBER = NULLPNTR;
	    break;
	  }
    DmUnfix( AD->DMOBJ );
  }


/*-------------------------------------------------------- IsSpiderMember --*/
/* Test if the POINT is a member of a spider; if so, return spider's owner,
   otherwise return NULLPNTR */

static int  IsSpiderMember( POINT, SPOWNER )
    long POINT;
    long *SPOWNER;
  {
    atomdesc *AD = AtxRead( POINT, 1 );
    atom_head *AH = (atom_head *)DmFix( AD->DMOBJ );
    if( AH->STATUS & TOOWNER ) /* If the pointer leads to owner */
      {
	int TYPE = AtxType( AH->POINT );
	if( SPIDER_B <= TYPE && TYPE <= SPIDER_E )
	  {
	    *SPOWNER = AH->POINT;
	    DmUnfix( AD->DMOBJ );
	    return( 1 );
	  }
      }
    *SPOWNER = NULLPNTR;
    DmUnfix( AD->DMOBJ );
    return( 0 );
  }


/*-------------------------------------------------------- AtxInsBack -------*/
/* Insert after the POINTER a BACKWARD atom if it does not exist; then
  augment the value of this atom by POINT */

static void  AtxInsBack( POINT, TYPE, POINTER )
    long POINT;         /* Point of reference */
    int TYPE;           /* Type of reference */
    long POINTER;       /* Referenced atom */
  {
    if( LOGICAL_B <= TYPE && TYPE <= HOLDER && POINTER != NULLPNTR )
      {
	long BACK = AtxNext( POINTER );
	if( AtxType( BACK ) == BACKWARD )
	  {
	    unsigned short int BA;
	    long *H;
	    atomdesc *AD = AtxRead( BACK, 0 );
	    AD->FLAGS |= UPDATED;
	    BA = AD->DMOBJ;
	    for( H = (long *)( DmFix( BA ) + ATOMHEAD );
					DmInside( (char *)H, BA ); H++ )
		if( *H == NULLPNTR )
		  {
		    *H = POINT;
		    DmUnfix( BA );
		    return;
		  }
	    DmUnfix( BA );
	    BA = AtxGetValue( BACK );
	    DmPush( BA, (char *)&POINT, 4 );
	    POINT = NULLPNTR;   /* Push NULLPNTR as a reserve */
	    DmPush( BA, (char *)&POINT, 4 );
	    AtxPutValue( BACK, BA );
	  }
	else
	  {
	    /* Create a new backward atom */
	    unsigned short int BACKAT = AtmCreate( BACKWARD, 0xFFF8 );
	    AtmDetValue( BACKAT, (char *)&POINT, 4L );
	    CURR_FILE = AtxFile( POINTER );
	    AtxInsert( BACKAT, NULLPNTR, POINTER );
	  }
      }
  }


/*-------------------------------------------------------- AtxDelBack -------*/
/* Delete a BACKWARD pointer POINT from BACK atom */

static void  AtxDelBack( BACK, POINT )
    long BACK, POINT;
  {
    if( BACK != NULLPNTR )
      { 
	int K = 1;
	long *H;
	atomdesc *AD = AtxRead( BACK, 0 );
	AD->FLAGS |= UPDATED;
	/* Set POINT = NULLPNTR and check if AD->DMOBJ contains other point */
	for( H = (long *)( DmFix( AD->DMOBJ ) + ATOMHEAD );
			DmInside( (char *)H, AD->DMOBJ ); H++ )
	    if( *H == POINT )
		*H = NULLPNTR;
	    else if( *H != NULLPNTR )
		K = 0;
	DmUnfix( AD->DMOBJ );
	if( K )
	  {
	    long SPOWNER;
	    if( IsSpiderMember( BACK, &SPOWNER ) )
		DiscFromSpider( SPOWNER, BACK );
	    else
		AtxDisconnect( BACK );
	    PsMakeGarbage( BACK );
	  }
      } 
  }


/*-------------------------------------------------------- AtxBack ----------*/
/* Get a BACKWARD atom related to the point */

static long  AtxBack( POINT )
    long POINT;
  {
    long SUB;
    int T = AtxType( POINT );
    if( LOGICAL_B <= T && T <= HOLDER &&
				( SUB = AtxSubord( POINT ) ) != NULLPNTR )
	return( AtxNext( SUB ) );
    return( NULLPNTR );
  }


/*----- External ----------------------------------------- AtmDetPointer ---*/
/* Determine a logical pointer of the created atom. The atom head must be
   filled in before invocation of the procedure. */

int AtmDetPointer( DMOBJECT, POINTER )
    unsigned short int DMOBJECT;
    long POINTER;
  {
    atom_head *ATOMMENT;
    if( POINTER != NULLPNTR )
	SkipHolders( POINTER );
    ATOMMENT = (atom_head *)DmPointer( DMOBJECT );
    if( LOGICAL_B <= ATOMMENT->TYPE && ATOMMENT->TYPE <= HOLDER )
      {
	DmLeave( DMOBJECT, 0L, ATOMHEAD );
	((atom_head *)DmPointer( DMOBJECT ))->VALUE = POINTER;
	return( 1 );
      }
    PS_ERROR |= b4; /* b4 unallowed operation during atom creation */
    return( 0 );
  }


/*----- External ----------------------------------------- AtmDetValue ------*/
/* Determine (change) a value of the created atom. The atom head must be
   filled in before invocation of the procedure, by AtmCreate */

int AtmDetValue( DMOBJECT, VALUE, VALSIZE )
    unsigned short int DMOBJECT;
    char *VALUE;
    long VALSIZE;
  {
    int TYPE;
    if( VALSIZE > MAXATOM - ATOMHEAD )
      {
	PS_ERROR |= b4 | b5; /* b4 b5, unallowed operation during creation
					of atom, too large atom */
	return( 0 );
      }
    TYPE = ((atom_head *)DmPointer( DMOBJECT ))->TYPE;
    if( TYPE <= HOLDER )
      {
	PS_ERROR |= b4; /*b4, wrong operation during creation of atom */
	return( 0 );
      }
    else
      {
	DmLeave( DMOBJECT, 0L, ATOMHEAD );
	if( TYPE <= LONGINT_E )
	  {
	    /* If integer */
	    long LONGVALUE = 0L;
	    if( VALUE != NULL )
		memcpy( (char *)&LONGVALUE, VALUE, 4 );
	    ((atom_head *)DmPointer( DMOBJECT ))->VALUE = LONGVALUE;
	  }
	else                                    /* If variable size atom */
	  {
	    ((atom_head *)DmPointer( DMOBJECT ))->VALUE = VALSIZE;
	    DmPush( DMOBJECT, VALUE, (unsigned)VALSIZE ); /*SUN:long->unsigned*/
	  }
      }
    return( 1 );
  }


/*----- External ----------------------------------------- AtmFind ----------*/
/* Find point of atom having given name/type; searching for the given owner */

long AtmFind( OWNER, NAME, TYPE )
    long OWNER;                 /* Owner point - entry to chain of members */
    unsigned short int NAME;            /* Atom name, 0xFFFF if not essential */
    int TYPE;                   /* Atom type; -1 if not essential */
  {
    long MEMBER = NULLPNTR;
    int OWNTYPE;
    if( OWNER == NULLPNTR )
      {
	PS_ERROR |= b14;
	return( NULLPNTR );
      }
    SkipHolders( OWNER );
    OWNTYPE = AtxType( OWNER );
    if( SPIDER_B <= OWNTYPE && OWNTYPE <= SPIDER_E )
      {
	unsigned short int DM_SPIDER = AtxRead( OWNER, 0 )->DMOBJ;
	spider_item *SPI = (spider_item *)( DmFix( DM_SPIDER ) + ATOMHEAD );
	for( ; DmInside( (char *)SPI,  DM_SPIDER ); SPI++ )
	    if( SPI->MEMBER != NULLPNTR )
	      {
		if( TYPE == -1 )
		  {
		    if( SPI->MEMBER_NAME == NAME )
		      {
			MEMBER = SPI->MEMBER;
			break;
		      }
		  }
		else if( NAME == 0xFFFF )
		  {
		    if( SPI->MEMBER_TYPE == TYPE )
		      {
			MEMBER = SPI->MEMBER;
			break;
		      }
		  }
		else
		  {
		    if( SPI->MEMBER_TYPE == TYPE &&  SPI->MEMBER_NAME == NAME )
		      {
			MEMBER = SPI->MEMBER;
			break;
		      }
		  }
	      }
	DmUnfix( DM_SPIDER );
      }
    else
	for( MEMBER = AtxSubord( OWNER ); MEMBER != NULLPNTR;
						MEMBER = AtxNext( MEMBER ) )
	  {
	    char *BGMEM = DmPointer( AtxRead( MEMBER, 1 )->DMOBJ );
	    unsigned short int MEMNAME;
	    MEMNAME = ((atom_head *)BGMEM)->NAME;
	    if( TYPE == -1 )
	      {         /* If member name matches NAME */
		if( MEMNAME == NAME )
		    break;
	      }
	    else if( NAME == 0xFFFF )
	      {         /* If member type matches TYPE */ 
		if( *BGMEM == TYPE )
		    break;
	      }
	    else if( MEMNAME == NAME && *BGMEM == TYPE ) /* If both match */
		break;
	  }
    return( MEMBER );
  }


/*----- External ----------------------------------------- AtmFindNext ------*/
/* Find point of atom having given name and/or type; start searching from
the position next to the given ENTRY point. */

long AtmFindNext( ENTRY, NAME, TYPE )
    long ENTRY;                 /* Point - entry to chain of members */
    unsigned short int NAME;            /* Atom name, 0xFFFF if not essential */
    int TYPE;                   /* Atom type; -1 if not essential */
  {
    long MEMBER = NULLPNTR;
    if( ENTRY == NULLPNTR )
	PS_ERROR |= b14;
    else
      {
	long OWNER;
	SkipHolders( ENTRY );
	/* Check if ENTRY is a member of a spider */
	if( IsSpiderMember( ENTRY, &OWNER ) )
	  {
	    int NOW = 0;
	    unsigned short int DM_SPIDER = AtxRead( OWNER, 0 )->DMOBJ;
	    spider_item *SPI = (spider_item *)( DmFix( DM_SPIDER ) + ATOMHEAD);
	    for( ; DmInside( (char *)SPI,  DM_SPIDER ); SPI++ )
		if( SPI->MEMBER != NULLPNTR )
		  {
		    if( NOW )
		      {
			if( TYPE == -1 )
			  {
			    if( SPI->MEMBER_NAME == NAME )
			      {
				MEMBER = SPI->MEMBER;
				break;
			      }
			  }
			else if( NAME == 0xFFFF )
			  {
			    if( SPI->MEMBER_TYPE == TYPE )
			      {
				MEMBER = SPI->MEMBER;
				break;
			      }
			  }
			else
			  {
			    if( SPI->MEMBER_TYPE == TYPE &&
						SPI->MEMBER_NAME == NAME )
			      {
				MEMBER = SPI->MEMBER;
				break;
			      }
			  }
		      }
		    else if( SPI->MEMBER == ENTRY )
			NOW = 1;
		  }
	    DmUnfix( DM_SPIDER );
	  }
	else
	    for( MEMBER = AtxNext( ENTRY ); MEMBER != NULLPNTR;
						MEMBER = AtxNext( MEMBER ) )
	      {
		char *BGMEM = DmPointer( AtxRead( MEMBER, 1 )->DMOBJ );
		unsigned short int MEMNAME = ((atom_head *)BGMEM)->NAME;
		if( TYPE == -1 )
		  {             /* If member name matches NAME */
		    if( MEMNAME == NAME )
			break;
		  }
		else if( NAME == 0xFFFF )
		  {             /* If member type matches TYPE */ 
		    if( *BGMEM == TYPE )
			break;
		  }
		else if( MEMNAME == NAME && *BGMEM == TYPE ) /* if both match*/
		    break;
	      }
      }
    return( MEMBER );
 }


/*----- External ----------------------------------------- AtmGet -----------*/
/* Get an atom from the given point; return 0 iff error; external */

unsigned short int AtmGet( POINT )
    long POINT;                 /* Point of the atom to be get */
  {
    SkipHolders( POINT );
    return( AtxRead( POINT, 0 )->DMOBJ );
  }


/*----- External ----------------------------------------- AtmGetValue ------*/
/* Get atom's value; external */

unsigned short int AtmGetValue( POINT )
    long POINT;
  {
    SkipHolders( POINT );
    return( AtxGetValue( POINT ) );
  }


/*----- External ---------------------------------------- AtxGetValue ------*/
/* Get atom's value */

unsigned short int  AtxGetValue( POINT )
    long POINT;
  {
    atomdesc *AD = AtxRead( POINT, 0 );
    long VALSIZE = AtxValSize( POINT );
    unsigned short int DMOBJECT = 0;
    if( *DmPointer( AD->DMOBJ ) > HOLDER )
      {
	DMOBJECT = DmCreate( 0 );
	DmPush( DMOBJECT, DmFixTop( AD->DMOBJ ) - VALSIZE, (unsigned)VALSIZE );  /*SUN:long->unsigned*/
	DmUnfix( AD->DMOBJ );
      }
    return( DMOBJECT );
  }


/*----- External ----------------------------------------- AtmInsCpLast -----*/
/* Copy the atom at POINT and insert it after last member of a ring;
   Return point of the atom's copy */

long AtmInsCpLast( POINT, OWNER )
    long POINT;       /* Db pointer to the atom */
    long OWNER;       /* Further owner of the atom, or NULLPNTR */
  {
    unsigned short int NN;
    int TT = AtmTypeAndName( POINT, &NN );
    return( PsDelHolder( AtmAssign( AtmInsLast( AtmCreate( TT, NN ),
							OWNER ), POINT ) ) );
  }


/*----- External ----------------------------------------- AtmInsLast -------*/
/* Insert an atom from DMOBJECT into VPOS as the last in a ring/spider */

long AtmInsLast( DMOBJECT, OWNER )
    unsigned short int DMOBJECT;/* Dynamic memory object with the atom */
    long OWNER;       /* Further owner of the atom, or NULLPNTR */
  {
    long LAST;
    SkipHolders( OWNER );
    LAST = AtxLast( OWNER );
    return( LAST == NULLPNTR ?  AtxInsert( DMOBJECT, OWNER, NULLPNTR ) :
				AtxInsert( DMOBJECT, NULLPNTR, LAST ) );
  }


/*----- External ----------------------------------------- AtmInsert --------*/
/* Insert a atom from DMOBJECT into VPOS. As for AtxInsert; external. */

long AtmInsert( DMOBJECT, OWNER, AFTER )
    unsigned short int DMOBJECT;/* Dynamic memory object with the atom */
    long OWNER;       /* Further owner of the atom, or NULLPNTR */
    long AFTER;       /* Further left neighbour of the atom, or NULLPNTR */
  {
    if( DM_COMPLAIN )
	PsAtxDel( );
    SkipParam( &OWNER, &AFTER );
    return( AtxInsert( DMOBJECT, OWNER, AFTER ) );
  }


/*-------------------------------------------------------- SkipParam --------*/
/* Skip holders or backward atoms for parameters of AtmInsert and AtmMove */

static void  SkipParam( OWNER, AFTER )
    long *OWNER, *AFTER;
  {
    if( *OWNER == NULLPNTR )
      {
	if( *AFTER == NULLPNTR )
	    *AFTER = VPOS_ENTRY;
	else
	  {
	    SkipHolders( *AFTER );
	    if( AtxType( AtxNext( *AFTER ) ) == BACKWARD )
		*AFTER = AtxNext( *AFTER );
	  }
      }
    else
	SkipHolders( *OWNER );
  }


/*-------------------------------------------------------- AtxInsert --------*/
/* Insert atom from DMOBJECT into VPOS. If OWNER != NULLPNTR insertion is made
   into beginning of its subordinated ring; otherwise AFTER is considered.
   If the atom is a logical
   pointer or reference, automatic action on the backward pointer is made.
   CURR_FILE determines the file where the object is inserted. Return point of
   the inserted atom or NULLPNTR iff error. DMOBJECT is captured by VPOS. */

long  AtxInsert( DMOBJECT, OWNER, AFTER )
    unsigned short int DMOBJECT;/* Dynamic memory object with the atom */
    long OWNER;       /* Further owner of the atom, or NULLPNTR */
    long AFTER;       /* Further left neighbour of the atom, or NULLPNTR */
  {  /* The atom is assigned with a free point from the file and then, this
	point is included properly into OWNER or after AFTER.Insertion does not
	mean physical storing of the atom in the file. */
    long POINT, SPOWNER, NEWOWNER;
    atom_head *OB;
    int IS_SPIDER = 0;
    CHANGES_IN_VPOS = 1;
	/* Find garbage area */
    /* For consistency of logical pointers */
    OB = (atom_head *)DmFix( DMOBJECT ); /* For preventing deletion from DM */
    POINT = PsFindArea( DmSize(DMOBJECT), OWNER != NULLPNTR ? OWNER : AFTER );
    AtxAllocate( UPDATED, POINT, DMOBJECT );

    /* Include the atom in a proper VPOS place */
    if( OWNER != NULLPNTR )
      {
	int OWNTYPE = AtxType( OWNER );
	if( SPIDER_B <= OWNTYPE && OWNTYPE <= SPIDER_E )
	  {
	    SPOWNER = OWNER;
	    IS_SPIDER = 1;
	  }
	else if( ! ( RING_B <= OWNTYPE && OWNTYPE <= RING_E ) ) /* Assertion */
	  {
	    PS_ERROR |= b4;
	    PsError();
	  }
      }
    else if( AFTER != NULLPNTR && IsSpiderMember( AFTER, &SPOWNER ) )
	IS_SPIDER = -1;
    NEWOWNER = SPOWNER;
    if( IS_SPIDER )
      {
	((atom_head *)DmPointer( DMOBJECT ))->STATUS |= TOOWNER;
	NEWOWNER =
	    InclToSpider( SPOWNER, POINT, IS_SPIDER < 0 ? AFTER : NULLPNTR );
      }
    else
	AtxInclude( POINT, OWNER, AFTER );
    /* If logical pointer or place holder - include backward pointer */
    /* For consistency of logical pointers */
    if( NEWOWNER != SPOWNER && LOGICAL_B <= OB->TYPE && OB->TYPE <= LOGICAL_E
						&& OB->VALUE == SPOWNER )
	OB->VALUE = NEWOWNER;
    DmUnfix( DMOBJECT );
    AtxInsBack( POINT, OB->TYPE, OB->VALUE );
    return( POINT );
  }


/*-------------------------------------------------------- InclToSpider ----*/
/* Include new point into the spider. The point must be disconnected from
VPOS. If AFTER is NULLPNTR the function includes the point as the first;
otherwise it includes it after AFTER item in the spider owner.
The function returns a new point of the spider's owner. Note that the procedure
is recursive, through AtxPutValue */

static long  InclToSpider( SPOWNER, POINT, AFTER )
    long SPOWNER, POINT, AFTER;
  {
    long NEWOWNER, FROM = 0L;
    unsigned short int SP_ATOM;
    spider_item *SPI;
    atomdesc *AD;
    atom_head *ATH;
    int NOTFOUND;
    /* Assign SPOWNER and TOOWNER flag to POINT atom */
    AD = AtxRead( POINT, 1 );
    AD->FLAGS |= UPDATED;
    ATH = (atom_head *)DmFix( AD->DMOBJ );
    ATH->POINT = SPOWNER;
    ATH->STATUS |= TOOWNER;
	/* Update the owner of the spider */
    SP_ATOM = AtxGetValue( SPOWNER );
    SPI = (spider_item *)DmFix( SP_ATOM );
    if( AFTER != NULLPNTR )
      {
	NOTFOUND = 1;
	for( ; DmInside( (char *)SPI, SP_ATOM ); SPI++,
						FROM += sizeof(spider_item) )
	    if( SPI->MEMBER == AFTER )
	      {
		NOTFOUND = 0;
		break;
	      }
	if( NOTFOUND )
	    PS_ERROR |= b15;
	SPI++;
	FROM += sizeof(spider_item);
      }

    NOTFOUND = 1;
    if( DmInside( (char *)SPI, SP_ATOM ) && SPI->MEMBER == NULLPNTR )
	NOTFOUND = 0; /* Canncelled item is found */
    else if( DmInside( (char *)(SPI+1), SP_ATOM ) && (SPI+1)->MEMBER == NULLPNTR )
      {         /* Cancelled item is after */
	*( SPI+1 ) = *SPI;
	NOTFOUND = 0;
      }
    else if( DmInside( (char *)(SPI-2), SP_ATOM ) && (SPI-2)->MEMBER == NULLPNTR )
      {         /* Cancelled item is before */
	*( SPI-2 ) = *( SPI-1 );
	SPI--;
	NOTFOUND = 0;
      }

    if( NOTFOUND )
      {         /* Cancelled item is not found */
	spider_item *SPIFROM;
	DmUnfix( SP_ATOM );

	/* Distend the atom;
			if FROM != 0, put one cancelled item as a reserve */
	DmDistend( SP_ATOM, FROM, FROM == 0L ?  sizeof(spider_item) :
				sizeof(spider_item) + sizeof(spider_item) );
	SPI = (spider_item *)DmFix( SP_ATOM );
	SPIFROM = (spider_item *)((char *)SPI + FROM );
	if( FROM != 0L )
	    (SPIFROM+1)->MEMBER = NULLPNTR;
	SPIFROM->MEMBER = POINT;
	SPIFROM->MEMBER_TYPE = ATH->TYPE;
	SPIFROM->MEMBER_NAME = ATH->NAME;
	
	/* If the top contains two cancelled items, remove them */
	SPIFROM = (spider_item *)DmTopPointer( SP_ATOM ) - 1;
	DmUnfix( SP_ATOM );     /* Note that SPIFROM refers to SP_ATOM ! */
	if( SPIFROM->MEMBER == NULLPNTR && (SPIFROM-1)->MEMBER == NULLPNTR )
	    DmPop( SP_ATOM, sizeof(spider_item) + sizeof(spider_item) );
      }
    else
      {
	SPI->MEMBER = POINT;
	SPI->MEMBER_TYPE = ATH->TYPE;
	SPI->MEMBER_NAME = ATH->NAME;
	DmUnfix( SP_ATOM );
      }
    DmUnfix( AD->DMOBJ );

	/* Assign new value for the spider owner */
    NEWOWNER = AtxPutValue( SPOWNER, SP_ATOM );

    if( NEWOWNER != SPOWNER )
      {
	/* Change pointers in all members to NEWOWNER */
	atomdesc *ADOWN = AtxRead( NEWOWNER, 0 );
	SPI = (spider_item *)( DmFix( ADOWN->DMOBJ ) + ATOMHEAD );
	for( ; DmInside( (char *)SPI, ADOWN->DMOBJ ); SPI++ )
	    if( SPI->MEMBER != NULLPNTR )
	      {
		AD = AtxRead( SPI->MEMBER, 1 );
		ATH = (atom_head *)DmPointer( AD->DMOBJ );
		AD->FLAGS |= UPDATED;
		ATH->POINT = NEWOWNER;
	      }
	DmUnfix( ADOWN->DMOBJ );
      }
   return( NEWOWNER );
  }


/*----- External ----------------------------------------- AtmIntValue ------*/
/* Return integer value of the atom; if inappropriate, set b14 of PS_ERROR */

long AtmIntValue( POINT )
    long POINT;
  {
    atom_head *ATOM;
    SkipHolders( POINT );
    ATOM = (atom_head *)DmPointer( AtxRead( POINT, 1 )->DMOBJ );
    if( TOEND <= ATOM->TYPE && ATOM->TYPE <= LONGINT_E )
	return( ATOM->VALUE );
    else
      {
	PS_ERROR |= b14;
	return( 0L ); 
      }
  }


/*----- External ----------------------------------------- AtmIntAssign -----*/
/* Assign integer value for the atom; if inappropriate, set b14 of PS_ERROR
   and return 0 */

int AtmIntAssign( POINT, VALUE )
    long POINT;         /* Point of the atom */
    long VALUE;         /* New atom's value */
  {
    int RES = 0;
    char *BG;
    CHANGES_IN_VPOS = 1;
    SkipHolders( POINT );
    BG = DmPointer( UpdDmObject( POINT ) );
    if( TOEND <= *BG && *BG <= LONGINT_E )
      {
	((atom_head *)BG)->VALUE = VALUE;
	RES = 1;
      }
    else
	PS_ERROR |= b14;
    return( RES );
  }


/*-------------------------------------------------------- AtxLast ----------*/
/* Give a last atom, or NULLPNTR if the ring has no members. For rings insert
or update TOEND atom. If OWNER == NULLPNTR, return last atom in the main ring*/

static long  AtxLast( OWNER )
    long OWNER;         /* Owner of the ring */
  {
    long LAST, ENDREF, POINT;
    if( OWNER == NULLPNTR )
	for( POINT = VPOS_ENTRY; POINT != NULLPNTR; POINT = AtxNext( POINT ) )
	    LAST = POINT;
    else
      {
	int TYPE = AtxType( OWNER );
	if( SPIDER_B <= TYPE && TYPE <= SPIDER_E )
	  {
	    atomdesc *AD = AtxRead( OWNER, 0 );
	    spider_item *SPI = (spider_item *)DmFixTop( AD->DMOBJ ) - 1;
	    for( ; DmInside( (char *)SPI - ATOMHEAD, AD->DMOBJ ); SPI-- )
		if( SPI->MEMBER != NULLPNTR )
		  {
		    DmUnfix( AD->DMOBJ );
		    return( SPI->MEMBER );
		  }
	    DmUnfix( AD->DMOBJ );
	    return( NULLPNTR );
	  }
	ENDREF = AtxSubord( OWNER );
	if( ENDREF == NULLPNTR )
	    return( NULLPNTR );
	else if( AtxNext( ENDREF ) == NULLPNTR )
	    return( ENDREF );
	else if( AtxType( ENDREF ) == TOEND )
	    return( AtmIntValue( ENDREF ) );
	else
	  {
	    ENDREF = AtxInsert( AtmCreate( TOEND, 0xFFFF ), OWNER, NULLPNTR );
	    for( POINT=ENDREF; POINT != NULLPNTR; POINT = AtxNext( POINT ) )
		LAST = POINT;
	    AtmIntAssign( ENDREF, LAST );
	  }
      }
    return( LAST );
  }


/*----- External ----------------------------------------- AtmMove ----------*/
/* Disconnect the atom at POINT with all its backward pointers and insert it
   into new place. Similar to AtxInsert. No physical moving of atoms.
   Note that if the owner is moved in such a way that it will be its own
   member, an inconsistency of the structure will appear, in a sense that some
   atoms may be inaccessible from the VPOS_ENTRY. */

void AtmMove( POINT, OWNER, AFTER )
    long POINT;       /* Db pointer to the atom */
    long OWNER;       /* Further owner of the atom, or NULLPNTR */
    long AFTER;       /* Further left neighbour of the atom, or NULLPNTR */
  {
    SkipHolders( POINT );
    SkipParam( &OWNER, &AFTER );
    AtxMove( POINT, OWNER, AFTER );
  }


/*----- External ----------------------------------------- AtmMoveLast ------*/
/* Disconnect the atom at POINT with all its backward pointers and insert it
   as a last member in the ring or spider. No physical moving of atoms.
   Note that if the owner is moved in such a way that it will be its own
   member, an inconsistency of the structure will appear, in a sense that some
   atoms may be inaccessible from the VPOS_ENTRY. */

void AtmMoveLast( POINT, OWNER )
    long POINT;       /* Db pointer to the atom */
    long OWNER;       /* Further owner of the atom, or NULLPNTR */
  {
    long LAST;
    SkipHolders( POINT );
    SkipHolders( OWNER );
    LAST = AtxLast( OWNER );
    if( LAST == NULLPNTR )
	AtxMove( POINT, OWNER, NULLPNTR );
    else
	AtxMove( POINT, NULLPNTR, LAST );
  }


/*-------------------------------------------------------- AtxMove ----------*/
/* Disconnect the atom at POINT with all its backward pointers and insert it
   into new place. Similar to AtxInsert; no physical moving of atoms. */

static void  AtxMove( POINT, OWNER, AFTER )
    long POINT;       /* Db pointer to the atom */
    long OWNER;       /* Further owner of the atom, or NULLPNTR */
    long AFTER;       /* Further left neighbour of the atom, or NULLPNTR */
  {
    long OLDOWNER, POSSIBACK, SPOWNER;
    int SPINC, POSS_BACK_TYPE;
    /* Transformation of parameters */
    CHANGES_IN_VPOS = 1;
    POSSIBACK = AtxNext( POINT );
    POSS_BACK_TYPE = POSSIBACK == NULLPNTR ? -1 : AtxType( POSSIBACK );
    if( POINT == OWNER || OWNER == NULLPNTR && POINT == AFTER ||
			POSSIBACK == AFTER && POSS_BACK_TYPE == BACKWARD )
							/* Do nothing */
	return;
    /* Disconnect the atom and its backward pointers from its old place;
	different action for rings and spiders */
    /* Check if the moved atom has backward pointers */
    if( POSS_BACK_TYPE != BACKWARD )
	POSSIBACK = NULLPNTR;
    if( IsSpiderMember( POINT, &OLDOWNER ) )
	DiscFromSpider( OLDOWNER, POINT );
    else
      {
	if( POSSIBACK != NULLPNTR )
	    AtxDiscNext( POINT );
	AtxDisconnect( POINT );
      }
    /* Test if inclusion has to be made into a spider */
    SPINC = 0;
    if( OWNER != NULLPNTR )
      {
	int OWNTYPE = AtxType( OWNER );
	if( SPIDER_B <= OWNTYPE && OWNTYPE <= SPIDER_E )
	  {
	    SPINC = 1;
	    SPOWNER = OWNER;
	    AFTER = NULLPNTR;
	  }
      }
    else if( AFTER != NULLPNTR && IsSpiderMember( AFTER, &SPOWNER ) )
	SPINC = 1;

    if( SPINC )
      {
	SPOWNER = InclToSpider( SPOWNER, POINT, AFTER );
	if( POSSIBACK != NULLPNTR )
	    InclToSpider( SPOWNER, POSSIBACK, POINT );
      }
    else
      {
	/* Connect the atom and its backward pointers in the new place */
	AtxInclude( POINT, OWNER, AFTER );
	if( POSSIBACK != NULLPNTR )
	    AtxInclude( POSSIBACK, NULLPNTR, POINT );
      }
  }


/*----- External ----------------------------------------- AtmName ----------*/
/* Give a name associated with the point */

unsigned short int AtmName( POINT )
    long POINT;
  {
    char *ATOM = DmPointer( AtxRead( POINT, 1 )->DMOBJ );
    if( *ATOM == HOLDER )
      {
	SkipHolders( POINT );
	ATOM = DmPointer( AtxRead( POINT, 1 )->DMOBJ );
      }
    return( ((atom_head *)ATOM)->NAME );
  }


/*----- External ----------------------------------------- AtmNext ----------*/
/* Give a next atom, or NULLPNTR if the end; external */

long AtmNext( POINT )
    long POINT;
  {
    if( DM_COMPLAIN )
	PsAtxDel( );
    SkipHolders( POINT );
    POINT = AtxNext( POINT );
    if( AtxType( POINT ) == BACKWARD )
	POINT = AtxNext( POINT );
    return( POINT );
  }


/*-------------------------------------------------------- AtxNext ----------*/
/* Give a pointer to the next atom, or NULLPNTR if the end */

long  AtxNext( POINT )
    long POINT;
  {
    atom_head *ATOM = (atom_head *)DmPointer( AtxRead( POINT, 1 )->DMOBJ );
    if( ATOM->STATUS & TOOWNER )
      {         /* Test if the atom is a member of a spider */
	long SPOWNER = ATOM->POINT;
	int OWN_TYPE = AtxType( SPOWNER );
	if( SPIDER_B <= OWN_TYPE && OWN_TYPE <= SPIDER_E )
	  { 
	    int K = 1;
	    unsigned short int SP_ATOM = AtxRead( SPOWNER, 0 )->DMOBJ;
	    spider_item *SPI = (spider_item *)( DmFix( SP_ATOM ) + ATOMHEAD );
	    for(; DmInside( (char *)SPI, SP_ATOM ); SPI++ )
		if( K )
		  {
		    if( SPI->MEMBER == POINT )
			K = 0;
		  }
		else if( SPI->MEMBER != NULLPNTR )
		  {
		    DmUnfix( SP_ATOM );
		    return( SPI->MEMBER );
		  }
	    DmUnfix( SP_ATOM );
	  }
	return( NULLPNTR );
      }
    else
	return( ATOM->POINT );
  }


/*-------------------------------------------------------- AtxPriNexOwn -----*/
/* Give next, owner, and prior atoms; return prior */

static long  AtxPriNexOwn( POINT, NEXT, OWNER )
    long POINT;
    long *NEXT;
    long *OWNER;
  {
    atom_head *ATOM;
    long P;
    ATOM = (atom_head *)DmPointer( AtxRead( POINT, 1 )->DMOBJ );
    P = ATOM->POINT;
    /* Get next */
    *NEXT = ATOM->STATUS & TOOWNER ? NULLPNTR : P;
    /* Get owner */
    while( P != NULLPNTR && ! ( ATOM->STATUS & TOOWNER ) )
      {
	ATOM = (atom_head *)DmPointer( AtxRead( P, 1 )->DMOBJ );
	P = ATOM->POINT;
      }
    *OWNER = P;
    /* Get prior */
    P = P == NULLPNTR ? VPOS_ENTRY : AtxSubord( P );
    if( P == POINT )
	return( NULLPNTR );
    else
      {
	do
	  {
	    ATOM = (atom_head *)DmPointer( AtxRead( P, 1 )->DMOBJ );
	    if( ATOM->POINT == POINT )
		return( P );
	    P = ATOM->POINT;
	  }
	while( 1 );
      }
  }


/*----- External ----------------------------------------- AtmOwner ---------*/
/* Give the point of owner, or NULLPNTR if FOV atom; external */

long AtmOwner( POINT )
    long POINT;
  {
    long OWNER;
    SkipHolders( POINT );
    if( ! IsSpiderMember( POINT, &OWNER ) )
      {
	atom_head *ATOM;
	do
	  {
	    ATOM = (atom_head *)DmPointer( AtxRead( POINT, 1 )->DMOBJ );
	    POINT = ATOM->POINT;
	  }
	while( ! ( POINT == NULLPNTR || ATOM->STATUS & TOOWNER ) );
	return( POINT );
      }
    return( OWNER );
  }


/*-------------------------------------------------------- AtxOwner ---------*/
/* Give a point of the owner atom, or NULLPNTR if no owner */

long  AtxOwner( POINT )
    long POINT;
  {
    long OWNER;
    if( ! IsSpiderMember( POINT, &OWNER ) )
      {
	atom_head *ATOM;
	do
	  {
	    ATOM = (atom_head *)DmPointer( AtxRead( POINT, 1 )->DMOBJ );
	    POINT = ATOM->POINT;
	  }
	while( ! ( POINT == NULLPNTR || ATOM->STATUS & TOOWNER ) );
	return( POINT );
      }
    return( OWNER );
  }


/*----- External ----------------------------------------- AtmPntAssign -----*/
/* Assign a new pointer for the atom; return 0 iff error. Make automatic
   action on a backward pointer. No physical movement of atoms. */

int AtmPntAssign( POINT, POINTER, NEWTYPE )
    long POINT;
    long POINTER;
    int NEWTYPE;
  {
    if( POINTER != NULLPNTR )
	SkipHolders( POINTER );
    SkipHolders( POINT );
    return( AtxPntAssign( POINT, POINTER, NEWTYPE ) );
  }


/*-------------------------------------------------------- AtxPntAssign -----*/
/* Assign a new pointer to the atom; return 0 iff error. Make automatic
   action on the backward pointer. No physical moving of atoms. */

static int  AtxPntAssign( POINT, POINTER, NEWTYPE )
    long POINT;
    long POINTER;
    int NEWTYPE;
  {
    int TYPE = AtxType( POINT );
    CHANGES_IN_VPOS = 1;
    if( LOGICAL_B <= TYPE && TYPE <= HOLDER || AtxSubord( POINT ) == NULLPNTR )
      {
	long BACK = AtxBack( POINT );
	unsigned short int DMOBJ = UpdDmObject( POINT );
	long DIFF = AtxSize( POINT ) - ATOMHEAD;
	((atom_head *)DmPointer( DMOBJ ))->VALUE = POINTER;
	/* Change TYPE into NEWTYPE */
	((atom_head *)DmPointer( DMOBJ ))->TYPE = NEWTYPE;
	if( DIFF > 0L )
	  {
	    /* Return unnecessary part to garbage */
	    PsGarbage( AtxFile( POINT ), AtxPoint( POINT ) + ATOMHEAD, DIFF );
	    DmLeave( DMOBJ, 0L, ATOMHEAD );
	  }
	AtxDelBack( BACK, POINT );
	AtxInsBack( POINT, NEWTYPE, POINTER );
	if( TYPE != NEWTYPE )
	     ChangeTypeInSpider( POINT, NEWTYPE );
	return( 1 );
      }
    PS_ERROR |= b6;
    return( 0 );
  }


/*----- External ---------------------------------------- AtmPoint ----------*/
/* Return relative position of the atom in the file; external */

long AtmPoint( POINT )
    long POINT;
  {
    SkipHolders( POINT );
    return( AtxPoint( POINT ) );
  }


/*----- External ----------------------------------------- AtmPutValue ------*/
/* Put the new value contained in DM object to the atom; return NULLPNTR if
   unsuccessful and new atom pointer otherwise */

long AtmPutValue( POINT, NEWOBJECT )
    long POINT;
    unsigned short int NEWOBJECT;
  {
    SkipHolders( POINT );
    DelSubordinates( POINT );
    return( AtxPutValue( POINT, NEWOBJECT ) );
  }


/*-------------------------------------------------------- AtxPutValue ------*/
/* Put the new value contained in DM object to the atom; return NULLPNTR if
   unsuccessful and new atom point otherwise */

long  AtxPutValue( POINT, NEWOBJECT )
    long POINT;
    unsigned short int NEWOBJECT;
  {
    unsigned short int DMOBJ = UpdDmObject( POINT );
    atom_head *BGNEW, *BGOLD = (atom_head *)DmFix( DMOBJ );
    int TYPE = BGOLD->TYPE;
    if( TYPE >= TOEND )
      {
	long NEWSIZE = DmSize( NEWOBJECT );
	if( TYPE <= LONGINT_E )
	  {
	    if( NEWSIZE != 4L )
		PS_ERROR |= b6;
	    BGOLD->VALUE = *(long *)DmFix( NEWOBJECT );
	    DmUnfix( DMOBJ );
	    DmUnfixDelete( NEWOBJECT );
	    return( POINT );
	  }
	else
	  {
	    long ATOMSIZE = NEWSIZE + ATOMHEAD;
	    long OLDSIZE = BGOLD->VALUE;
	    /* Augment NEWOBJECT with organizational information */
	    DmDistend( NEWOBJECT, 0L, ATOMHEAD );
	    BGNEW = (atom_head *)DmFix( NEWOBJECT );
	    memcpy( (char *)BGNEW, (char *)BGOLD, ATOMHEAD );
	    BGNEW->VALUE = NEWSIZE;
	    DmUnfix( DMOBJ );
	    DmUnfix( NEWOBJECT );
	    if( OLDSIZE >= NEWSIZE || PsFindAfter( POINT, NEWSIZE - OLDSIZE ) )
	      {
		atomdesc *AD;
		/* If atom size is not increased or reserve is sufficient */
		if( OLDSIZE > NEWSIZE )
		    /* Return redundant bytes to garbage */
		    PsGarbage( AtxFile( POINT ), AtxPoint( POINT ) + ATOMSIZE,
				OLDSIZE - NEWSIZE );
		AD = AtxLocate( POINT );
		AD->DMOBJ = NEWOBJECT;
		AD->FLAGS &= ~b2;
		DmDelete( DMOBJ );
		return( POINT );
	      }
	    else  /* If the atom size is increased and insufficient reserve */
	      {
		long NEWPOINT = FindAreaUpdRef( POINT, ATOMSIZE );
		AtxAllocate( UPDATED, NEWPOINT, NEWOBJECT );
		if( TYPE == BACKWARD )
		    PsMakeGarbage( POINT );
		else
		  { /* Return area to garbage; leave space for place holder */
		    PsGarbage(CURR_FILE, AtxPoint(POINT) + ATOMHEAD, OLDSIZE );
		    /* Set HOLDER at POINT, pointing NEWPOINT */
		    AtxLeaveHolder( POINT, NEWPOINT );
		    DmDelete( DMOBJ );
		  }
		return( NEWPOINT );
	      }
	  }
      }
    DmUnfix( DMOBJ );
    PS_ERROR |= b6;     /* b6 - unallowed assignment to atom value */ 
    return( NULLPNTR );
  }


/*-------------------------------------------------------- FindAreaUpdRef ---*/
/* Find area of size ATOMSIZE and update all references to POINT */

static long  FindAreaUpdRef( POINT, ATOMSIZE )
    long POINT, ATOMSIZE;
  {
    long *REF, SPOWNER, NEWPOINT;
    unsigned short int WORKOBJ;
    CURR_FILE = AtxFile( POINT );
    NEWPOINT = PsFindArea( ATOMSIZE, POINT );
    if( IsSpiderMember( POINT, &SPOWNER ) )
      {
	atomdesc *AD = AtxRead( SPOWNER, 0 );
	unsigned short int SP_ATOM = AD->DMOBJ;
	spider_item *SPI = (spider_item *)( DmFix( SP_ATOM ) + ATOMHEAD );
	AD->FLAGS |= UPDATED;
	for(; DmInside( (char *)SPI, SP_ATOM ); SPI++ )
	    if( SPI->MEMBER == POINT )
	      {
		SPI->MEMBER = NEWPOINT;
		break;
	      }
	SPI++;
	WORKOBJ = DmInside((char *)SPI,SP_ATOM) && SPI->MEMBER != NULLPNTR &&
					SPI->MEMBER_NAME == 0xFFF8 ?
		AtxGetValue( SPI->MEMBER ) : DmCreate( 0 );
	DmUnfix( SP_ATOM );
      }
    else
      {
	long AFTER, UPDPOINT, BACK;
	AFTER = AtxPriNexOwn( POINT, &BACK, &UPDPOINT );
	/* Determine points of updating in WORKOBJ */
	WORKOBJ = AtxType( BACK ) == BACKWARD ?
				AtxGetValue( BACK ) : DmCreate( 0 );
	if( AFTER == NULLPNTR )
	    DmPush( WORKOBJ, (char *)&UPDPOINT, 4 );
	if( AtxNext( POINT ) == NULLPNTR ) /* Change TOEND */
	  {
	    long OWNER = AtxOwner( POINT );
	    if( OWNER != NULLPNTR )
	      {
		long TOENDREF = AtxSubord( OWNER );
		if( AtxType( TOENDREF ) == TOEND )
		    AtmIntAssign( TOENDREF, NEWPOINT );
	      }
	  }
	/* Update AFTER */
	if( AFTER != NULLPNTR )
	    ((atom_head *)DmPointer( UpdDmObject( AFTER )))->POINT = NEWPOINT;
      }

    /* Update all references */
    for( REF = (long *)DmFix( WORKOBJ );DmInside((char *)REF, WORKOBJ ); REF++)
	if( *REF != NULLPNTR )
	    ((atom_head *)DmPointer( UpdDmObject( *REF )))->VALUE = NEWPOINT;
    DmUnfixDelete( WORKOBJ );
    return( NEWPOINT );
  }


/*----- External ----------------------------------------- AtmRename --------*/
/* Rename a atom; if it is a member of a spider, change properly its dir */

void AtmRename( POINT, NAME )
    long POINT;
    unsigned short int NAME;
  {
    long SPOWNER;
    SkipHolders( POINT );
    ((atom_head *)DmPointer( UpdDmObject( POINT )))->NAME = NAME;
    if( IsSpiderMember( POINT, &SPOWNER ) )
      {
	atomdesc *AD = AtxRead( SPOWNER, 0 );
	spider_item *SPI = (spider_item *)( DmFix( AD->DMOBJ ) + ATOMHEAD );
	AD->FLAGS |= UPDATED;
	for( ; DmInside( (char *)SPI, AD->DMOBJ ); SPI++ )
	    if( SPI->MEMBER == POINT )
	      {
		SPI->MEMBER_NAME = NAME;
		break;
	      }
	DmUnfix( AD->DMOBJ );
      }
  }


/*-------------------------------------------------------- AtxSize ----------*/
/* Return the atom size */

long  AtxSize( POINT )
    long POINT;
  {
    atom_head *ATOM = (atom_head *)DmPointer( AtxRead( POINT, 1 )->DMOBJ );
    return( ATOM->TYPE <= LONGINT_E ? (long)ATOMHEAD
				    : ATOM->VALUE + ATOMHEAD );
  }


/*----- External ----------------------------------------- AtmValSize -------*/
/* Return the size of value of the atom. -1L iff error; external */

long AtmValSize( POINT )
    long POINT;
  {
    SkipHolders( POINT );
    return( AtxValSize( POINT ) );
  }


/*-------------------------------------------------------- AtxValSize -------*/
/* Return the size of value of the atom. -1L iff error */

static long  AtxValSize( POINT )
    long POINT;
  {
    atom_head *ATOM = (atom_head *)DmPointer( AtxRead( POINT, 1 )->DMOBJ );
    return( ATOM->TYPE <= HOLDER  ? -1L :
	  ( ATOM->TYPE <= LONGINT_E ?  4L : ATOM->VALUE ) );
  }


/*----- External ----------------------------------------- AtmSubord --------*/
/* Give first subordinated atom or navigate via a logical pointer.
   Return NULLPNTR iff not applicable; external */

long AtmSubord( POINT )
    long POINT;
  {
    unsigned short int DMOBJ;
    atom_head *ATOM;
    long RES = NULLPNTR;
    if( DM_COMPLAIN )
	PsAtxDel( );
    DMOBJ = AtxRead( POINT, 0 )->DMOBJ;
    ATOM = (atom_head *)DmPointer( DMOBJ );
    if( ATOM->TYPE == HOLDER )
      {
	SkipHolders( POINT );
	DMOBJ = AtxRead( POINT, 0 )->DMOBJ;
	ATOM = (atom_head *)DmPointer( DMOBJ );
      }
    if( ATOM->TYPE < HOLDER )
	RES = ATOM->VALUE;
    else if( SPIDER_B <= ATOM->TYPE && ATOM->TYPE <= SPIDER_E )
      {
	spider_item *SPI = (spider_item *)(DmFix( DMOBJ ) + ATOMHEAD );
	for( ; DmInside( (char *)SPI, DMOBJ ); SPI++ )
	    if( SPI->MEMBER != NULLPNTR )
	      {
		RES = SPI->MEMBER;
		break;
	      }
	DmUnfix( DMOBJ );
      }
    if( RES != NULLPNTR && AtxType( RES ) == TOEND )
	RES = AtxNext( RES );
    return( RES );
  }


/*-------------------------------------------------------- AtxSubord ------*/
/* Give the first subordinated atom or the value of a logical pointer. Return
NULLPNTR iff not applicable */

long  AtxSubord( POINT )
    long POINT;
  {
    atom_head *ATOM = (atom_head *)DmPointer( AtxRead( POINT, 1 )->DMOBJ );
    if(  ATOM->TYPE <= HOLDER )
	return( ATOM->VALUE );
    else if( SPIDER_B <= ATOM->TYPE && ATOM->TYPE <= SPIDER_E )
      {
	unsigned short int DMOBJ = AtxRead( POINT, 0 )->DMOBJ;
	spider_item *SPI = (spider_item *)(DmFix( DMOBJ ) + ATOMHEAD );
	for( ; DmInside( (char *)SPI, DMOBJ ); SPI++ )
	    if( SPI->MEMBER != NULLPNTR )
	      {
		DmUnfix( DMOBJ );
		return( SPI->MEMBER );
	      }
	DmUnfix( DMOBJ );
      }
    return( NULLPNTR );
  }


/*----- External ----------------------------------------- AtmType ---------*/
/* Give type of the atom; external */

int AtmType( POINT )
    long POINT;
  {
    int T = AtxType( POINT );
    if( T == HOLDER )
      {
	SkipHolders( POINT );
	T = AtxType( POINT );
      }
    return( T );
  }


/*----- External ----------------------------------------- AtmTypeAndName --*/
/* Give type and name of the atom; external */

int AtmTypeAndName( POINT, NAME )
    long POINT;
    unsigned short int *NAME;
  {
    char *BGA = DmPointer( AtxRead( POINT, 1 )->DMOBJ );
    if( *BGA == HOLDER )
      {
	SkipHolders( POINT );
	BGA = DmPointer( AtxRead( POINT, 1 )->DMOBJ );
      }
    *NAME = ((atom_head *)BGA)->NAME;
    return( *BGA );
  }


/*----- External ----------------------------------------- AtmUpdated -------*/
/* Set UPDATED flag for atom; no action for atoms stored on secondary files */

void AtmUpdated( POINT )
    long POINT;
  {
    atomdesc *AD;
    SkipHolders( POINT );
    CHANGES_IN_VPOS = 1;
    if( ( AD = AtxLocate( POINT ) ) != NULL )
	AD->FLAGS |= UPDATED;
  }


/*----- External ----------------------------------------- AtmValAssign -----*/
/* Assign a given value; return NULLPNTR if unsuccessful or new atom point
   otherwise. If the atom is located in new place, leave place holder in old */

long AtmValAssign( POINT, TYPE, VALUE, VALSIZE )
    long POINT;
    int TYPE;
    char *VALUE;
    long VALSIZE;
  {
    SkipHolders( POINT );
    /* Remove subordinates for RING or SPIDER atoms */
    DelSubordinates( POINT );
    return( AtxValAssign( POINT, TYPE, VALUE, VALSIZE ) );
  }



/*-------------------------------------------------------- AtxValAssign -----*/
/* Assign a given value; return NULLPNTR if unsuccessful and new atom pointer
   otherwise; if the atom is located in new place, leave place holder in old */

static long  AtxValAssign( POINT, TYPE, VALUE, VALSIZE )
    long POINT;
    int  TYPE;
    char *VALUE;
    long VALSIZE;
  {
    unsigned short int DMOBJ;
    long OLDSIZE, NEWSIZE;
    atomdesc *AD = AtxRead( POINT, 1 );
    atom_head *AH;
    CHANGES_IN_VPOS = 1;
    AD->FLAGS |= UPDATED;
    DMOBJ = AD->DMOBJ;
    AH = (atom_head *)DmFix( DMOBJ );
    if( AH->TYPE < TOEND && AtxSubord( POINT ) != NULLPNTR )
      {
	PS_ERROR |= b6;
	DmUnfix( DMOBJ );
	return( NULLPNTR );
      }
    if( VALSIZE + ATOMHEAD > MAXATOM )
      {
	PS_ERROR |= b5;
	DmUnfix( DMOBJ );
	return( NULLPNTR );
      }
    OLDSIZE = AtxSize( POINT ); /* Old size of an atom */
    NEWSIZE = TYPE <= LONGINT_E ? (long)ATOMHEAD : VALSIZE + ATOMHEAD;
    if( OLDSIZE >= NEWSIZE || PsFindAfter( POINT, NEWSIZE - OLDSIZE ) )
      {/* If the atom size is not increased or the reserve is sufficient */
	if( OLDSIZE > NEWSIZE )
	    /* Return redundant bytes to garbage */
	    PsGarbage( AtxFile( POINT ), AtxPoint( POINT ) + NEWSIZE,
				OLDSIZE - NEWSIZE );
	/* Update physically DMOBJ */
	AH->TYPE = TYPE;
	AD->FLAGS &= ~b2;       /* Cancelling the flag "head only" */
	DmUnfix( DMOBJ );
	DmLeave( DMOBJ, 0L, ATOMHEAD );
	if( RING_B <= TYPE && TYPE <= RING_E )
	    ((atom_head *)DmPointer( DMOBJ ))->VALUE = NULLPNTR;
	else if( TYPE >= BYTES_B )
	  {
	    ((atom_head *)DmPointer( DMOBJ ))->VALUE = VALSIZE;
	    DmPush( DMOBJ, VALUE, (unsigned)VALSIZE ); /*SUN: long->unsigned*/
	  }
	else
	  {
	    long LONGVALUE = 0L; /* For alignment problems */
	    if( VALUE != NULL )
		memcpy( (char *)&LONGVALUE, VALUE, 4 );
	    if( LOGICAL_B <= TYPE && TYPE <= HOLDER )
		((atom_head *)DmPointer( DMOBJ ))->VALUE = 
			  VALSIZE == 0 ? NULLPNTR: LONGVALUE;
	    else /* if( TYPE <= LONGINT_E ) */
		((atom_head *)DmPointer( DMOBJ ))->VALUE = LONGVALUE;
	  }
	return( POINT );
      }
    else        /* If the atom size is increased and insufficient reserve */
      {
	unsigned short int NEWOBJECT = DmCreate( 0 );
	DmPush( NEWOBJECT, VALUE, (unsigned)VALSIZE ); /*SUN: long->unsigned*/
	if( OLDSIZE == ATOMHEAD )
	  {
	    AH->TYPE = TYPE;
	    AH->VALUE = 0L;
	  }
	DmUnfix( DMOBJ );
	POINT = AtxPutValue( POINT, NEWOBJECT );
	return( POINT );
      }
/* Never executed
    DmUnfix( DMOBJ );
    PS_ERROR |= b6;
    return( NULLPNTR );
*/
  }


/*----- External ----------------------------------------- AtmChangeSize ----*/
/* Change the size of atom's value;return NULLPNTR if unsuccessful or new atom
   point otherwise. If the atom is located in new place, leave holder in old.*/

long AtmChangeSize( POINT, VALSIZE )
    long POINT;
    long VALSIZE;
  {
    SkipHolders( POINT );
    return( AtxChangeSize( POINT, VALSIZE ) );
  }


/*-------------------------------------------------------- AtxChangeSize ----*/
/* Change the size of atom's value;return NULLPNTR if unsuccessful or new atom
   point otherwise. If the atom is located in new place, leave holder in old.*/

static long  AtxChangeSize( POINT, VALSIZE )
    long POINT;
    long VALSIZE;
  {
    unsigned short int DMOBJ;
    long OLDSIZE, NEWSIZE;
    atomdesc *AD = AtxRead( POINT, 0 );
    int TYPE = AtxType( POINT );
    CHANGES_IN_VPOS = 1;
    AD->FLAGS |= UPDATED;
    DMOBJ = AD->DMOBJ;
    if( VALSIZE + ATOMHEAD > MAXATOM )
      {
	PS_ERROR |= b5;
	return( NULLPNTR );
      }
    if( TYPE <= LONGINT_E )
      {
	PS_ERROR |= b14;
	return( NULLPNTR );
      }
    OLDSIZE = AtxSize( POINT ); /* Old size of the atom */
    NEWSIZE = VALSIZE + ATOMHEAD;
    if( OLDSIZE == NEWSIZE )
	return( POINT );
    DmChangeSize( DMOBJ, NEWSIZE );
	/* The DMOBJ is fixed to prevent sending it to a VPOS file */
    ((atom_head *)DmFix( DMOBJ ))->VALUE = VALSIZE;
    if( OLDSIZE > NEWSIZE )
	/* If atom size is not increased, return redundant bytes to garbage */
	PsGarbage( AtxFile( POINT ), AtxPoint( POINT ) + NEWSIZE,
				OLDSIZE - NEWSIZE );
    else if( ! PsFindAfter( POINT, NEWSIZE - OLDSIZE ) )
      { /* Insufficient reserve; find new area for the atom */
	long NEWPOINT = FindAreaUpdRef( POINT, NEWSIZE );
	AtxAllocate( UPDATED, NEWPOINT, DMOBJ );
	/* Return area to garbage */
	if( TYPE == BACKWARD )
	  {
	    PsGarbage( CURR_FILE, AtxPoint( POINT ), OLDSIZE );
	    AtxDeAllocate( POINT );
	  }
	else
	  {
	    PsGarbage( CURR_FILE, AtxPoint(POINT) + ATOMHEAD,
							OLDSIZE - ATOMHEAD );
	    /* Set HOLDER at POINT, pointing NEWPOINT */
	    AtxLeaveHolder( POINT, NEWPOINT );
	  }
	DmUnfix( DMOBJ );
	return( NEWPOINT );
      }
    DmUnfix( DMOBJ );
    return( POINT );
  }


/*-------------------------------------------------------- PsError ---------*/
/* Reaction on a fatal error */

static void  PsError( )
  {
    if( PS_ERROR )
      {
	int I;
	unsigned short int MASK = 0x0001;
	DmCheckCons();
	if( PS_ERROR & b3 )
	    printf( "File writing error" );
	if( PS_ERROR & b2 )
	    printf( "File reading error" );
	if( PS_ERROR & b1 )
	    printf( "Cannot open the file" );
	if( PS_ERROR & b12 )
	    printf( "Cannot close a file" );
	printf(": PS_ERROR");
	for( I = 0; I < 16; I++ )
	  {
	    if( PS_ERROR & MASK )
		printf( " b%d", I );
	    MASK = MASK << 1;
	  }
	printf(", cannot continue\n");
	getchar();
	exit(1); 
      }
  }


/*----- External ----------------------------------------- PsCancUpdates ---*/
/* Cancel all updates made so far */

int PsCancUpdates( )
  {
    int E = 1;
    char MAINNAME[ FMSIZE ];
    atomdesc *INV;
    unsigned INVS;
    filedesc *FD;
    /* Remove all DM objects containing atoms */
    CURR_FPOINTER = NULL;
    ACTUAL_FILE = -1;
    for( INV = BEGAMAIN, INVS = 0; INVS < AINVSIZE; INV++, INVS++ )
      {
	if( ! ( INV IS_CANCELLED || INV->FLAGS & OVERFLOW ) )
	    DmDelete( INV->DMOBJ );
	INV->FLAGS = CANCELLED;
      }
    for( INV = BEGAOVER; DmInside( (char *)INV, AINVOVER ); INV++ )
      {
	if( ! ( INV IS_CANCELLED ) )
	  {
	    DmDelete( INV->DMOBJ );
	    INV->FLAGS = CANCELLED;
	  }
      }
    FD = (filedesc *)DmFix( FILETABLE );
    strcpy( MAINNAME, FD->DESC.PATHNODE );
/*
    strmfe( MAINNAME, FD->DESC.PATHNODE, FD->DESC.EXT );
*/
	/* Remove all apparent and new files, close old files */
    for(; DmInside( (char *)FD, FILETABLE ); FD++ )
	if( ! ( FD IS_CANCELLED ) )
	  {
	    if( FD->GARBAGE )
		DmDelete( FD->GARBAGE );
	    if( FD->FPOINTER != NULL )
	      {
		if( fclose( FD->FPOINTER ) == -1 )
		  {
		    PS_ERROR |= b12;
		    PsError();
		  }
	      }
	    if( FD->FLAGS & NEWFILE )
	      {
		char FNAME[ FMSIZE ];
		strcpy( FNAME, FD->DESC.PATHNODE );
		strcat( FNAME, WORK_EXT );
/*
		strmfe( FNAME, FD->DESC.PATHNODE, WORK_EXT );
*/
		unlink( FNAME );
	      }
	  }
    DmUnfix( FILETABLE );
    return( E );
  }


/*----- External ----------------------------------------- PsClose ----------*/
/* Close physical data structure */

int PsClose(  )
  {
    filedesc *FD, *FITEM;
    DmDelete( PsNoHolder( ) );    /* Remove all place holders */
    FD = (filedesc *)DmFix( FILETABLE );
    /* Remove all temporary files */
    for( FITEM = FD + 1; DmInside( (char *)FITEM, FILETABLE ); FITEM++ )
	if( ! ( FITEM IS_CANCELLED ) && FITEM->FLAGS & TEMPORARY )
	    PsDeleteFile( (int)(FITEM - FD) );
    /* Shorten the file table */
    while( ( (filedesc *)( DmTopPointer( FILETABLE ) ) - 1 ) IS_CANCELLED )
      {
	DmUnfix( FILETABLE );
	DmPop( FILETABLE, sizeof( filedesc ) );
	FD = (filedesc *)DmFix( FILETABLE );
      }
    /* Close permanent files */
    for( FITEM = FD + 1; DmInside( (char *)FITEM, FILETABLE ); FITEM++ )
	if( ! (FITEM IS_CANCELLED ) && FITEM IS_UPDATED )
	    PsCloseFile( (int)(FITEM - FD) );
    if( CHANGES_IN_VPOS )
	/* Store updates on the main file. If the file is apparent, all
	   organizational data must be created and then included into proper
	   rings; otherwise their points should be found and updated. The
	   action concerns: permanent names, file table, and garbage
	   descriptors. Garbage description for the first file is stored a few
	   times, to take into account updates of it when storing it. */
      {
	long OWNGAR, POINT, PNPOINT;
	unsigned short int DM_ATOM, FITAB;
	unsigned COPYOFNAMES;
	CURR_FILE = 0;

	/* Store boundaries of types of atoms */
	PNPOINT = AtmFind( VPOS_ENTRY, 0xFFF9, -1 );
	if( PNPOINT == NULLPNTR ) /* If such an atom does not exist */
	  {
	    unsigned char BOUNDS[ 4 ];
	    BOUNDS[ 0 ] = RING_E;
	    BOUNDS[ 1 ] = LOGICAL_E;
	    BOUNDS[ 2 ] = LONGINT_E;
	    BOUNDS[ 3 ] = BYTES_E;
	    DM_ATOM = AtmCreate( BYTES_E, 0xFFF9 );
	    AtmDetValue( DM_ATOM, BOUNDS, 4L );
	    PNPOINT = AtxInsert( DM_ATOM, VPOS_ENTRY, NULLPNTR );
	  }
	/* Find/create root of garbage descriptors; delete garbage
		descriptor for the main file */
	if( ( OWNGAR = AtmFind( VPOS_ENTRY, 0xFFFB, -1 ) ) == NULLPNTR )
	    OWNGAR = AtmCreateAndInsert( RING_B, 0xFFFB, VPOS_ENTRY, NULLPNTR );
	else
	    AtmDelete( AtmSubord( OWNGAR ) );

	/* Write permanent names */
	if( ( PNPOINT = AtmFind( VPOS_ENTRY, 0xFFFD, -1 ) ) == NULLPNTR )
	    PNPOINT = AtmCreateAndInsert( BYTES_E, 0xFFFD, VPOS_ENTRY,NULLPNTR);
	COPYOFNAMES = DmCreate( 0 );
	DmMove( COPYOFNAMES, PERMNAMES );
	PsDelHolder( AtmPutValue( PNPOINT, COPYOFNAMES ) );
				/* COPYOFNAMES is captured by VPOS */
	/* Form file table stored in VPOS */
	FITAB = DmCreate( 0 );
	for( FITEM = FD + 1; DmInside( (char *)FITEM, FILETABLE ); FITEM++ )
	  {
	    if( FITEM IS_CANCELLED )
		memset( FITEM->DESC.PATHNODE, 0, (unsigned)FMSIZE );
	    DmPush( FITAB, (char *)FITEM, sizeof( dbfiledesc ) );
	  }
	/* Write file table to VPOS */
	if( ( PNPOINT = AtmFind( VPOS_ENTRY, 0xFFFC, -1 ) ) == NULLPNTR )
	    PNPOINT = AtmCreateAndInsert( BYTES_E, 0xFFFC, VPOS_ENTRY,
								NULLPNTR );
	PsDelHolder( AtmPutValue( PNPOINT, FITAB ) );

	/* Write garbage description for each non-main file */
	for( POINT = PNPOINT = AtmSubord( OWNGAR ), FITEM = FD + 1;
	     DmInside( (char *)FITEM, FILETABLE );
	     FITEM++, POINT = PNPOINT, PNPOINT = AtmNext( PNPOINT ) )
	  {
	    if( PNPOINT == NULLPNTR )
	      {
		DM_ATOM = AtmCreate( BYTES_E, 0xFFFF );
		if( ! (FITEM IS_CANCELLED) && FITEM->GARBAGE )
		  {
		    AtmDetValue( DM_ATOM, DmFix( FITEM->GARBAGE ),
						DmSize( FITEM->GARBAGE ) );
		    DmUnfixDelete( FITEM->GARBAGE );
		    FITEM->GARBAGE = 0;
		  }
		PNPOINT = AtmInsert( DM_ATOM, POINT == NULLPNTR ? OWNGAR :
							NULLPNTR, POINT );
	      }
	    else
	      {
		PNPOINT = PsDelHolder( AtmPutValue( PNPOINT, FITEM->GARBAGE ));
		FITEM->GARBAGE = 0; /* Garbage DM object is captured by VPOS */
	      }
	  }

	/* Write the atom with apparent size to organizational data of VPOS */
	if( ( PNPOINT = AtmFind( VPOS_ENTRY, 0xFFF5, -1 ) ) == NULLPNTR )
	    PNPOINT = AtmCreateAndInsert( LONGINT_B, 0xFFF5,
							VPOS_ENTRY, NULLPNTR );

	DmDelete( PsNoHolder( ) );    /* Remove all place holders */

	/* Create and insert a garbage description for the main file;
	it is always written to the end of the main file (at APPASIZE point),
	thus the description content cannot be changed */

	POINT = FD->APPASIZE;
	DM_ATOM = AtmCreate( BYTES_E, 0xFFFF ); /* For garbage of main file */
	if( FD->GARBAGE )
	  {
	    AtmDetValue( DM_ATOM, DmFix( FD->GARBAGE ), DmSize( FD->GARBAGE ));
	    DmUnfixDelete( FD->GARBAGE );
	    FD->GARBAGE = 0;
	  }
	FD->APPASIZE += DmSize( DM_ATOM );
	AtxAllocate( UPDATED, POINT, DM_ATOM ); /* Include DM_ATOM into VPOS */
	/* Updating of pointers in OWNGAR and DM_ATOM */
      {
	long NEXTGAR = AtmSubord( OWNGAR );
	unsigned short int DMOWNGAR = AtmGet( OWNGAR );
	atom_head *PME;
	((atom_head *)DmPointer( DMOWNGAR ))->VALUE = POINT;
	AtmUpdated( OWNGAR );
	PME = (atom_head *)DmPointer( DM_ATOM );
	if( NEXTGAR == NULLPNTR )
	  {
	    PME->POINT = OWNGAR;
	    PME->STATUS |= TOOWNER;
	  }
	else
	  {
	    PME->POINT = NEXTGAR;
	    PME->STATUS &= TONEXT;
	  }
      }
	/* Assign new apparent size of the file */
/*
	FD->APPASIZE +=  AtxSize( POINT );
*/
	AtmIntAssign( PNPOINT, FD->APPASIZE );
	
	/* Possible optimization - make an order in the main data section:
	put names of entities before names of procedures and put all cancelled
	items at the end */
	PsCloseFile( 0 );
	/* Renaming NEWFILE files */
	for( FITEM = FD; DmInside( (char *)FITEM, FILETABLE ); FITEM++ )
	    if( ! (FITEM IS_CANCELLED) && FITEM->FLAGS & NEWFILE )
	      {
		char OLDNAME[ FMSIZE ], ACTNAME[ FMSIZE ], WORKNAME[ FMSIZE ];
		strcpy( OLDNAME, FITEM->DESC.PATHNODE );
		strcat( OLDNAME, OLD_EXT );
		strcpy( ACTNAME, FITEM->DESC.PATHNODE );
		strcpy( WORKNAME, FITEM->DESC.PATHNODE );
		strcat( WORKNAME, WORK_EXT );
		unlink( OLDNAME );
		rename( ACTNAME,  OLDNAME );
		rename( WORKNAME, ACTNAME );
	      }
      }
    DmUnfixDelete( AINVMAIN );
    DmUnfixDelete( AINVOVER );
    DmUnfixDelete( FILETABLE );
    AINVMAIN = AINVOVER = FILETABLE = 0;
    return( 1 );
  }


/*-------------------------------------------------------- WrtAtom ----------*/
/* Internal procedure necessary for PsCloseFile and RemDmAtom */

static void  WrtAtom( FDESC, FILENO, INV, FRQ )
    filedesc *FDESC;
    int FILENO;
    atomdesc *INV;
    int FRQ;
  {
    long POINT = INV->POINT;
    unsigned short int DMOBJ = INV->DMOBJ;
    if( AtxFile( POINT ) == FILENO )
      {
	if( INV IS_UPDATED )
	  {
	    long ASIZE = AtxSize( POINT );
	    long SEEKPOINT = AtxPoint( POINT );
	    if( SEEKPOINT <= FDESC->DESC.SIZE )
		fseek( FDESC->FPOINTER, SEEKPOINT, 0 );
	    else
		/* Write garbage and leave seekpntr at EOF*/
		PsExtFile( FDESC->FPOINTER, SEEKPOINT - FDESC->DESC.SIZE );
	    AtxWrite( FDESC->FPOINTER, DMOBJ );
	    if( SEEKPOINT + ASIZE > FDESC->DESC.SIZE )
		FDESC->DESC.SIZE = SEEKPOINT + ASIZE;
	  }
	if( DmFrequency( DMOBJ ) <= FRQ && DmExists( DMOBJ ) == -1 )
	  {
	    DmDelete( DMOBJ );
	    AtxDeAllocate( POINT );
	  }
	else
	    INV->FLAGS &= ~UPDATED;
      }
  }


/*----- External ---------------------------------------- PsAtxDel ---------*/
/* The procedure for removing atoms from DM and for updating atom inventory.
   Called after DM complains for overloading. */

/* Minimal number of atoms to be removed in 1st stage */
#define MINREMOVED 20

/* Frequency of atoms to be removed */
#define CRIT_FRQ_1 128
#define CRIT_FRQ_2 128

void PsAtxDel(  )
  {
    int AREST;
    atomdesc *AT;
    unsigned REMOVED = 0;                  /* The number of removed atoms */
    DM_COMPLAIN = 0;
    /* Remove non-updated atoms with frequency less than CRITICALFRQ */
    for( AT = BEGAOVER; DmInside( (char *)AT, AINVOVER ); AT++ )
	if( ! ( AT IS_CANCELLED ) )
	    REMOVED += RemNUDmAtom( AT, CRIT_FRQ_1 );
    for( AT = BEGAMAIN, AREST = AINVSIZE; AREST; AT++, AREST-- )
	if( ! ( AT->FLAGS & OVERFLOW || AT IS_CANCELLED ) )
	    REMOVED += RemNUDmAtom( AT, CRIT_FRQ_1 );
    if( REMOVED < MINREMOVED )
      {
	/* Remove updated atoms with frequency less than FRQ */
	for( AT = BEGAOVER; DmInside( (char *)AT, AINVOVER ); AT++ )
	    if( ! ( AT IS_CANCELLED ) )
		RemDmAtom( AT, CRIT_FRQ_2 );
	for( AT = BEGAMAIN; DmInside( (char *)AT, AINVMAIN ); AT++ )
	    if( ! ( AT->FLAGS & OVERFLOW || AT IS_CANCELLED ) )
		RemDmAtom( AT, CRIT_FRQ_2 );
      }
  }


/*-------------------------------------------------------- RemDmAtom --------*/
/* Remove an atom (atoms) from DM; auxiliary procedure for PsAtxDel */

static void  RemDmAtom( AT, FRQ )
    atomdesc *AT;
    int FRQ;
 {
    unsigned short int DMO = AT->DMOBJ;
    if( DmFrequency( DMO ) <= FRQ && DmExists( DMO ) == -1 )
      {
	if( AT IS_UPDATED )
	  {
	    int FILENO = AtxFile( AT->POINT );
	    filedesc *FDESC = (filedesc *)DmFix( FILETABLE ) + FILENO;
	    if( FDESC->FLAGS & NEWFILE )
	      {
		PsFileOpen( AT->POINT );
		WrtAtom( FDESC, FILENO, AT, FRQ );
	      }
	    else
		PsCloFile( FILENO, FRQ );
	    DmUnfix( FILETABLE );
	  }
	else
	  {
	    DmDelete( DMO );
	    AtxDeAllocate( AT->POINT );
	  }
      }
  }


/*-------------------------------------------------------- RemNUDmAtom ------*/
/* Remove non-updated atoms from DM; auxiliary procedure for PsAtxDel */

int  RemNUDmAtom( AT, FRQ )
    atomdesc *AT;
    int FRQ;
 {
    unsigned short int DMO = AT->DMOBJ;
    if( DmFrequency( DMO ) <= FRQ && DmExists( DMO ) == -1 && !(AT IS_UPDATED))
      {
	DmDelete( DMO );
	AtxDeAllocate( AT->POINT );
	return( 1 );
      }
    return( 0 );
  }


/*----- External ----------------------------------------- PsCreateFile -----*/
/* Create the given file in a filetable; set CURR_FILE on the created file.
   Return 0 iff the number of files too high, > 256, or if the file exists.
   Two files with identical path + name are not allowed. */

int PsCreateFile( FILENAME, TEMPOR )
    char *FILENAME;     /* The path + name of created file */
    int TEMPOR;         /* 0 if permanent, 1 if temporary */
  {
    filedesc *BT, *FD, *HITCANC;
    char PATHNODE[ FMSIZE ];
    CHANGES_IN_VPOS = 1;
    strcpy( PATHNODE, FILENAME );

    /* Note that it would be nice to check the path!!! */
    HITCANC = NULL;
    for( BT = FD = (filedesc *)DmFix( FILETABLE );
				DmInside( (char *)FD, FILETABLE ); FD++ )
	if( FD IS_CANCELLED )
	  {
	    if( HITCANC == NULL )
		HITCANC = FD;
	  }
	else if( ! strcmp( PATHNODE, FD->DESC.PATHNODE ) )
	  {
	    PS_ERROR |= b8;     /* b8 - file already exist */
	    DmUnfix( FILETABLE );
	    return( 0 );
	  }
    if( HITCANC == NULL )
      {
	DmUnfix( FILETABLE );
	if( FD - BT >= 256 )
	  {
	    PS_ERROR |= b9;      /* b9 - too many data files */
	    return( 0 );
	  }
	DmExtend( FILETABLE, sizeof( filedesc ) );
	BT = (filedesc *)DmFix( FILETABLE );
	HITCANC = (filedesc *)( DmTopPointer( FILETABLE ) ) - 1;
      }
    memset( HITCANC->DESC.PATHNODE, 0, (unsigned)FMSIZE );
    strcpy( HITCANC->DESC.PATHNODE, PATHNODE );
/* No extension considered
    strcpy( HITCANC->DESC.EXT, EXT );
*/
    HITCANC->DESC.SIZE = HITCANC->APPASIZE = 0L;
    HITCANC->FLAGS = APPARENT;
    HITCANC->GARBAGE = 0;
    if( TEMPOR )
	HITCANC->FLAGS |= TEMPORARY;
    HITCANC->FPOINTER = NULL;
    DmUnfix( FILETABLE );
    CURR_FILE = HITCANC - BT;
    if( CURR_FILE == 0 )
      { /* If created file is the main file, write (apparently) the root of
	   organizational data at its beginning. */
	HITCANC->APPASIZE = (long)ATOMHEAD;
	AtxAllocate( UPDATED, VPOS_ENTRY, AtmCreate( RING_B, 0xFFFE ) );
      }
    return( 1 );
  }


/*----- External ----------------------------------------- PsTemporary ------*/
/* Return 1 if the file is temporary and 0 otherwise */

int PsTemporary( FILENO )
    int FILENO;
  {
    return( ((filedesc *)DmPointer( FILETABLE ) + FILENO)->FLAGS & TEMPORARY );
  }


/*----- External ----------------------------------------- PsDeleteFile ----*/
/* Delete the file together with all data on it */

int PsDeleteFile( FILENO )
    int FILENO;
  {
    char FNAME[ FMSIZE ];
    filedesc *FITEM;
    CHANGES_IN_VPOS = 1;
    FITEM = (filedesc *)DmFix( FILETABLE ) + FILENO;
    if( ! (FITEM IS_CANCELLED) )
      {
	long POINT = 0L;
	long SIZE;
	if( FITEM->GARBAGE == 0 )
	    FITEM->GARBAGE = DmCreate( 0 );
	while( FITEM->APPASIZE > 0L )
	  {
	    if( DmSize( FITEM->GARBAGE ) > 0L )
	      {
		gardesc *GD = (gardesc *)DmFix( FITEM->GARBAGE );
		PsGarLong( GD, POINT, SIZE );
		POINT = POINT > 0L ? 0L : SIZE;
		DmUnfix( FITEM->GARBAGE );
	      }
	    else
		POINT = 0L;
/* Note difference of SUN and PC: below the next command for PC */
	    *((char *)&POINT + 3 ) = FILENO;
/*	    *(unsigned char *)&POINT = FILENO;  */
	    AtxDelete( NULLPNTR, POINT );
	  }
	DmDelete( FITEM->GARBAGE );
	FITEM->GARBAGE = 0;
	if( FITEM->FLAGS & NEWFILE )
	  {
	    if( FITEM->FPOINTER != NULL )
	      {
		if( fclose( FITEM->FPOINTER ) == -1 )
		  {
		    PS_ERROR |= b12;
		    PsError();
		  }
		FITEM->FPOINTER = NULL;
	      }
	    ACTUAL_FILE = -1;
	    CURR_FPOINTER = NULL;
	    strcpy( FNAME, FITEM->DESC.PATHNODE );
	    strcat( FNAME, WORK_EXT );
/*
	    strmfe( FNAME, FITEM->DESC.PATHNODE, WORK_EXT );
*/
	    unlink( FNAME );
	  }
	FITEM->FLAGS = CANCELLED;
      }
    DmUnfix( FILETABLE );
    return( 1 );
  }


/*----- External ----------------------------------------- PsDetTypes -------*/
/* Determine intervals of basic atom types of VPOS */
/* The following conditions must be satisfied:
 0 <= RING_TYPE < LOGIC_TYPE < LONG_TYPE - 2;
	LONG_TYPE < BYTES_TYPE < 255 */

void PsDetTypes( RING_TYPE, LOGIC_TYPE, LONG_TYPE, BYTES_TYPE )
    int RING_TYPE, LOGIC_TYPE, LONG_TYPE, BYTES_TYPE;
  {                                     /* Below there are defaults: */
    RING_B       = 0;
    RING_E       = RING_TYPE;           /* 0..5    - ring types */
    LOGICAL_B    = RING_TYPE + 1;
    LOGICAL_E    = LOGIC_TYPE;          /* 6..8    - logical pointer atoms */
    HOLDER       = LOGIC_TYPE + 1;      /* 9       - place holder atom */
    TOEND        = LOGIC_TYPE + 2;      /* 10      - to-end-of-ring reference*/
    LONGINT_B    = LOGIC_TYPE + 3;
    LONGINT_E    = LONG_TYPE;           /* 11..20  - long integer atoms */
    BYTES_B      = LONG_TYPE + 1;
    BYTES_E      = BYTES_TYPE;          /* 21..240 - variable size atoms */
    SPIDER_B     = BYTES_TYPE + 1;
    SPIDER_E     = 254;                 /* 241..254 - spider types */
    BACKWARD     = 255;                 /* 255     - backward pointers atom */
  }


/*----- External ----------------------------------------- PsInit -----------*/
/* Initialization of structures necessary for the package */

int PsInit(  )
  {
    atomdesc *INV;
    unsigned INVC;
    PS_ERROR = 0;
    CURR_FILE = 0;      /* Current file for storing new atoms */ 
    OPENFILES = 0;      /* The number of opened files */
    CHANGES_IN_VPOS = 0;        /* The indicator if the filetable has been changed */
    if( FILETABLE )
	DmClear( FILETABLE );
    else
	FILETABLE = DmCreate( 0 );
    if( PERMNAMES )
	DmClear( PERMNAMES );
    else
	PERMNAMES = DmCreate( 0 );
    if( ! AINVMAIN )
      {
	AINVMAIN = DmCreate( AINVSIZE * sizeof( atomdesc ) );
	BEGAMAIN = (atomdesc *)DmFix( AINVMAIN );
      }
    for( INV = BEGAMAIN, INVC = AINVSIZE; INVC--; INV++ )
	INV->FLAGS = CANCELLED;
    if( ! AINVOVER )
      {
	AINVOVER = DmCreate( 0 );
	BEGAOVER = (atomdesc *)DmFix( AINVOVER );
      }
    return( 1 );
  }


/*-------------------------------------------------------- AtxLeaveHolder ---*/
/* Leave HOLDER at POINT, pointing NEWPOINT */

static void  AtxLeaveHolder( POINT, NEWPOINT )
    long POINT, NEWPOINT;
  {
    long OWNH;
    atomdesc *AD = AtxLocate( POINT );
    /* Update atom inventory */
    AD->FLAGS = UPDATED;
    AD->DMOBJ = AtmCreate( HOLDER, 0xFFFF );
    /* Include the holder into the holder ring */
    if( ( OWNH = AtmFind( VPOS_ENTRY, 0xFFFA, -1 )) == NULLPNTR )
	/* Create an owner of holders */
	OWNH = AtmCreateAndInsert( RING_B, 0xFFFA, VPOS_ENTRY, NULLPNTR );
    AtxInclude( POINT, OWNH, NULLPNTR );
    ((atom_head *)DmPointer( AD->DMOBJ ))->VALUE = NEWPOINT;
    AtxInsBack( POINT, HOLDER, NEWPOINT );
  }


/*----- External ----------------------------------------- PsDelHolder ------*/
/* Delete all place holders (if exist) that are associated with the point. No
logical pointer can lead to a holder. Return true atom's point. */

long PsDelHolder( POINT )
    long POINT;
  {
    long TRUEPOINT;
    SkipHolders( POINT );
    TRUEPOINT = POINT;
    /* Find BACKWARD atom */
    POINT = AtxNext( POINT );
    if( POINT != NULLPNTR && AtxType( POINT ) == BACKWARD )
      {
	long *H;
	unsigned short int BA = AtxGetValue( POINT );
	for( H = (long *)DmFix( BA ); DmInside( (char *)H, BA ); H++ )
	    if( *H != NULLPNTR && AtxType( *H ) == HOLDER )

		AtxDelete( NULLPNTR, *H );
	DmUnfixDelete( BA );
      }
    return( TRUEPOINT );
  }


/*----- External ----------------------------------------- PsNoHolder -------*/
/* Remove all place holders and return the DM table containing pairs
	<holder point, original atom point> */

unsigned short int PsNoHolder(  )
  {
    unsigned short int TABLE = DmCreate( 0 );
    long HOWN;
    HOWN = AtmFind( VPOS_ENTRY, 0xFFFA, -1 );
    if( HOWN != NULLPNTR )
      {
	long H, ATOM;
	for( H = AtxSubord( HOWN ); H != NULLPNTR; H = AtxNext( H ) )
	  {
	    ATOM = H;
	    DmPush( TABLE, (char *)&H, sizeof( long ) );
	    SkipHolders( ATOM );
	    DmPush( TABLE, (char *)&ATOM, sizeof( long ) );
	  }
	AtxDelete( NULLPNTR, HOWN );
      }
    return( TABLE );
  }


/*----- External ----------------------------------------- PsOpen -----------*/
/* Open physical data structure */

int PsOpen( MAINFILE )
    char * MAINFILE;    /* The main file of VPOS */
    /* The beginning atom of the main file is an entry point to the field 
	of vision of the whole VPOS. Naming convention is the following:
	0xFFFF - means "no name"; the name of garbage descriptors and holders,
	0xFFFE - name of a root atom for organizational data; always stored
		 at the beginning of the main file.
	0xFFFD - name of permanent dictionary atom,
	0xFFFC - name of file table atom, 
	0xFFFB - name of a root atom (subordinated to 0xFFFE) for atoms with
		 garbage description; each subordinated atom has name 0xFFFF,
	0xFFFA - name of a root atom for all place holders,
	0xFFF9 - name of an atom with boundaries of types,
	0xFFF8 - name of backward atoms.
	Names from 0xFFF0 to 0xFFF7 are reserved to similar special purposes */
  {
    /* Checking if the file exists and if it represents data file */
    FILE *MAINFP;
    atom_head AH;
    if( ( MAINFP = fopen( MAINFILE, "rb" ) ) == NULL )
      {
       if( PsInit( ) )
	  { /* If the database does not exist, it is created */
	    PsCreateFile( MAINFILE, 0 );
	    return(1);
	  }
	else
	  {
	    PS_ERROR |= b1;
	    PsError();
	    return( 0 );
	  }
      }
    else
	if( fread( (char *)&AH, ATOMHEAD, 1, MAINFP ) != 1 ||
				AH.TYPE != RING_B || AH.NAME != 0xFFFE )
	  {
	    PS_ERROR |= b1;
	    fclose( MAINFP );
	    printf( 
		"\nSorry, file %s exists, but it is not a valid database!\n",
		MAINFILE );
	    PsError();
	    return( 0 );
	  }

    if( PsInit( ) )
      {
	filedesc *FD;
	long ATOMPOINT;
	long ATOMFSIZE;
	long ATOMGDESC;
	unsigned short int TEMPTAB;
	dbfiledesc *SV;
	/* Open the main file */
	DmExtend( FILETABLE, sizeof( filedesc ) );
	FD = (filedesc *)DmFix( FILETABLE );
	/* Initial filling FILETABLE. Main file has always the number 0 */
	FD->FPOINTER = CURR_FPOINTER = MAINFP;
	FD->FLAGS = ACTUAL_FILE = 0;
	FD->GARBAGE = 0;
	FD->APPASIZE = INT_MAX; /* To prevent reaction from assertion */
	strcpy( FD->DESC.PATHNODE, MAINFILE );
		/* Get the file size */
	if( ( ATOMFSIZE = AtmFind( VPOS_ENTRY, 0xFFF5, -1 ) ) == NULLPNTR )
	  {
	    fseek( FD->FPOINTER, 0L, 2 );
	    FD->DESC.SIZE = ftell( FD->FPOINTER );
	  }
	else
	    FD->DESC.SIZE = AtmIntValue( ATOMFSIZE );
	FD->APPASIZE = FD->DESC.SIZE;
	DmUnfix( FILETABLE );

		/* Find and set boundaries of types of atoms */
	ATOMPOINT = AtmFind( VPOS_ENTRY, 0xFFF9, -1 );
	if( ATOMPOINT != NULLPNTR )
	  {
	    char *BOUND;
	    TEMPTAB = AtmGetValue( ATOMPOINT );
	    BOUND = DmPointer( TEMPTAB );
	    PsDetTypes( *BOUND, *(BOUND + 1), *(BOUND + 2), *(BOUND + 3) );
	    DmDelete( TEMPTAB );
	  }

		/* Rewrite file table from storage to FILETABLE */
	TEMPTAB = AtmGetValue( AtmFind( VPOS_ENTRY, 0xFFFC, -1 ) );
	    /* File table written on the main file does not contain the
		description of the main file. Each item of the file table
		is a stored file description */
	for( SV = (dbfiledesc *)DmFix( TEMPTAB );
				DmInside( (char *)SV, TEMPTAB ); SV++ )
	  {
	    filedesc OP;
	    OP.DESC = *SV;
	    OP.FLAGS = *OP.DESC.PATHNODE ? 0 : CANCELLED;
	    OP.APPASIZE = OP.DESC.SIZE;
	    OP.GARBAGE = 0;
	    OP.FPOINTER = NULL;
	    DmPush( FILETABLE, (char *)&OP, sizeof(filedesc) );
	  }
	DmUnfixDelete( TEMPTAB );
		/* Read garbage description for each file */
	ATOMGDESC = AtmFind( VPOS_ENTRY, 0xFFFB, -1 );
	ATOMPOINT = AtmSubord( ATOMGDESC );
	for( FD = (filedesc *)DmFix( FILETABLE );
		ATOMPOINT != NULLPNTR; ATOMPOINT = AtmNext( ATOMPOINT ), FD++ )
	    if( AtmValSize( ATOMPOINT ) )
		FD->GARBAGE = AtmGetValue( ATOMPOINT );
	DmUnfix( FILETABLE );
		/* Rewrite table of permanent names to PERMNAMES */
	DmDelete( PERMNAMES );
	PERMNAMES = AtmGetValue( AtmFind( VPOS_ENTRY, 0xFFFD, -1) );

/* For preventing file growth, delete atoms with file size and garbage desc. */
	if( ATOMFSIZE != NULLPNTR )
	    AtxDelete( NULLPNTR, ATOMFSIZE );
	AtxDelete( NULLPNTR, ATOMGDESC );
	CHANGES_IN_VPOS = 0;
	return( 1 );
      }
    else
      {
	printf( "\nCannot initialize!\n" );
	PsError();
      }
    return( 0 );
  }


/*-------------------------------------------------------- AtxInclude -------*/
/* Connect point into new place; similar to AtxInsert. The POINT must be
   previously disconnected from VPOS rings */

static void  AtxInclude( POINT, OWNER, AFTER )
    long POINT;       /* Db pointer to the atom */
    long OWNER;       /* Further owner of the atom, or NULLPNTR */
    long AFTER;       /* Further left neighbour of the atom, or NULLPNTR */
  {
    /* If OWNER = AFTER = NULLPNTR, insert into the main ring */
    unsigned short int DMOBJECT;
    atom_head *AS;
    long PO;
    atom_head *OB, *AH;
    CHANGES_IN_VPOS = 1;
    if( OWNER == NULLPNTR && AFTER == NULLPNTR )
	AFTER = VPOS_ENTRY;
    if( OWNER != NULLPNTR )
      {
	PO = AtxSubord( OWNER );
	if( PO != NULLPNTR && AtxType( PO ) == TOEND )
	  {
	    OWNER = NULLPNTR;
	    AFTER = PO;
	  }
      }
    if( OWNER == NULLPNTR && AFTER != NULLPNTR )
      {
	/* Shift AFTER to the next backward pointer */
	long NX = AtxNext( AFTER );
	if( NX != NULLPNTR && AtxType( NX ) == BACKWARD )
	    AFTER = NX;
      }
    DMOBJECT = UpdDmObject( POINT );
    OB = (atom_head *)DmFix( DMOBJECT );
    if( OWNER != NULLPNTR )
      {
	AS = (atom_head *)DmPointer( UpdDmObject( OWNER ) );
	PO = AS->VALUE;
	AS->VALUE = POINT;
	if( PO == NULLPNTR )
	  {
	    OB->POINT = OWNER;
	    OB->STATUS |= TOOWNER;
	  }
	else
	  {
	    OB->POINT = PO;
	    OB->STATUS &= TONEXT;
	  }
      }
    else        /* if( AFTER != NULLPNTR ) */
      {
	AH = (atom_head *)DmPointer( UpdDmObject( AFTER ) );
	PO = AH->POINT;
	OB->POINT = PO;
	AH->POINT = POINT;
	OB->STATUS = (OB->STATUS & TONEXT) | (AH->STATUS & TOOWNER);
	AH->STATUS &= TONEXT;
	if( OB->STATUS & TOOWNER && PO != NULLPNTR )
	  {
	    long TOENDREF = AtxSubord( PO );
	    if( AtxType( TOENDREF ) == TOEND )
		AtmIntAssign( TOENDREF, POINT );
	  }
      }
    DmUnfix( DMOBJECT );
  }


/*-------------------------------------------------------- AtxDisconnect ----*/
/* Disconnect the atom from the ring */

void  AtxDisconnect( POINT )
    long POINT;
  {
    long NEXT, OWNER, PREVIOUS;
    CHANGES_IN_VPOS = 1;
    PREVIOUS = AtxPriNexOwn( POINT, &NEXT, &OWNER );
    if( PREVIOUS != NULLPNTR )
      {
	atom_head *AH = (atom_head *)DmPointer( UpdDmObject( PREVIOUS ) );
	if( NEXT != NULLPNTR )
	    AH->POINT = NEXT;
	else
	  {
	    AH->POINT = OWNER;
	    AH->STATUS |= TOOWNER;
	    if( OWNER != NULLPNTR )
	      {
		long TOENDREF = AtxSubord( OWNER );
		if( TOENDREF != NULLPNTR && AtxType( TOENDREF ) == TOEND )
		    AtmIntAssign( TOENDREF, PREVIOUS );
	      }
	  }
      }
    else        /* Update owner */
	((atom_head *)DmPointer( UpdDmObject(OWNER) ))->VALUE = NEXT;
  }


/*-------------------------------------------------------- AtxDiscNext ------*/
/* Disconnect the next atom from the ring; return 0 iff PREVIOUS == NULLPNTR or
   next atom does not exist */

static int  AtxDiscNext( PREVIOUS )
    long PREVIOUS;
  {
    long POINT, NEXT;
    unsigned short int DMOBJECT;
    atom_head *AH;
    if( PREVIOUS == NULLPNTR || ( POINT = AtxNext( PREVIOUS )) == NULLPNTR )
	return( 0 );
    CHANGES_IN_VPOS = 1;
    NEXT = AtxNext( POINT );
    DMOBJECT = UpdDmObject( PREVIOUS );
    AH = (atom_head *)DmFix( DMOBJECT );
    if( NEXT != NULLPNTR )
	AH->POINT = NEXT;
    else
      {
	long OWNER = AtxOwner( POINT );
	AH->POINT = OWNER;
	AH->STATUS |= TOOWNER;
	if( OWNER != NULLPNTR )
	  {
	    long TOENDREF = AtxSubord( OWNER );
	    if( TOENDREF != NULLPNTR && AtxType( TOENDREF ) == TOEND )
		AtmIntAssign( TOENDREF, PREVIOUS );
	  }
      }
    DmUnfix( DMOBJECT );
    return( 1 );
  }


/*-------------------------------------------------------- UpdDmObject ------*/
/* Return DM object associated with a point and update atom inventory */

static unsigned short int  UpdDmObject( POINT )
    long POINT;
  {
    atomdesc *AD = AtxRead( POINT, 1 );
    CHANGES_IN_VPOS = 1;
    AD->FLAGS |= UPDATED;
    return( AD->DMOBJ );
  }



/*---------------------------------------------------------------------------*/
/* The following organization of the name dictionary is assumed:
1. The dictionary is a dynamic memory object PERMNAMES;
2. The code of name is its unsigned relative position in the dictionary.
3. Names are stored consecutively, with end-of-string character, with no order.
4. If the name is cancelled, the filler 0xFF is assumed. The sequence of
   fillers is ended by the end-of-string character. */


/*----------------------------------------------------- PsInsNameAbs -------*/
/* Insert the name into name table and return its code; 0xFFFF if overflow */

unsigned short int PsInsNameAbs( NAME )
    char *NAME;
  {
    unsigned short int CODE;    /* The returned  code of the name */
    char *BEGTAB, *VARTAB;
    int NAMESIZE;
    CHANGES_IN_VPOS = 1;
    /* Include the name appropriately */
    NAMESIZE = strlen( NAME );
    /* Looking for proper place */
    for( BEGTAB = VARTAB = DmFix( PERMNAMES );
	 DmInside( VARTAB, PERMNAMES ); VARTAB += strlen( VARTAB ) + 1 )
	if( *(unsigned char *)VARTAB == 0xFF && strlen( VARTAB ) >= NAMESIZE )
	   {
	    strcpy( VARTAB, NAME );
	    DmUnfix( PERMNAMES );
	    CODE = (unsigned short int)(VARTAB - BEGTAB);
	    return( CODE );
	   }
    DmUnfix( PERMNAMES );
    if( (unsigned short int)( DmSize( PERMNAMES ) >= 0xFFEF ) )
      {
	PS_ERROR |= b7;
	return( 0xFFFF );
      }
    CODE = DmSize( PERMNAMES ); 
    DmPush( PERMNAMES, NAME, NAMESIZE + 1 );
    return( CODE );
  }


/*----- External ----------------------------------------- PsInsName --------*/
/* Insert (check the presence of) the name into name table
   and return its code. Return 0xFFFF if overflow */

unsigned short int PsInsName( NAME )
    char *NAME;
  {
    unsigned short int CODE;    /* The returned  code of the name */
    if( ( CODE = PsNameCode( NAME ) ) != 0xFFFF )
	return( CODE );
    return( PsInsNameAbs( NAME ) );
  }


/*----- External ----------------------------------------- PsNameCode -------*/
/* Change atom name into its unsigned code. Return 0xFFFF if the name is
   not found */

unsigned short int PsNameCode( NAME )
    char *NAME;
  {
    char *BEGTAB, *VARTAB;
    for( BEGTAB = VARTAB = DmFix( PERMNAMES );
		DmInside( VARTAB, PERMNAMES ); VARTAB += strlen( VARTAB ) + 1 )
	if( !strcmp( NAME, VARTAB ) )
	  {
	    DmUnfix( PERMNAMES );
	    return( (unsigned short int)(VARTAB - BEGTAB) );
	  } 
    DmUnfix( PERMNAMES );
    return( PsInsNameAbs( NAME ) );     /* No name -> inserted */
  }


/*----- External ----------------------------------------- PsCodeName -------*/
/* Change unsigned name code into atom name. Return 0 iff the name is
   not found */

int PsCodeName( CODE, NAME )
    unsigned short int CODE;
    char *NAME;
  {
    char *BGNAME;
    if( CODE >= (unsigned short int)0xFFF0 )
      {
	sprintf( NAME, FindString( sSTANDARD_NAME_0x ), CODE );
	return( 1 );
      }
    if( CODE >= DmSize( PERMNAMES ) - 1 )
      {
	sprintf( NAME, FindString( sUNKNOWN_NAME_0x ), CODE );
	return( 0 );
      }
    BGNAME = DmFix( PERMNAMES ) + CODE;
    if( ! DmInside( BGNAME, PERMNAMES) ||
	 *(unsigned char *)BGNAME == 0xFF || *BGNAME == '\0' ||
	 ( CODE != 0 && *( BGNAME - 1 ) != '\0' ) )
      {
	DmUnfix( PERMNAMES );
	*NAME = '\0';
	return( 0 );
      }
    else
      {
	strcpy( NAME, BGNAME );
	DmUnfix( PERMNAMES );
	return( 1 );
      }
  }


/*----- External ----------------------------------------- PsNameSize -------*/
/* Return the size of name (excluding null terminator), or -1 iff error */

int PsNameSize( CODE )
    unsigned short int CODE;
  {
    return( (int)( CODE < DmSize( PERMNAMES ) ?
	strlen( DmPointer( PERMNAMES )+CODE ) : -1 ) );
  }


/*----- External ----------------------------------------- AtmCreate -------*/
/* The atom is created in the specified DM object. Defaults: for integer - 0L
value, for variable size - empty value; all pointers are NULLPNTR. */

unsigned short int AtmCreate( TYPE, NAME )
    int TYPE;
    unsigned short int NAME;
  {
    unsigned short int DMOBJECT;
    atom_head *AH;
    /* Four reserve items for spiders */
    DMOBJECT = DmCreate( SPIDER_B <= TYPE && TYPE <= SPIDER_E ?
			ATOMHEAD + 4 * sizeof( spider_item ) : ATOMHEAD );
    AH = (atom_head *)DmPointer( DMOBJECT );
    AH->TYPE = TYPE;
    AH->NAME = NAME;
    AH->STATUS = 
    AH->RESERVED[0] =
    AH->RESERVED[1] =
    AH->RESERVED[2] =
    AH->RESERVED[3] = 0;
    AH->POINT = NULLPNTR;
    if( TYPE <= TOEND )
	AH->VALUE = NULLPNTR;
    else if( SPIDER_B <= TYPE && TYPE <= SPIDER_E )
      {
	spider_item *PHI = (spider_item *)( (char *)AH + ATOMHEAD );
	AH->VALUE = 4 * sizeof( spider_item );
	PHI->MEMBER = (PHI+1)->MEMBER = (PHI+2)->MEMBER = (PHI+3)->MEMBER =
								NULLPNTR;
      }
    else
	AH->VALUE = 0L;
    return( DMOBJECT );
  }


/*----- External -------------------------------------- AtmCreateAndInsert -*/
/* The atom is created as for AtmCreate and then inserted as for AtmInsert */

long AtmCreateAndInsert( TYPE, NAME, OWNER, AFTER )
    int TYPE;
    unsigned short int NAME;
    long OWNER;       /* Further owner of the atom, or NULLPNTR */
    long AFTER;       /* Further left neighbour of the atom, or NULLPNTR */
  {
    unsigned short int DMOBJECT;
    atom_head *AH;
    /* Four reserve items for spiders */
    DMOBJECT = DmCreate( SPIDER_B <= TYPE && TYPE <= SPIDER_E ?
			ATOMHEAD + 4 * sizeof( spider_item ) : ATOMHEAD );
    AH = (atom_head *)DmPointer( DMOBJECT );
    AH->TYPE = TYPE;
    AH->NAME = NAME;
    AH->STATUS = 
    AH->RESERVED[0] =
    AH->RESERVED[1] =
    AH->RESERVED[2] =
    AH->RESERVED[3] = 0;
    AH->POINT = NULLPNTR;
    if( TYPE <= TOEND )
	AH->VALUE = NULLPNTR;
    else if( SPIDER_B <= TYPE && TYPE <= SPIDER_E )
      {
	spider_item *PHI = (spider_item *)( (char *)AH + ATOMHEAD );
	AH->VALUE = 4 * sizeof( spider_item );
	PHI->MEMBER = (PHI+1)->MEMBER = (PHI+2)->MEMBER = (PHI+3)->MEMBER =
								NULLPNTR;
      }
    else
	AH->VALUE = 0L;
    if( DM_COMPLAIN )
	PsAtxDel( );
    SkipParam( &OWNER, &AFTER );
    return( AtxInsert( DMOBJECT, OWNER, AFTER ) );
  }


/*-------------------------------------------------------- AtxRead ----------*/
/* Read the atom to DM, return reference to inventory; if error return NULL */

atomdesc *  AtxRead( POINT, MODE )
    long POINT;
    int MODE;   /* MODE = 0 -> read whole atom;
		   MODE = 1 -> ATOMHEAD only */
  {
    atomdesc *AD = NULL;
    atomdesc *IT = BEGAMAIN + HashPoint( POINT );
    if( ! ( IT IS_CANCELLED ) ) /* Locate the point at AD */
      {
	if( IT->FLAGS & OVERFLOW )
	  {
	    unsigned short int *P = (unsigned short int *)&IT->POINT;
	    atomdesc *OVEREND = BEGAOVER + *( P + 1 );
	    for( IT = BEGAOVER + *P; IT <= OVEREND; IT++ )
		if( ! (IT IS_CANCELLED) && IT->POINT == POINT )
		  {
		    AD = IT;
		    break;
		  }
	  }
	else if( IT->POINT == POINT )
	    AD = IT;
      }
    if( AD != NULL )    /* If the atom is present */
      {
	if( ! MODE )    /* Check if the whole atom is present */
	    if( AD->FLAGS & HEADONLY )          /* Read the whole value */
		if( AtxRdPhys( POINT, AD->DMOBJ, &MODE ) )
		    AD->FLAGS &= ~HEADONLY;     /* Set HEADONLY flag to 0 */
		else
		    return( NULL );
      }
    else        /* The atom is not present */
      {
	unsigned short int DMO = DmCreate( 0 );
	if( ! AtxRdPhys( POINT, DMO, &MODE ) ) /* MODE may be changed */
	    return( NULL );
	/* New position in atom inventory */
	AD = AtxAllocate( MODE ? HEADONLY : 0, POINT, DMO );
      }
    return( AD );
  }


/*-------------------------------------------------------- AtxRdPhys --------*/
/* Read the atom to DM to given DMOBJ; If error return 0; otherwise 1 */

static int  AtxRdPhys( POINT, DMOBJ, MODE )
    long POINT;
    unsigned short int DMOBJ;
    int *MODE; /* MODE = 1 -> only ATOMHEAD should be read */
  {
    FILE *FP;
    atom_head *AH;
    char *BEGOBJ;
    long SIZE, CSIZE;
    int ERR = 0;
    int TYPE;
    FP = AtxFile( POINT ) == ACTUAL_FILE ? CURR_FPOINTER : PsFileOpen( POINT );
    if( DmSize( DMOBJ ) == 0 )
      {
	ERR = fseek( FP, AtxPoint( POINT ), 0 ) == -1;
	DmExtend( DMOBJ, ATOMHEAD );
	BEGOBJ = DmPointer( DMOBJ );
	ERR |= fread( BEGOBJ, ATOMHEAD, 1, FP ) != 1;
/*
	ERR |= read( fileno( FP ), BEGOBJ, ATOMHEAD ) != ATOMHEAD;
*/
      }
    AH = (atom_head *)DmPointer( DMOBJ );
    SIZE = AH->VALUE;
    TYPE = AH->TYPE;
    if( TYPE <= LONGINT_E || SIZE == 0 )
	*MODE = 0;
    else if( ! *MODE )
      {
	/* UINT_MAX from the PC version is replaced by USHRT_MAX on SUN */
	for( CSIZE = SIZE; CSIZE >= USHRT_MAX; CSIZE -= USHRT_MAX )
	    DmExtend( DMOBJ, USHRT_MAX );
	if( CSIZE > 0L )
	    DmExtend( DMOBJ, (int)CSIZE );
	BEGOBJ = DmPointer( DMOBJ ) + ATOMHEAD;
	ERR |= fseek( FP, AtxPoint( POINT ) + ATOMHEAD, 0 ) == -1;
	if( SIZE <= INT_MAX )
	    ERR |= fread( BEGOBJ, (int)SIZE, 1, FP ) != 1;
	else
	    ERR |= read( fileno( FP ), BEGOBJ, SIZE ) != SIZE;
      }
    if( ERR )
	PS_ERROR |= b2; /* File reading error */
    return( ! ERR  );
  }


/*----- External ----------------------------------------- AtmFile ----------*/
/* Return file number associated with the point or -1 iff error; external */

int AtmFile( POINT )
    long POINT;
  {
    if( POINT == NULLPNTR )
	return( -1 );
    SkipHolders( POINT );
    return( ( int )AtxFile( POINT ) );
  }


/*-------------------------------------------------------- PsFileOpen -------*/
/* Open the data file containing the POINT. Returns file pointer
   to a proper data file, or NULL iff error */

static FILE *  PsFileOpen( POINT )
    long POINT;
  {
    filedesc *BGFITAB, *FITEM /* , *DV */;
    if( AtxFile( POINT ) == ACTUAL_FILE )
	return( CURR_FPOINTER );
    BGFITAB = (filedesc *)DmFix( FILETABLE );
    FITEM =  BGFITAB + AtxFile( POINT ); /* Shift to relative position */
    /* Assertion K.S. */
    if( ! DmInside( (char *)FITEM, FILETABLE ) || ( FITEM IS_CANCELLED ) )
      {
	PS_ERROR |= b11;
	printf( FindString( sWRONG_FILE_FOR_POINT ),
				AtxFile( POINT ), AtxPoint( POINT ) );
	getchar();
	exit(1);
      }
    if( FITEM->FPOINTER == NULL )
      { /* Open the data file */
	/* Determine mode of the opening */
	char * OPMODE, FNAME[ FMSIZE ];
	OPMODE = FITEM->FLAGS & NEWFILE ? "rb+" : "rb";

	if( FITEM->FLAGS & NEWFILE )
	  {
	    strcpy( FNAME, FITEM->DESC.PATHNODE );
	    strcat( FNAME, WORK_EXT );
	  }
	else
	    strcpy( FNAME, FITEM->DESC.PATHNODE );
	if( ( FITEM->FPOINTER = fopen( FNAME, OPMODE ) ) == NULL )
	    PS_ERROR |= b1;
	else
	  {
	    ACTUAL_FILE = AtxFile( POINT );
	    CURR_FPOINTER = FITEM->FPOINTER; 
	    OPENFILES++;
	  }
      }
    DmUnfix( FILETABLE );
    return( FITEM->FPOINTER );
  }


/*-------------------------------------------------------- PsExtFile --------*/
/* Extend the file with SIZE bytes, and leave a seek point on the file end */

#define MAXBUF  10000
#define MINBUF  0x00400L

static void  PsExtFile( FPOINTER, SIZE )
    FILE *FPOINTER;
    long SIZE;
  {
    char *BUFFER[MAXBUF]; 
    fseek( FPOINTER, 0L, 2 );
    while(SIZE)
      {
	unsigned BLOCK = SIZE < MAXBUF ? SIZE : MAXBUF;
	if( write( fileno( FPOINTER ), BUFFER, BLOCK ) != BLOCK )
	  {
	    PS_ERROR |= b3;
	    PsError();
	  }
	SIZE -= BLOCK;
      }
  }


/*-------------------------------------------------------- AtxWrite ---------*/
/* Write the atom or its part stored in DMOBJect to the given file */

void  AtxWrite( FPOINTER, DMOBJ )
    FILE *FPOINTER;
    unsigned short int DMOBJ;
  {
    long        SIZE    = DmSize( DMOBJ );
    char        *BEGBUF = DmPointer( DMOBJ );

    if( SIZE <= INT_MAX )
      {
	if( fwrite( BEGBUF, (int)SIZE, 1, FPOINTER ) != 1 )
	  {
	    PS_ERROR |= b3;
	    PsError();
	  }
      }
    else
      {
	fflush( FPOINTER );
	if( write( fileno( FPOINTER ), BEGBUF, SIZE ) != SIZE )
	  {
	    PS_ERROR |= b3;
	    PsError();
	  }
      }
  }


/*----- External ---------------------------------------- PsRewrite --------*/
/* Write SIZE bytes from file FROM to file TO */
/* Return 0 iff error */

int PsRewrite( FROM, TO, SIZE  )
    char *FROM, *TO;
    long SIZE;
  {
    long BLOCK;
    unsigned short int BUF;
    char *BEGBUF;
    FILE *FROMFP;
    FILE *TOFP;
    if( (FROMFP = fopen( FROM, "rb" )) == NULL )
      {
	PS_ERROR |= b1;
	return( 0 );
      }
    /* Description of "fopen" for SUN and flags in SunOS Reference Manual,
       pages 979 */
    if( (TOFP   = fopen( TO,   "wb" )) == NULL )
      {
	fclose( FROMFP );
	PS_ERROR |= b1;
	return( 0 );
      }
    BUF = DmCreate( 0 );
    BLOCK = SIZE > MINBUF ? MINBUF : SIZE ;
    DmChangeSize( BUF, BLOCK );
    BEGBUF = DmPointer( BUF );
    fseek( TOFP, 0L, 2 );
    while( SIZE > 0L )
      {
	int RBL, WBL;
	if( BLOCK > SIZE )
	    BLOCK = SIZE;
	RBL = fread(  BEGBUF, BLOCK, 1, FROMFP );
	WBL = fwrite( BEGBUF, BLOCK, 1, TOFP   );
	if( RBL != 1 || WBL != 1 )
	  {
	    if( RBL != 1 )
		PS_ERROR |= b2;
	    if( WBL != 1 )
		PS_ERROR |= b3;
	    PsError();
	  }
	SIZE -= BLOCK;
      }
    DmDelete( BUF );
    if( fclose( FROMFP ) || fclose( TOFP ) )
      {
	PS_ERROR |= b12;
	PsError();
      }
    return( 1 );
  }


/*----- External ----------------------------------------- PsCloseFile ------*/
/* Close the file if any atom has been updated; write all updated data to it.
*  The following cases are possible:
*       1. The file has a flag APPARENT. In this case file is opened, all data
*          from DM are written to it and the file obtains flag NEWFILE; flag
*          APPARENT is removed. Then the file is closed.
*       2. The file has a flag NEWFILE. In this case all DM data with flag
*          UPDATED are written to it, and the file is closed.
*       3. The file has no flag APPARENT nor NEWFILE. In this case new file is
*          opened with extension WORK_EXT, data from old file are rewritten to
*          the new file, up to the actual required size and both files are
*          closed. Next, as for point 2.
*  Flag TEMPORARY is not considered; all data from temporary file are treated
*  as data from permanent files. The procedure does not cancel the file
*  description, and removes all unfixed data related to the file from DM.
*/ 

int PsCloseFile( FILENO )
    int FILENO; /* Relative position in the file description table */
  {
    return( PsCloFile( FILENO, 255 ) );
  }


/*-------------------------------------------------------- PsCloFile --------*/
/* Internal file closing */

static int  PsCloFile( FILENO, FRQ )
    int FILENO; /* Relative position in the file description table */
    int FRQ;    /* Frequency limit: atoms with less or equal frequency are
			removed from DM */
  {
    char FNAME[ FMSIZE ];
    filedesc *FDESC;    /* Pointer to the file description item */
    atomdesc *INV;
    unsigned INVS;
    CURR_FPOINTER = NULL;
    ACTUAL_FILE = -1;
    FDESC = (filedesc *)DmFix( FILETABLE ) + FILENO;
    if( FDESC IS_CANCELLED )
      {
	DmUnfix( FILETABLE );
	return( 1 );
      }

    strcpy( FNAME, FDESC->DESC.PATHNODE );
    strcat( FNAME, WORK_EXT );
/*
    strmfe( FNAME, FDESC->DESC.PATHNODE, WORK_EXT );
*/
    if( FDESC->APPASIZE == 0L && FDESC->FLAGS & NEWFILE )
      { /*If the apparent size is 0, remove the NEWFILE file */
	if( FDESC->FPOINTER != NULL )
	  {
	    if( fclose( FDESC->FPOINTER ) == -1 )
	      {
		PS_ERROR |= b12;
		PsError();
	      }
	    FDESC->FPOINTER = NULL;
	    ACTUAL_FILE = -1;
	    CURR_FPOINTER = NULL;
	  }
	unlink( FNAME );
	FDESC->DESC.SIZE = 0L;
	FDESC->FLAGS &= ~NEWFILE;
	FDESC->FLAGS |= APPARENT;
	DmUnfix( FILETABLE );
	return( 1 );
      }

    if( ! ( FDESC->FLAGS & (APPARENT | NEWFILE ))) /* If old file is actual */
      {
	/* Test if any atom has been updated; if so, open NEWFILE */
	unsigned NOTUPDATED = 1; 
	for( INV = BEGAOVER; DmInside( (char *)INV, AINVOVER ); INV++ )
	    if( ! ( INV IS_CANCELLED )
			&& INV IS_UPDATED && AtxFile( INV->POINT ) == FILENO )
	      {
		NOTUPDATED = 0;
		break;
	      }
	if( NOTUPDATED )
	    for( INV = BEGAMAIN, INVS = 0; INVS < AINVSIZE; INV++, INVS++ )
		if( !(INV IS_CANCELLED || INV->FLAGS & OVERFLOW)
			&& INV IS_UPDATED && AtxFile( INV->POINT ) == FILENO )
		  {
		    NOTUPDATED = 0;
		    break;
		  }
	if( ! NOTUPDATED )
	  {
	    /* Case 3: A file with WORK_EXT extension is opened, in same
	    directory as the old file, and data from the old are rewritten,
	    to size = min(FDESC->APPASIZE, FDESC->DESC.SIZE). */
	    char  OLDNAME[ FMSIZE ];
	    long MINSIZE = FDESC->DESC.SIZE < FDESC->APPASIZE ?
				FDESC->DESC.SIZE : FDESC->APPASIZE;
	    if( FDESC->FPOINTER != NULL )
	      {
		if( fclose( FDESC->FPOINTER ) == -1 )
		  {
		    PS_ERROR |= b12;
		    PsError();
		  }
		FDESC->FPOINTER = NULL;
		ACTUAL_FILE = -1;
		CURR_FPOINTER = NULL;
	      }
	    FDESC->DESC.SIZE = MINSIZE;
	    strcpy( OLDNAME, FDESC->DESC.PATHNODE );
/*
	    strmfe( OLDNAME, FDESC->DESC.PATHNODE, FDESC->DESC.EXT );
*/
	    PsRewrite( OLDNAME, FNAME, MINSIZE  );
	    if( ( FDESC->FPOINTER = fopen( FNAME, "rb+" ) ) == NULL )
	      {
		DmUnfix( FILETABLE );
		PS_ERROR |= b1; /* b1 - cannot open */
		return( 0 );
	      }
	    FDESC->FLAGS |= NEWFILE | UPDATED;
	  }
      }
    else if( FDESC->FLAGS & APPARENT )  /* Case 1: Create new file */
      {
	FDESC->FLAGS &= ~APPARENT;              /* Cancel flag APPARENT */
	FDESC->FLAGS |= NEWFILE | UPDATED;
	if( ( FDESC->FPOINTER = fopen( FNAME, "wb+" ) ) == NULL )
	  {
	    DmUnfix( FILETABLE );
	    PS_ERROR |= b1; /* b1 - cannot open */
	    return( 0 );
	  }
      }
    ACTUAL_FILE = FILENO;
    CURR_FPOINTER = FDESC->FPOINTER;
    /* Cancel non-updated or write updated atoms to the new file and then
	remove them from DM if FRQ is less */
    for( INV = BEGAOVER; DmInside( (char *)INV, AINVOVER ); INV++ )
	if( ! ( INV IS_CANCELLED ) )
	    WrtAtom( FDESC, FILENO, INV, FRQ );
    for( INV = BEGAMAIN, INVS = 0; INVS < AINVSIZE; INV++, INVS++ )
	if( ! (INV IS_CANCELLED || INV->FLAGS & OVERFLOW ) )
	    WrtAtom( FDESC, FILENO, INV, FRQ );


    if( fclose( FDESC->FPOINTER ) == -1 )
      {
	PS_ERROR |= b12 | b3;
	PsError();
      }
    FDESC->FPOINTER = NULL;
    ACTUAL_FILE = -1;
    CURR_FPOINTER = NULL;
    DmUnfix( FILETABLE );
    return( 1 );
  }


/*---------------------------------------------------------------------------*/
/* Note on organization of atom inventory:
The atom inventory is stored in two fixed DM objects AINVMAIN and AINVOVER. The
first object has a fixed size given by AINVSIZE. The object beginnings are
given in BEGAMAIN and BEGAOVER. Maintenance and searching in the inventory is
organized by the hash-coding method. A hash macro HashPoint takes a VPOS point
of an atom and returns a relative position of a corresponding AINVMAIN
inventory item. If two or more points hit the same item, it means overflow.
In such a case all candidate items are stored in a second DM object AINVOVER,
without order; its size is not fixed. In case of overflow a corresponding item
of AINVMAIN has flag OVERFLOW set to 1; two first bytes of POINT store lower
bound (relatively to the beginning) of searching interval in AINVOVER, two next
bytes of POINT store upper bound of the searching interval, and DMOBJ store
the number of overflow items stored in AINVOVER.

static unsigned CONFLICTS = 0; Number of conflicts allocated in the BEGAOVER
*/

static unsigned ALLOCATED = 0; /* Number of allocated atoms */
#define ALDENSITY 45    /* allowed % of allocated objects in comparison to AINVSIZE */

/*-------------------------------------------------------- AtxAllocate ------*/
/* Create new atom inventory item */

static atomdesc *  AtxAllocate( FLAGS, POINT, DMOBJ )
    int  FLAGS;
    long POINT;
    unsigned short int DMOBJ;
  {
    atomdesc *IT = BEGAMAIN + HashPoint( POINT );

    /* Assertion */
    if( PsIsGarbage( POINT ) )
      {
	printf( "! A reference to garbage %ld", POINT );
	PS_ERROR |= b11;
	PsError(); 
      }
    ALLOCATED++;
    if( IT IS_CANCELLED )
      {
	IT->FLAGS = FLAGS;
	IT->POINT = POINT;
	IT->DMOBJ = DMOBJ;
      }
    else
      {
	atomdesc *OVER;
	unsigned REL;
	int NOTFOUND = 1;
	unsigned short int *P = (unsigned short int *)&IT->POINT;
	if( (long)ALLOCATED * 100 >= (long)AINVSIZE * ALDENSITY )
	    DM_COMPLAIN = 1;
	if( IT->FLAGS & OVERFLOW )
	  {
/*          CONFLICTS++; */
	    for( OVER = BEGAOVER; DmInside( (char *)OVER, AINVOVER ); OVER++ )
		if( OVER IS_CANCELLED )
		  {
		    NOTFOUND = 0;
		    break;
		  }
		else /* Assertion */
		    if( POINT == OVER->POINT )
		      {
printf( "1 Conflicting point %ld\n", POINT );
			PS_ERROR |= b11;
			PsError();
		      }
	    if( NOTFOUND )
	      {
		DmUnfix( AINVOVER );
		DmExtend( AINVOVER, sizeof( atomdesc ) );
		BEGAOVER = (atomdesc *)DmFix( AINVOVER );
		OVER = (atomdesc *)( DmTopPointer( AINVOVER ) ) - 1;
	      }
	    REL = (unsigned)( OVER - BEGAOVER );
	    if( REL < *P )
		*P = REL;
	    else if( *( P + 1 ) < REL )
		*( P + 1 ) = REL;
	    IT->DMOBJ++;
	  }
	else    /* The single busy position - make overflow */
	  {
/*          CONFLICTS += 2; */
	    /* Assertion */
	    if( POINT == IT->POINT )
	      {
printf( "2 Conflicting point %ld\n", POINT );
		PS_ERROR |= b11;
		PsError();
	      }
	    for( OVER = BEGAOVER; DmInside( (char *)OVER, AINVOVER ) &&
			  DmInside( (char *)(OVER + 1), AINVOVER ); OVER++ )
		if( (OVER + 1) IS_CANCELLED )
		  {
		    if( OVER IS_CANCELLED )
		      {
			NOTFOUND = 0;
			break;
		      }
		  }
		else
		    OVER++;
	    if( NOTFOUND )
	      {
		DmUnfix( AINVOVER );
		DmExtend( AINVOVER, sizeof( atomdesc ) + sizeof( atomdesc ) );
		BEGAOVER = (atomdesc *)DmFix( AINVOVER );
		OVER = (atomdesc *)( DmTopPointer( AINVOVER ) ) - 2;
	      }
	    *( OVER+1 ) = *IT;
	    REL = (unsigned)( OVER - BEGAOVER );
	    IT->FLAGS = OVERFLOW;
	    *P = REL;
	    *(P + 1) = REL + 1;
	    IT->DMOBJ = 2;
	  }
	OVER->FLAGS = FLAGS;
	OVER->POINT = POINT;
	OVER->DMOBJ = DMOBJ;
	return( OVER );
      }
    return( IT );
  }


/*-------------------------------------------------------- AtxDeAllocate ----*/
/* Remove an item from atom inventory; no action for DMOBJ with the atom */

static void  AtxDeAllocate( POINT )
    long POINT;
  {
    atomdesc *IT = BEGAMAIN + HashPoint( POINT );
    if( ! ( IT IS_CANCELLED ) )
      {
	ALLOCATED--;
	if( IT->FLAGS & OVERFLOW )
	  {
	    atomdesc *OVER;
	    int K = 0;
	    unsigned short int *P = (unsigned short int *)&IT->POINT;
	    atomdesc *OVEREND = BEGAOVER + *( P + 1 );
/*          CONFLICTS--; */
	    for( OVER = BEGAOVER + *P; OVER <= OVEREND; OVER++ )
		if( !(OVER IS_CANCELLED) && OVER->POINT == POINT )
		  {
		    OVER->FLAGS = CANCELLED;
		    K = 1;
		    break;
		  }
	    if( K )
	      {
		IT->DMOBJ--;
		if( IT->DMOBJ == 1 ) /* Reduce the overflow */
		  {
/*                  CONFLICTS--; */
		    for( OVER = BEGAOVER + *P; OVER <= OVEREND; OVER++)
		      {
			if( !(OVER IS_CANCELLED) &&
				BEGAMAIN + HashPoint( OVER->POINT ) == IT )
			  {
			    *IT = *OVER;
			    OVER->FLAGS = CANCELLED;
			    break;
			  }
		      }
		  }
	      }
	  }
	else if( IT->POINT == POINT )
	    IT->FLAGS = CANCELLED;
      }
  }


/*-------------------------------------------------------- AtxLocate --------*/
/* Return pointer to atom description in atom inventory,or NULL if not found */

static atomdesc *  AtxLocate( POINT )
    long POINT;
  {
    atomdesc *IT = BEGAMAIN + HashPoint( POINT );
    if( ! (IT IS_CANCELLED) )
      {
	if( IT->FLAGS & OVERFLOW )
	  {
	    unsigned short int *P = (unsigned short int *)&IT->POINT;
	    atomdesc *OVER = BEGAOVER + *P;
	    atomdesc *OVEREND = BEGAOVER + *( P + 1 );
	    for( ; OVER <= OVEREND; OVER++ )
		if( ! (OVER IS_CANCELLED) && OVER->POINT == POINT )
		    return( OVER );
	  }
	else if( IT->POINT == POINT )
	    return( IT );
      }
    return( NULL );
  }  


/*-------------------------------------------------------- PsLongGar --------*/
/* Change long POINT and SIZE into garbage description */

void  PsLongGar( GARBAGE, GARPOINT, GARSIZE )
    gardesc *GARBAGE;
    long GARPOINT;
    long GARSIZE;
  {
    GARBAGE->GARPOINT = GARPOINT;
    GARBAGE->GARSIZE  = GARSIZE;
  }


/*-------------------------------------------------------- PsFindArea -------*/
/* Give the pointer to garbage area of a required SIZE, update proper
   garbage table, and, if necessary, the file size */

/* Suggestion is too strong: in this way the file tends to grow. Some criterion
should be introduced to make possible utilization of the beginning of the file.
E.g. randomly switch off the suggestion */

static RANDOMSUGG = 0;

static long  PsFindArea( SIZE, SUGGEST )
    long SIZE;
    long SUGGEST;
  {
    long RESULT;
    filedesc *FD;
    CHANGES_IN_VPOS = 1;
    FD = (filedesc *)DmFix( FILETABLE ) + CURR_FILE;
    RANDOMSUGG = (RANDOMSUGG + 1) % 4;
    SUGGEST = RANDOMSUGG &&
	      SUGGEST != NULLPNTR &&
	      AtxFile( SUGGEST ) == CURR_FILE ?
						AtxPoint( SUGGEST ) : 0L;
    if( FD->GARBAGE != 0 && DmSize( FD->GARBAGE) > 0L ) /* If garbage exists */
      {
	gardesc *GD, *BGD = (gardesc *)DmFix( FD->GARBAGE );
	long MINDISTANCE = MAXFILE;
	gardesc *HIT = NULL;
	/* Find the garbage item most close to the suggestion. Do not leave
	the garbage smaller than ATOMHEAD */
	for( GD = (gardesc *)( DmTopPointer( FD->GARBAGE ) ) - 1;
				DmInside( (char *)GD, FD->GARBAGE ); GD-- )
	  {
	    long S, GARSIZE;
	    PsGarLong( GD, S, GARSIZE );
	    if( GARSIZE == SIZE )
	      {
		HIT = GD;
		break;
	      }
	    else if( GARSIZE >= SIZE + ATOMHEAD )
	      {
		long DISTANCE = S > SUGGEST ? S - SUGGEST : SUGGEST - S;
		if( MINDISTANCE > DISTANCE )
		  {
		    MINDISTANCE = DISTANCE;
		    HIT = GD;
		  }
		else
		    break;
	      }
	  }
	if( HIT != NULL )
	  {
	    long S, GARSIZE;
	    PsGarLong( HIT, RESULT, GARSIZE );
	    GARSIZE -= SIZE;
	    S = RESULT + SIZE;
	    PsLongGar( HIT, S, GARSIZE );
	    DmUnfix( FD->GARBAGE );
	    if( GARSIZE == 0L )
		DmExtract( FD->GARBAGE, (long)((char *)HIT - (char *)BGD),
							sizeof(gardesc));
	    DmUnfix( FILETABLE );
/* Note difference of SUN and PC: in PC the below command looks like */
	    *(((char *)&RESULT) + 3) = CURR_FILE;
/*	    *(unsigned char *)&RESULT = CURR_FILE;       09.06.96 AJ*/
	    return( RESULT );
	  }
	DmUnfix( FD->GARBAGE );
      }
    RESULT = FD->APPASIZE;
    FD->APPASIZE += SIZE;       /* Increase apparent size */
    DmUnfix( FILETABLE );
    if( RESULT <= MAXFILE )
      {
/* Note difference of SUN and PC: in PC the below command looks like */
	*(((char *)&RESULT)+3) = (char)CURR_FILE;

/*	*(unsigned char *)&RESULT = (unsigned char)CURR_FILE;  09.06.96 */
	return( RESULT );
      }
    PS_ERROR |= b13; /* Too large file */
    return( NULLPNTR );
  }


/*-------------------------------------------------------- PsFindAfter ------*/
/* Return 1 if after the point there is a free area of a required SIZE,
   and update properly garbage table */

static int  PsFindAfter( POINT, SIZE )
    long POINT;
    long SIZE;
  {
    long S;
    long CPOINT;
    long CSIZE;
    filedesc *FD;
    CHANGES_IN_VPOS = 1;
    CPOINT = AtxPoint( POINT );
    CSIZE = AtxSize( POINT );
    /* Fix a proper garbage table */
    FD = (filedesc *)DmFix( FILETABLE ) + AtxFile( POINT );
    if( FD->GARBAGE != 0 )      /* If garbage exists */
      {
	gardesc *GD, *BGD;
	for( GD = BGD = (gardesc *)DmFix( FD->GARBAGE );
		DmInside( (char *)GD, FD->GARBAGE ); GD++ )
	  {
	    long GARSIZE;
	    PsGarLong( GD, S, GARSIZE );
	    if( CPOINT + CSIZE < S )
		break; /* No reserve atfter the atom */
	    else if( CPOINT + CSIZE == S )
	      {
		if( SIZE > GARSIZE )    /* The reserve is insufficient */
		    break;
		/* The reserve is sufficient */ 
		if( SIZE == GARSIZE )   /* The reserve is exact */
		  {
		    DmUnfix( FD->GARBAGE );
		    DmExtract( FD->GARBAGE, (long)((char *)GD - (char *)BGD),
				sizeof( gardesc ) );
		  }
		else
		  {
		    GARSIZE -= SIZE;
		    S += SIZE;
		    PsLongGar( GD, S, GARSIZE );
		    DmUnfix( FD->GARBAGE );
		  }
		DmUnfix( FILETABLE );
		return( 1 );
	      } 
	  }
	DmUnfix( FD->GARBAGE );
      }
    if( CPOINT + CSIZE == FD->APPASIZE )        /* If the end of file */
      {
	FD->APPASIZE += SIZE;
	DmUnfix( FILETABLE );
	return( 1 );
      }
    DmUnfix( FILETABLE );
    return( 0 );
  }


/*-------------------------------------------------------- PsMakeGarbage ----*/
/* Return the atom area to garbage */

static void  PsMakeGarbage( POINT )
    long POINT;
  {
    atomdesc *AD;
    int  FILENO  = AtxFile(  POINT );
    long CPOINT  = AtxPoint( POINT );
    long CSIZE   = AtxSize(  POINT );
	/* Remove the atom from atom inventory */
    if( ( AD = AtxLocate( POINT ) ) != NULL )
      {
	DmDelete( AD->DMOBJ );
	AtxDeAllocate( AD->POINT );
      }
    PsGarbage( FILENO, CPOINT, CSIZE );
  }


/*----- External ----------------------------------------- PsIsGarbage ------*/
/* Return 1 if the point belongs to garbage and 0 otherwise */

int PsIsGarbage( POINT )
    long POINT;
  {
    long NPOINT, NSIZE;
    unsigned short int DMOBJ;
    int FILENO = AtxFile( POINT );
    long CPOINT = AtxPoint( POINT );
    filedesc *FD = (filedesc *)DmFix( FILETABLE ) + FILENO;
    gardesc *GI;
    if( ! DmInside( (char *)FD, FILETABLE ) || FD IS_CANCELLED ||
	CPOINT >= FD->APPASIZE )
      {
	DmUnfix( FILETABLE );
	return( 1 );
      }
    DMOBJ = FD->GARBAGE;
    if( ! DMOBJ )
      {
	DmUnfix( FILETABLE );
	return( 0 );
      }
    for( GI = (gardesc *)DmFix( DMOBJ ); DmInside( (char *)GI, DMOBJ ); GI++ )
      {
	PsGarLong( GI, NPOINT, NSIZE );
	if( NPOINT <= CPOINT && CPOINT < NPOINT + NSIZE )
	  {
	    DmUnfix( FILETABLE );
	    DmUnfix( DMOBJ );
	    return( 1 );
	  }
	if( NPOINT > CPOINT )
	    break;
      }
    DmUnfix( FILETABLE );
    DmUnfix( DMOBJ );
    return( 0 );
  }


/*-------------------------------------------------------- PsGarbage --------*/
/* Make garbage; given file, point, and size */

static void  PsGarbage( FILENO, CPOINT, CSIZE )
    int FILENO;
    long CPOINT;
    long CSIZE;
  {
    unsigned short int DMOBJ;
    long PPOINT, PSIZE, NPOINT, NSIZE, FROM;
    filedesc *FD;
    gardesc *BITEM, *GITEM;
    int STICK = 0;


    if( CSIZE <= 0L )
	return;
    CHANGES_IN_VPOS = 1;
    FD = (filedesc *)DmFix( FILETABLE ) + FILENO;
    if( ! DmInside( (char *)FD, FILETABLE ) || ( FD IS_CANCELLED ) )
      {
	PS_ERROR |= b11;
	printf( FindString( sWRONG_FILE ), FILENO );
	getchar();
	return;
      }
    if( FD->GARBAGE == 0 )
	FD->GARBAGE = DmCreate( 0 );
    DMOBJ = FD->GARBAGE;
    /* Find proper item in the garbage inventory */
    for( BITEM = GITEM = (gardesc *)DmFix( DMOBJ );
				DmInside( (char *)GITEM, DMOBJ ); GITEM++ )
      {
	PsGarLong( GITEM, NPOINT, NSIZE );
	if( NPOINT > CPOINT )
	    break;
      }
/* 4 cases: new item,stick with previous, stick with next,or stick with both */
    /* Check possibility to stick with previous */
    if( DmInside( (char *)(GITEM - 1), DMOBJ ) )
      {
	PsGarLong( GITEM - 1, PPOINT, PSIZE );
	if( PPOINT + PSIZE >= CPOINT )  /* Toleration to internal errors: previous may overlap */
	    STICK = 1;
      }
    /* Check possibility to stick with next */          
    if( DmInside( (char *)GITEM, DMOBJ ) )
	if( CPOINT + CSIZE >= NPOINT )  /* Toleration to internal errors: next may overlap */
	    STICK += 2;
    FROM = (long)( (char *)GITEM - (char *)BITEM );
    /* 4 cases */
    switch( STICK )
      {
	case  0:        /* No sticking */
	    DmUnfix( DMOBJ );
	    DmDistend( DMOBJ, FROM, sizeof( gardesc ) );
	    GITEM = (gardesc *)(DmPointer( DMOBJ ) + FROM );
	    PsLongGar( GITEM, CPOINT, CSIZE );
	    break;
	case  1:        /* Stick with previous */
	    PsLongGar( GITEM - 1, PPOINT, CPOINT + CSIZE - PPOINT );
	    DmUnfix( DMOBJ );
	    break;
	case  2:        /* Stick with next */
	    PsLongGar( GITEM, CPOINT, NPOINT + NSIZE - CPOINT );
	    DmUnfix( DMOBJ );
	    break;
	case  3:        /* Stick with previous and next */
	    PSIZE = NPOINT + NSIZE - PPOINT;
	    PsLongGar( GITEM - 1, PPOINT, PSIZE );
	    PsLongGar( GITEM, PPOINT, PSIZE );
	    DmUnfix( DMOBJ );
	    DmExtract( DMOBJ, FROM, sizeof( gardesc ) );
	    break;
      }

    /* If possible, decrease the apparent size of the file */
    GITEM = (gardesc *)( DmTopPointer( DMOBJ ) ) - 1; /*Last garbage*/
    PsGarLong( GITEM, NPOINT, NSIZE );
    if( NPOINT + NSIZE == FD->APPASIZE )
      {
	FD->APPASIZE -= NSIZE;
	DmPop( DMOBJ, sizeof(gardesc) );
      }
    DmUnfix( FILETABLE );
/* Testing 
printf( "---------------------------------- Garbage Table - APPASIZE %ld\n",
FD->APPASIZE );
    for( GITEM = (gardesc *)DmFix( DMOBJ );
				DmInside( (char *)GITEM, DMOBJ ); GITEM++ )
      {
	PsGarLong( GITEM, NPOINT, NSIZE );
printf("            %4ld -%4ld\n", NPOINT, NPOINT+NSIZE-1);
      }
    DmUnfix( DMOBJ );
printf( "+++++++++++++++++++++++++++++++++++++++++++++++++++++++\n" );
*/ 
  }


/*----- External ----------------------------------------- AtmReferences ----*/
/* Return in a DM object all points of logical pointers pointing the atom */

unsigned short int AtmReferences( POINT )
    long POINT;
  {
    SkipHolders( POINT );
    POINT = AtxNext( POINT );
    if( AtxType( POINT ) == BACKWARD )
      {
	long *REF;
	unsigned short int TABLE = DmCreate( 0 ), TEMP = AtxGetValue( POINT );
	/* Remove backward pointers implied by place holders */
	for( REF = (long *)DmFix( TEMP ); DmInside( (char *)REF, TEMP ); REF++)
	  {
	    int T = AtxType( *REF ); /* if NULLPNTR, T is -1 */
	    if( LOGICAL_B <= T && T <= LOGICAL_E )
		DmPush( TABLE, (char *)REF, sizeof( long ) );
	  }
	DmUnfixDelete( TEMP );
	return( TABLE );
      }
    return( DmCreate( 0 ) );
  }


/*----- External ----------------------------------------- AtmFlag ---------*/
/* Give the flag stored in the atom; external */

int AtmFlag( POINT, FLAG )
    long POINT;
    int FLAG;   /* A number from 0 to 7 */
  {
    SkipHolders( POINT );
    return( ((atom_head *)DmPointer( AtxRead( POINT, 1 )->DMOBJ ))->STATUS
							& ( 0x01 << FLAG ) );
  }


/*----- External ----------------------------------------- AtmFlagOn -------*/
/* Set the give flag of atom to 1; external */

void AtmFlagOn( POINT, FLAG )
    long POINT;
    int FLAG;   /* A number from 0 to 7 */
  {
    atomdesc *AD;
    if( FLAG != 1 )
      {
	SkipHolders( POINT );
	AD = AtxRead( POINT, 1 );
	AD->FLAGS |= UPDATED;
	((atom_head *)DmPointer( AD->DMOBJ ))->STATUS |= 0x01 << FLAG;
      }
    else
	PS_ERROR |= b14;
  }


/*----- External ----------------------------------------- AtmFlagOff ------*/
/* Set the give flag of atom to 0; external */

void AtmFlagOff( POINT, FLAG )
    long POINT;
    int FLAG;   /* A number from 0 to 7 */
  {
    atomdesc *AD;
    if( FLAG != 1 )
      {
	SkipHolders( POINT );
	AD = AtxRead( POINT, 1 );
	AD->FLAGS |= UPDATED;
	((atom_head *)DmPointer( AD->DMOBJ ))->STATUS &= ~(0x01 << FLAG);
      }
    else
	PS_ERROR |= b14;
  }


/*==========================================================================*/
