/*-------------------------------------------------------------------------+
| Virtual persistent object store, miscellaneous functions (SUN version)   |
|                                                                          |
| Date of last updating: 18 November 1991                                  |
|                                                                          |
| Copyright (C) 1991 by Kazimierz Subieta.                                 |
| This software is subject to a licence agreement.                         |
| No restrictions when used for didactic purposes.                         |
+_________________________________________________________________________*/


#include	<fcntl.h>
#include	<stdio.h>
#include	<string.h>

#include	<messages.h>
#include	<dynmem.h>
#include	<vpos.h>
#include	<vpos-mis.h>

/* ======================================================================== */

static char *BG;
static unsigned short int NEWDICT, SIZE;

typedef struct /* The file description stored in DM */
  {
    char KIND;		/* 0 - cancelled, P - permanent, T - temporary */
    char DUMMY; 	/* for alignment */
    unsigned short int GARPIECES; /* How many pieces in the garbage */
    long TOTALGAR;		/* Total size of garbage areas */
    long APPASIZE;		/* Apparent file size */
  }fis;

/*----- External ----------------------------------------- PsStatistics -----*/
/* Return statistics of VPOS */

void PsStatistics( DMATOMS, FILESTAT )
    unsigned *DMATOMS;	/* Number of atoms currently in DM */
    unsigned short int *FILESTAT; /* DM table showing file statistics:
			type, size, garbage size, nbr of garbage areas;
                        see the fis struct */
  {
    if( DMATOMS != NULL )
      {
	atomdesc *AT;
	*DMATOMS = 0;
	for( AT = BEGAOVER; DmInside( (char *)AT, AINVOVER ); AT++ )
	    if( ! ( AT IS_CANCELLED ) )
		(*DMATOMS)++;
	for( AT = BEGAMAIN; DmInside( (char *)AT, AINVMAIN ); AT++ )
	    if( ! ( AT->FLAGS & OVERFLOW || AT IS_CANCELLED ) )
		(*DMATOMS)++;
      }
    if( FILESTAT != NULL )
      {
	filedesc *FD;
	fis FIS;
	*FILESTAT = DmCreate( 0 );
	for( FD = (filedesc *)DmFix( FILETABLE );
			DmInside( (char *)FD, FILETABLE ); FD++ )
	  {
	    FIS.TOTALGAR = FIS.APPASIZE = 0L;
	    FIS.GARPIECES = 0;
	    FIS.KIND = FD IS_CANCELLED ? '0':(FD->FLAGS & TEMPORARY) ? 'T':'P';
	    if( ! ( FD IS_CANCELLED ) )
	      {
		FIS.APPASIZE = FD->APPASIZE;
		if( FD->GARBAGE )
		  {
		    gardesc *G;
		    for( G = (gardesc *)DmFix( FD->GARBAGE );
				DmInside( (char *)G, FD->GARBAGE ); G++ ) 
		      {
			long GP, GS;
			FIS.GARPIECES++;
			PsGarLong( G, GP, GS );
			FIS.TOTALGAR += GS;
		      }
		    DmUnfix( FD->GARBAGE );
		  }
	      }
	    DmPush( *FILESTAT, (char *)&FIS, sizeof(fis) );
	  }
	DmUnfix( FILETABLE );
      }
  }


/*----- External ----------------------------------------- PsNamesVolume ---*/
/* Get the total size of names in permanent dictionary */

unsigned short int PsNamesVolume()
  {
    char *V;
    unsigned RES = 0;
    for( V = DmFix( PERMNAMES ); DmInside( V, PERMNAMES ); )
      {
	int L = strlen( V ) + 1;
	if( *(unsigned char *)V != 0xFF )
	    RES += L;
	V += L;
      }
    DmUnfix( PERMNAMES );
    return( RES );
  }


/*-------------------------------------------------------- ConvPoint -------*/
int ConvPoint( POINT, BUF )
    long POINT;
    char *BUF;
  {
    char BUF1[20];
    if( POINT == NULLPNTR )
	strcpy( BUF, "NULLPNTR" );
    else
      {
	sprintf( BUF, "%id", AtxFile( POINT ) );
        strcat( BUF, "." );
	sprintf( BUF1, "%ld", AtxPoint( POINT ) );
        strcat( BUF, BUF1 );
      }
    return( strlen(BUF) );
  }


/*-------------------------------------------------------- InsertExNames ---*/
/* Auxiliary recursive procedure for PsCleanDict */

static void InsertExNames( P )
    long P;
  {
    unsigned short int NAME;
    int T = AtmTypeAndName( P, &NAME );  
    /* Insert the name if appropriate */
    if( NAME < SIZE && *(unsigned char *)(BG + NAME ) == 0xFF )
      {
	if( NAME )	/* Filler before is also finished by end-of-string */
	    *(BG + NAME - 1) = '\0';	
	PsCodeName( NAME, BG + NAME );
      }
    if( RING_B <= T && T <= RING_E || SPIDER_B <= T && T <= SPIDER_E )
	for( P = AtmSubord( P ); P != NULLPNTR; P = AtmNext( P ) )
	    InsertExNames( P );
  }


/*----- External ----------------------------------------- PsCleanDict -----*/
/* Main procedure for cleaning the dictionary; assumed the database is open */

void PsCleanDict( )
  {
    long POINT;
    char *V;
    unsigned CUT = 0;
    CHANGES_IN_VPOS = 1;
    SIZE = DmSize( PERMNAMES );
    /* Create a new dictionary of the same size as PERMNAMES */
    NEWDICT = DmCreate( SIZE );
    /* Fill the NEWDICT with the filler */
    BG = DmFix( NEWDICT );
    for( V = BG; DmInside( V, NEWDICT ); V++ )
	*V = 0xFF;
    for( POINT = 0L; POINT != NULLPNTR; POINT = AtmNext( POINT ) )
	InsertExNames( POINT );
    /* Cut the end of NEWDICT if possible */
    for( V = BG + SIZE - 1; DmInside( V, NEWDICT ); V-- )
      {
	if( *(unsigned char *)V != 0xFF )
	    break;
	CUT++;
      }
    DmUnfix( NEWDICT );
    DmPop( NEWDICT, CUT );
    DmDelete( PERMNAMES );
    PERMNAMES = NEWDICT;
  }

/*--------------------------------------------------------------------------*/
static int VPOS_CONSISTENT;	/* Shows consistency of VPOS */


/* Testing procedure: display all information about the given point; begin the
display in a new line, precede it by LEVEL charaters -.
The procedure is recursive.*/

void DispPoint( FP, POINT, LEVEL )
    FILE *FP;
    long POINT;
    int LEVEL;
  {
    char	STRNAME[ 200 ];
    char	*TYPENAME;
    int		SPACES, C, TYPE, STATUS;
    int I;
    char MASK = 0x01;
    unsigned short int	DMOBJ, NAME;
    atom_head *AH;
    if( ! VPOS_CONSISTENT )
       return;
    fprintf( FP, "\n" );
    for( SPACES = 0; SPACES < LEVEL; SPACES++ )
	fprintf( FP, "- " );
    fprintf( FP, "%d.%-4ld ", AtxFile( POINT ), AtxPoint( POINT ) );
    if( PsIsGarbage( POINT ) )
      {
	fprintf( FP, FindString( sINCONSISTENCY_GARBAGE ) );
	VPOS_CONSISTENT = 0;	/* Shows consistency of VPOS */
	return;
      }

    DMOBJ	= AtxRead( POINT, 0 )->DMOBJ;
    AH		= (atom_head *)DmFix( DMOBJ );
    TYPE	= AH->TYPE;
    NAME	= AH->NAME;
    STATUS	= AH->STATUS;

    if( DM_COMPLAIN )
	PsAtxDel( );

    TYPENAME =	TYPE == RING				? "RING    "	:
		TYPE == AGGREGATE			? "AGGREGATE"	:
		TYPE == VPOSLINK			? "LINK    "	:
		TYPE == ANCESTOR			? "ANCESTOR"	:
		TYPE == VPOSINT				? "INTEGER "	:
		TYPE == VPOSREAL			? "REAL    "	:
		TYPE == VPOSSTRING			? "STRING  "	:
		TYPE == LOQISTEXT			? "LOQISTEXT"	:
		TYPE == PRGAGNES			? "PRGAGNES"	:
		TYPE == PRGDUMMY			? "PRGDUMMY"	:
		TYPE == STR_TABLE			? "STR_TABLE"	:
		TYPE == FREEDATA			? "FREEDATA"	:
		TYPE == PRGELEM				? "PRGELEM"	:
		TYPE == SPIDER				? "OBJECT"	:
		TYPE == HOLDER				? "HOLDER"	:
		TYPE == TOEND				? "TOEND   "	:
		TYPE == BACKWARD			? "BACKWARD"	:
		RING_B    <= TYPE && TYPE <= RING_E	? "?RING"	:
		LOGICAL_B <= TYPE && TYPE <= LOGICAL_E	? "?LOGICAL"	:
		LONGINT_B <= TYPE && TYPE <= LONGINT_E	? "?INTEGER"	:
		BYTES_B   <= TYPE && TYPE <= BYTES_E	? "?BYTES  "	:
		SPIDER_B  <= TYPE && TYPE <= SPIDER_E	? "?SPIDER"	:
							  "????";
    fprintf( FP, TYPENAME );
    fprintf( FP, " %3d <", TYPE );
    if( NAME < 0xFFE0 )
      {
	PsCodeName( NAME, STRNAME );
	fprintf( FP, "%s>", STRNAME );
      }
    else {
	STRNAME[0] = '\0';
        sprintf( STRNAME, "0x%X", NAME );
	fprintf( FP, "%s>", STRNAME );
     }

    fprintf( FP, "[" );
    for( I = 0; I < 8; I++ )
      {
	if( MASK & STATUS )
	    fprintf( FP, "%d", I );
	MASK = MASK << 1;
      }
    fprintf( FP, "]" );

    if( RING_B <= TYPE && TYPE <= RING_E ||
					SPIDER_B <= TYPE && TYPE <= SPIDER_E )
      {
	long SUB;
	if( SPIDER_B <= TYPE && TYPE <= SPIDER_E )
	  {
	    spider_item *REF = (spider_item *)( (char *)AH + ATOMHEAD );
	    fprintf( FP, " ->" );
	    for( ; DmInside( (char *)REF, DMOBJ ); REF++ )
	        if( REF->MEMBER == NULLPNTR )
		    fprintf( FP, " NULLPNTR " );
	        else
		    fprintf( FP, " %d.%ld", AtxFile(  REF->MEMBER ),
		                            AtxPoint( REF->MEMBER ) );
	  
	  }
	for( SUB = AtxSubord(POINT); SUB != NULLPNTR; SUB = AtxNext(SUB) )
	  {
	    long OW;
	    DispPoint( FP, SUB, LEVEL + 1 );
	    if( ! VPOS_CONSISTENT )
	       break;
	    OW = AtxOwner( SUB );
	    if( OW != POINT )
		fprintf( FP, FindString( sINCONSISTENCY_OWNER ),
			AtxFile( OW ), AtxPoint( OW ),
			AtxFile( POINT ), AtxPoint( POINT ) );
	  }
      }
    else if( LOGICAL_B <= TYPE && TYPE <= LOGICAL_E ||
					TYPE == HOLDER || TYPE == TOEND )
      {
	if( AH->VALUE == NULLPNTR )
	    fprintf( FP, " NULLPNTR " );
	else
	    fprintf( FP, " -> %d.%ld ", AtxFile( AH->VALUE ),
						AtxPoint( AH->VALUE ) );
      }
    else if( LONGINT_B <= TYPE && TYPE <= LONGINT_E )
	fprintf( FP, " %ld ", AH->VALUE );
    else if( TYPE == FREEDATA && NAME == 0xFFF7 )
      {
	unsigned short int *REF = (unsigned short int *)( (char *)AH + ATOMHEAD );
	fprintf( FP, " EXPORTED->" );
	for( ; DmInside( (char *)REF, DMOBJ ); REF++ )
	  {
	    PsCodeName( *REF, STRNAME );
	    fprintf( FP, " %s", STRNAME );
	  }
      }
    else if( BYTES_B <= TYPE && TYPE <= BYTES_E )
      {
	fprintf( FP, FindString( sOF_SIZE ), AH->VALUE );
	if( TYPE == VPOSSTRING || TYPE == 44 /* STANDARD_TYPE */ )
	  {
	    C = AH->VALUE <= 40 ? AH->VALUE : 40;
	    memcpy( STRNAME, (char *)AH + ATOMHEAD, C );
	    *(STRNAME + 40) = '\0';
	    fprintf( FP, "%s", STRNAME );
	  }
	else if( TYPE == VPOSREAL )
	    fprintf( FP, "%f", *(double *)((char *)AH + ATOMHEAD) );
      }
    else if( BACKWARD == TYPE )
      {
	long *REF = (long *)( (char *)AH + ATOMHEAD );
	fprintf( FP, " ->" );
	for( ; DmInside( (char *)REF, DMOBJ ); REF++ )
	    if( *REF == NULLPNTR )
		fprintf( FP, " NULLPNTR " );
	    else
		fprintf( FP, " %d.%ld", AtxFile( *REF ), AtxPoint( *REF ) );
      }
    DmUnfix( DMOBJ );
  }


/*--------------------------------------------------------------------------*/
void DispWholeVPOS( FP )
    FILE *FP;
  {
    long POINT;
    filedesc *FD;
    VPOS_CONSISTENT = 1;	/* Shows consistency of VPOS */
    FD = (filedesc *)DmFix( FILETABLE );

    fprintf( FP, 
    "\n-------------------------------------------------------------------\n");
    fprintf( FP, 
    "VIRTUAL OBJECT STORE CONTENT\nMain file: %s\n", FD->DESC.PATHNODE );
    DmUnfix( FILETABLE );
    fprintf( FP, 
    "Each line of the display contains:\n");
    fprintf( FP, 
    "ObjectId 0bjTypeName ObjTypeNbr <ObjName> [Flags] Value\n");
    fprintf( FP, 
    "-------------------------------------------------------------------");
    for( POINT = 0L; POINT != NULLPNTR; POINT = AtxNext( POINT ) )
      {
	DispPoint( FP, POINT, 0 );
	if( ! VPOS_CONSISTENT )
	  {
	    fprintf( FP,
		"\n! The object store is inconsistent; cannot continue!");
	    break;
	  }
      }
    fprintf( FP, 
    "\n-------------------------------------------------------------------\n");
  }


/*-------------------------------------------------------- InternalReact2 --*/
/* Internal reaction to inconsistency: optional removing of the atom */

static void  InternalReact2( POINT, OWNER )
    long POINT, OWNER;
  {
    int RESULT = 0;
    if( 1 /* w_confirm( FindString( sTRY_DELETE ), "Y", 10, 25, 20, 56 ) */ >0 )
      {
	int TYPE = AtxType( POINT );
/*
	PushInfoTextMsg( sTRIAL_REMOVE );
	NextInfoLine( SYSTEM_MSG );
*/
	if( RING_B <= TYPE && TYPE <= RING_E ||
					TOEND <= TYPE && TYPE < BACKWARD )	      
	  {
	    /* Disconnect the atom making no action on garbage */
	    if( OWNER == NULLPNTR )
	      {
		int O_TYPE;
		OWNER = AtxOwner( POINT );
		O_TYPE = AtxType( OWNER );
		if( SPIDER_B <= O_TYPE && O_TYPE <= SPIDER_E )
		    DiscFromSpider( OWNER, POINT );
		else
		  {
		    /* If after the point there is a BACKWARD atom, remove it */
		    if( AtxType( AtxNext( POINT ) ) == BACKWARD )
			AtxDelete( NULLPNTR, AtxNext( POINT ) );
		    AtxDisconnect( POINT );
		  }
		RESULT = 1;
	      }
	    else
	      {
		/* Owner of the point is improper thus it cannot be used; 
		   instead the OWNER is true owner */
		int O_TYPE = AtxType( OWNER );
		if( SPIDER_B <= O_TYPE && O_TYPE <= SPIDER_E )
		  {
		    DiscFromSpider( OWNER, POINT );
		    RESULT = 1;
		  }
	      }
	  }
	/* The other actions are not introduced so far; the reaction will be
	   programmed according to current needs */
/*
	if( RESULT )
	  {
	    PushInfoTextMsg( sSUCCESSFULLY_DONE );
	    NextInfoLine( SYSTEM_MSG );
	  }
	else
	  {
	    printf( FindString( sSORRY_CANNOT_DO_ ) );
	    PushInfoTextMsg( sSORRY_CANNOT_DO );
	    NextInfoLine( SYSTEM_MSG );
	  }
*/
      }
  }

/*==========================================================================*/
/* A new version */

static unsigned short int OCCUPIED = 0; /* DM object storing inventory of DM objects
				 which for particular files store areas
				 currently occupied by atoms. For each such
				 object the occcupied area is stored as
				 two long-s: beginning and one after end.
				 Sticking is done as usual */ 
static unsigned short int *OCTAB;		/* Pointer to table of DM object with loading*/

static long NBR_OF_ATOMS;	/* Number of atoms in VPOS */
static long SIZE_OF_VALUES;     /* Total size of values utilized in VPOS */

/*==========================================================================*/
/* The procedure for reorganization of a database by: making physical order
of atoms corresponding to the logical order, removing all garbage areas,
removing all cancelled items in spiders and backward atoms. The procedure
assumes as input a closed database.
     Stages of the procedure:
  1. Open a database.
  2. 1-st full pass through the database: For each spider and each backward
     atom all cancelled items are removed.
  3. All garbage atoms are cleaned, that is, they will contain empty lists
     of garbage items.
  4. 2-nd full pass through the database: For each file and for each atom a new
     point is computed; for atoms of type LOGICAL and for atoms pointed by
     LOGICAL the pair <old atom point, new atom point> is stored at a special
     file (or DM object). The same for atoms pointed by TOEND atoms.
  5. 3-st full pass through the database: the whole database is rewritten to
     new files; for all logical pointers and backward pointers new points are
     computed according to the TRANSFORMATION table obtained in the second
     pass through the database.
  6. Renaming of files as for close;
  7. Cancel updates in the database.
  */

typedef struct
  {
    FILE *FPOINTER;		/* File pointer */
    char PATHNODE[ FMSIZE ]; 	/* Path + file node */
  } f_size_pointer;
  
typedef struct
  {
    long  OLD, NEW;
  } point_pair;

static FILE *TRANSFORMATION;
static long *L_OCTAB;
static f_size_pointer *F_OCTAB;

static long		ATOM_NBR;
static long		TRAN_SIZE, CURRENT_NBR, CURRENT_SEEK;
static point_pair	CURRENT_PAIR;
static unsigned short int		DM_INDX;
#define MAX_BLOCK	4096

/*------------------------------------------------------- RemoveCancelled --*/
/* Remove all cancelled items if the point is spider or backward and remove
   them in all subordinates of the point; remove all TOEND atoms */
   
static void  RemoveCancelled( POINT )
    long POINT;
  {
    int TYPE = AtxType( POINT );
    if( DM_COMPLAIN )
	PsAtxDel( );
    if( TYPE == TOEND )
	AtxDelete( NULLPNTR, POINT );
    else if( RING_B <= TYPE && TYPE <= RING_E )
      {
	long NEXT_POINT;
	for( POINT = AtxSubord( POINT ); POINT != NULLPNTR; POINT = NEXT_POINT)
	  {
	    NEXT_POINT = AtxNext( POINT );
	    RemoveCancelled( POINT );
	  }
      }
    else if( SPIDER_B <= TYPE && TYPE <= SPIDER_E )
      {
	unsigned short int NEW_SPIDER = DmCreate( 0 );
	unsigned short int DM_SPIDER = AtxRead( POINT, 0 )->DMOBJ;
	spider_item *SI = (spider_item *)( DmFix( DM_SPIDER ) + ATOMHEAD );
	for( ; DmInside( (char *)SI, DM_SPIDER ); SI++ )
	    if( SI->MEMBER != NULLPNTR )
	      {
		RemoveCancelled( SI->MEMBER );
		DmPush( NEW_SPIDER, (char *)SI, sizeof( spider_item ) );
	      }
	DmUnfix( DM_SPIDER );
	if( DmSize( NEW_SPIDER ) == DmSize( DM_SPIDER ) - ATOMHEAD )
	    DmDelete( NEW_SPIDER );
	else
	    AtxPutValue( POINT, NEW_SPIDER );
      }
    else if( TYPE == BACKWARD )
      {
	unsigned short int NEW_BACK = DmCreate( 0 );
	unsigned short int DM_BACK = AtxRead( POINT, 0 )->DMOBJ;
	long *BP = (long *)( DmFix( DM_BACK ) + ATOMHEAD );
	for( ; DmInside( (char *)BP, DM_BACK ); BP++ )
	    if( *BP != NULLPNTR )
		DmPush( NEW_BACK, (char *)BP, 4 );
	DmUnfix( DM_BACK );
	if( DmSize( NEW_BACK ) == 0 )
	  {
	    DmDelete( NEW_BACK );
	    AtxDelete( NULLPNTR, POINT );	  
	  }
	else if( DmSize( NEW_BACK ) == DmSize( DM_BACK ) - ATOMHEAD )
	    DmDelete( NEW_BACK );
	else
	    AtxPutValue( POINT, NEW_BACK );
      }
  }



/*---------------------------------------------------------- SetNextArea ---*/
/* Set the next area for the point; fill in the OCCUPIED and TRANSFORMATION
   tables */

static void  SetNextArea( POINT )
    long POINT;
  {
    int ATOMFILE = AtxFile( POINT );
    int TYPE = AtxType( POINT );
    long NEW_POINT = ATOMFILE * 16777216L + *( L_OCTAB + ATOMFILE ); 
    ATOM_NBR++;
    if( DM_COMPLAIN )
	PsAtxDel( );
    fwrite( (char *)&POINT,     4, 1, TRANSFORMATION );
    fwrite( (char *)&NEW_POINT, 4, 1, TRANSFORMATION );
    /* Increase the size for the atom's file */
    *( L_OCTAB + ATOMFILE ) += AtxSize( POINT );
    if( RING_B <= TYPE && TYPE <= RING_E ||
					SPIDER_B <= TYPE && TYPE <= SPIDER_E )
	for( POINT = AtxSubord( POINT ); POINT != NULLPNTR;
						POINT = AtxNext( POINT ) )
	    SetNextArea( POINT );
  }


/*---------------------------------------------------- SearchForNewPoint ---*/
static long  SearchForNewPoint( OLD_POINT )
    long OLD_POINT;
  {
    if( OLD_POINT != NULLPNTR )
      {
	long INCREM = 1;
	if( CURRENT_PAIR.OLD == OLD_POINT )
	    return( CURRENT_PAIR.NEW );
	while( INCREM <= ATOM_NBR )
	  {
	    long SEEK_POINT = CURRENT_SEEK + INCREM * sizeof( point_pair );
	    if( SEEK_POINT >= 0 && SEEK_POINT < TRAN_SIZE )
	      {
		unsigned short int WHICH_OBJ =
			 *( (unsigned short int *)DmPointer( DM_INDX ) +
						SEEK_POINT / MAX_BLOCK );
		CURRENT_PAIR = *(point_pair *)( DmPointer( WHICH_OBJ ) +
						SEEK_POINT % MAX_BLOCK );
		if( CURRENT_PAIR.OLD == OLD_POINT )
		  {
		    CURRENT_SEEK = SEEK_POINT;
		    return( CURRENT_PAIR.NEW );
		  }	      
	      }
	    if( INCREM > 0 )
	        INCREM = - INCREM;
	    else
	        INCREM = - INCREM + 1;
	  }
	/* Only for testing */
	printf( FindString( sWRONG_SEARCH ), OLD_POINT );
      }
     return( NULLPNTR );
  }
  

/*------------------------------------------------------- WriteToNewFile ---*/
static void  WriteToNewFile( POINT )
    long POINT;
  {
    long NEXT_POINT;
    int ATOMFILE = AtxFile( POINT );
    int TYPE = AtxType( POINT );
    f_size_pointer *FSP = F_OCTAB + ATOMFILE;
    /* Make a copy of the atom */
    unsigned short int OLD_ATOM = AtxRead( POINT, 0 )->DMOBJ;
    unsigned short int NEW_ATOM = DmCreate( 0 );
    char *BG_NEW_ATOM;
    CURRENT_NBR++;
    DmMove( NEW_ATOM, OLD_ATOM );
    if( DM_COMPLAIN )
	PsAtxDel( );
    BG_NEW_ATOM = DmFix( NEW_ATOM );
	/* Changing TO_NEXT/TO_OWNER pointer in the atom */
    ((atom_head *)BG_NEW_ATOM)->POINT =
		SearchForNewPoint( ((atom_head *)BG_NEW_ATOM)->POINT );        
	/* Changing pointers in the NEW_ATOM */
    if( RING_B <= TYPE && TYPE <= RING_E ||
		LOGICAL_B <= TYPE && TYPE <= LOGICAL_E || TYPE == TOEND )
	((atom_head *)BG_NEW_ATOM)->VALUE =
		SearchForNewPoint( ((atom_head *)BG_NEW_ATOM)->VALUE );        
    else if( SPIDER_B <= TYPE && TYPE <= SPIDER_E )
      {
	spider_item *SI = (spider_item *)( BG_NEW_ATOM + ATOMHEAD );      
	for( ; DmInside( (char *)SI, NEW_ATOM ); SI++ )
	   SI->MEMBER = SearchForNewPoint( SI->MEMBER );
      }
    else if( TYPE == BACKWARD )
      {
	long *BP = (long *)( BG_NEW_ATOM + ATOMHEAD );      
	for( ; DmInside( (char *)BP, NEW_ATOM ); BP++ )
	   *BP = SearchForNewPoint( *BP );
      }
    DmUnfix( NEW_ATOM );
    /* Write the NEW_ATOM to new file */
    AtxWrite( FSP->FPOINTER, NEW_ATOM );
    DmDelete( NEW_ATOM );
    /* Write subordinates of the POINT */
    if( RING_B <= TYPE && TYPE <= RING_E ||
			SPIDER_B <= TYPE && TYPE <= SPIDER_E )
	for( POINT = AtxSubord( POINT ); POINT != NULLPNTR; POINT = NEXT_POINT)
	  {
	    NEXT_POINT = AtxNext( POINT );
	    WriteToNewFile( POINT );
	    /* Remove the atom from DM */
	    RemNUDmAtom( AtxRead( POINT, 1 ), 255 );
	  }
  }


/*----- External --------------------------------------- PsReorganize --*/
/* Reorganize the database; return 0 iff error */

int PsReorganize( MAINFILE )
    char *MAINFILE;
  {
    if( PsOpen( MAINFILE ) )
      {
	unsigned short int *TT;
	long POINT, NEXT_POINT, REST_OF_TS, GAROWNER;
	long CUR = 0L, GAR;
	filedesc *FD;
	/* Remove the atom storing the size of main file */
	if( ( POINT = AtmFind( 0L, 0xFFF5, -1 ) ) != NULLPNTR )
	     AtxDelete( NULLPNTR, POINT );
	/* Remove all cancelled items in all spiders and in all backward atoms;
	   remove all TOEND atoms */
	for( POINT = 0L; POINT != NULLPNTR; POINT = NEXT_POINT )
	  {
	    NEXT_POINT = AtxNext( POINT );
	    RemoveCancelled( POINT );
	  }

	/* Clean all garbage atoms */
	GAROWNER = AtmFind( 0L, 0xFFFB, -1 );
	if( GAROWNER != NULLPNTR )
	    for( GAR = AtxSubord( GAROWNER );
				GAR != NULLPNTR; GAR = AtxNext( GAR ) )
	      {
	        atomdesc *AD = AtxRead( GAR, 0 );
	        AD->FLAGS |= UPDATED;
	        ((atom_head *)DmPointer( AD->DMOBJ ))->VALUE = 0;
	        DmLeave( AD->DMOBJ, 0L, ATOMHEAD );
	      }
	else	/* Create owner of the garbage and all empty memebers */
	  {
	    GAROWNER = AtmCreateAndInsert( RING_B,0xFFFB, VPOS_ENTRY,NULLPNTR );
	    /* Create garbage descriptor for each item of the file table */
	    for( FD = (filedesc *)DmFix( FILETABLE ); 
				DmInside( (char *)FD, FILETABLE ); FD++ )
		AtmCreateAndInsert( BYTES_E, 0xFFFF, GAROWNER, NULLPNTR );
	    DmUnfix( FILETABLE );
	  }
	/* Making the object with lastly occupied areas and the file
	   storing point transformation table */
	TRANSFORMATION = fopen( "TRANSFOR.$1$", "w" );
					/* Point transformation table */
	OCCUPIED = DmCreate( 0 ); /* Actually occupied sizes for pds files */
	for( FD = (filedesc *)DmFix( FILETABLE ); 
				DmInside( (char *)FD, FILETABLE ); FD++ )
	    DmPush( OCCUPIED, (char *)&CUR, 4 ); /* Clean all sizes */
	DmUnfix( FILETABLE );
	L_OCTAB = (long *)DmFix( OCCUPIED );

	/* Fill in the TRANSFORMATION table */
	ATOM_NBR = 0;
	for( POINT = 0L; POINT != NULLPNTR; POINT = AtxNext( POINT ) )
	    SetNextArea( POINT );
	
	DmUnfix( OCCUPIED );
	fclose( TRANSFORMATION );
	DmClear( OCCUPIED );

	/* Open new files */
	for( FD = (filedesc *)DmFix( FILETABLE ); 
				DmInside( (char *)FD, FILETABLE ); FD++ )
	  {
	    f_size_pointer FSP;
	    FSP.FPOINTER = NULL;
	    *FSP.PATHNODE = '\0';
	    if( ! ( FD IS_CANCELLED ) )
	      {		/* Open files for writing */
		char TEMP_FILE_NAME[ FMSIZE ];

		strcpy( TEMP_FILE_NAME, FD->DESC.PATHNODE );
		strcat( TEMP_FILE_NAME, "$0$" );
		strcpy( FSP.PATHNODE, FD->DESC.PATHNODE );
		FSP.FPOINTER = fopen( TEMP_FILE_NAME, "w" );
		/* Clean garbage tables */
		if( FD->GARBAGE )
		  {
		    DmDelete( FD->GARBAGE );
		    FD->GARBAGE = 0;
		  }
	      }
	    DmPush( OCCUPIED, (char *)&FSP, sizeof( f_size_pointer ) );
	  }
	DmUnfix( FILETABLE );
	/* Read the transformation table to several DM objects */
	DM_INDX = DmCreate( 0 );	/* Index of DM objects */
	TRANSFORMATION = fopen( "TRANSFOR.$1$", "r" );
	fseek( TRANSFORMATION, 0L, 2 );
	REST_OF_TS = TRAN_SIZE = ftell( TRANSFORMATION );
	fseek( TRANSFORMATION, 0L, 0 );
	while( REST_OF_TS > 0L )
	  {
	    int BLOCK = REST_OF_TS > MAX_BLOCK ? MAX_BLOCK : REST_OF_TS;
	    unsigned short int DM_TRANS = DmCreate( BLOCK );
	    DmPush( DM_INDX, (char *)&DM_TRANS, 2 );
	    fread( DmPointer( DM_TRANS ), BLOCK, 1, TRANSFORMATION );
	    REST_OF_TS -= BLOCK;
	  }
	fclose( TRANSFORMATION );
	unlink( "TRANSFOR.$1$" );
	CURRENT_NBR = CURRENT_SEEK = 0L;
	CURRENT_PAIR = *(point_pair *)DmPointer( 
				*(unsigned short int *)DmPointer( DM_INDX ) );
	F_OCTAB = (f_size_pointer *)DmFix( OCCUPIED );

	/* Rewrite atoms to new files with extension "$0$" */
	for( POINT = 0L; POINT != NULLPNTR; POINT = AtxNext( POINT ) )
	    WriteToNewFile( POINT );

	/* Remove TRANSFORMATION tables */
	for( TT = (unsigned short int *)DmFix( DM_INDX );
				DmInside( (char *)TT, DM_INDX ); TT++ )
	    DmDelete( *TT );
	DmUnfix( DM_INDX );
	DmDelete( DM_INDX );

	/* Close and renaming the files */
	PsCancUpdates();
	for( ; DmInside( (char *)F_OCTAB, OCCUPIED ); F_OCTAB++ )
	    if( *F_OCTAB->PATHNODE )			/* If not cancelled */
	      {
		char OLDNAME[ FMSIZE ], ACTNAME[ FMSIZE ], WORKNAME[ FMSIZE ];
		fclose( F_OCTAB->FPOINTER );

		strcpy( OLDNAME, F_OCTAB->PATHNODE );
		strcat( OLDNAME, OLD_EXT );
		strcpy( ACTNAME, F_OCTAB->PATHNODE );
		strcpy( WORKNAME, ACTNAME );
		strcat( WORKNAME, "$0$" );
		unlink( OLDNAME );
		rename( ACTNAME, OLDNAME );
		rename( WORKNAME, ACTNAME );
	      }
	DmUnfixDelete( OCCUPIED );
	return( 1 );
      } 
    return( 0 );
  }


/*-------------------------------------------------------- PsWrongStructure */
/* Return 1 iff name and RESERVE in atom are incorrect and 0 otherwise */

static int  PsWrongStructure( POINT )
    long POINT;
  {    
    char N[ 50 ];
    unsigned short int DMOBJ = AtxRead( POINT, 1 )->DMOBJ;
    atom_head *AH = (atom_head *)DmFix( DMOBJ );
    if( AH->NAME < 0xFFE0 && ! PsCodeName( AH->NAME, N ) ||
                                       AH->RESERVED[0] || AH->RESERVED[1] )
      {
	DmUnfix( DMOBJ );
	return( 1 );
      }
    DmUnfix( DMOBJ );
    return( 0 );
  }
  

/*--------------------------------------------------------------------------*/
    
#define OLD_PRGELEM	2
#define PRGELEM		253

static void  CheckPdsPoint( POINT )
    long POINT;
  {
    int TYPE, ATFILE;
    unsigned short int DMLOAD;
    
    TYPE = AtxType( POINT );
    ATFILE = AtxFile( POINT );
    DMLOAD = *( OCTAB + ATFILE );

    NBR_OF_ATOMS++;	/* Number of atoms in VPOS */
    if( LOGICAL_B <= TYPE && TYPE <= LOGICAL_E ||
        LONGINT_B <= TYPE && TYPE <= LONGINT_E )
	SIZE_OF_VALUES +=4;     /* Total size of values utilized in VPOS */
    else if( BYTES_B <= TYPE && TYPE <= BYTES_E )
        SIZE_OF_VALUES += AtmValSize( POINT );
        
    if( ! DMLOAD || ! DmInside( (char *)( OCTAB + ATFILE ), OCCUPIED ) )
      {
	char INFOR[ 120 ], BUF[ 20 ];
	 /* File does not exist */
	ConvPoint( POINT, BUF );
	sprintf( INFOR, FindString( sPOINT_NOT_EXISTS ), BUF );
/*
	PushInfoText( INFOR );
	NextInfoLine( SYSTEM_MSG );
*/
	printf( "\n%s", INFOR );
	VPOS_CONSISTENT = 0;
      }
    else
      {
	int RESULT = 1;
	long BGPOINT, EDPOINT;
	long *BCL, *CL;
	if( DM_COMPLAIN )
	    PsAtxDel( );
	BGPOINT = AtxPoint( POINT );		/* Beginning of the point */ 
	EDPOINT = BGPOINT + AtxSize( POINT );	/* After the end of point */
	/* Comparison of the point range with ranges on the description of
	   occupied areas */
	for( BCL = CL = (long *)DmFix( DMLOAD );
				DmInside( (char *)CL, DMLOAD ); CL += 2 )
	  {
	    long BGOC = *CL, EDOC = *( CL + 1 );
	    /* Is there a part common with the area occupied by the point? */
	    if( BGOC <= BGPOINT && BGPOINT < EDOC ||
		BGOC <= EDPOINT - 1 && EDPOINT - 1 < EDOC ||
		BGPOINT <= BGOC && EDPOINT >= EDOC )
	      {	/* The areas overlap! */
		char INFOR[ 120 ], NAME[ 80 ], BUF[ 20 ];
		ConvPoint( POINT, BUF );
		PsCodeName( AtmName( POINT ), NAME );
		sprintf( INFOR, FindString( sATOM_OVERLAPS ), BUF, NAME );
/*
		PushInfoText( INFOR );
		NextInfoLine( SYSTEM_MSG );
*/
		printf( "\n%s", INFOR );
		InternalReact2( POINT, NULLPNTR );
		RESULT = VPOS_CONSISTENT = 0;
		break;
	      }
	    else if( EDPOINT <= BGOC )
		break;
	  }
	if( RESULT )
	  {
	    int STICK = 0;
	    long FROM;
	    /* Try to stick area occupied by the atom with existing areas */
	    /* Check the possibility to stick with previous */
	    if( DmInside( (char *)(CL - 2), DMLOAD ) &&
						*( CL - 1 ) == BGPOINT )
		STICK = 1;
	    /* Check the possibility to stick with next */		
	    if( DmInside( (char *)CL, DMLOAD ) && *CL == EDPOINT )
		STICK += 2;
	    /* 4 cases */
	    FROM = (long)( (char *)CL - (char *)BCL );
	    switch( STICK )
	      {
		case  0:	/* No sticking */
		    DmUnfix( DMLOAD );
		    DmDistend( DMLOAD, FROM, 8 );
		    CL = (long *)(DmPointer( DMLOAD ) + FROM );
		    *CL = BGPOINT;
		    *( CL + 1 ) = EDPOINT;
		    break;
		case  1:	/* Stick with previous */
		    *( CL - 1 ) = EDPOINT;
		    DmUnfix( DMLOAD );
		    break;
		case  2:	/* Stick with next */
		    *CL = BGPOINT;
		    DmUnfix( DMLOAD );
		    break;
		case  3:	/* Stick with previous and next */
		    *( CL - 1 ) = *( CL+ 1 );
		    *CL = *( CL - 2 ); 
		    DmUnfix( DMLOAD );
		    DmExtract( DMLOAD, FROM, 8 );
		    break;
	      }
	  }
	else
	    DmUnfix( DMLOAD );
	if( RING_B<=TYPE && TYPE<=RING_E || SPIDER_B<=TYPE && TYPE<=SPIDER_E )
	  {
	    long SUB, NEXTSUB;
	    for( SUB = AtxSubord( POINT ); SUB != NULLPNTR; SUB = NEXTSUB )
	      {
		long OW;
		NEXTSUB = AtxNext( SUB );
		if( PsWrongStructure( SUB ) || PsIsGarbage( SUB ) )
		  {
		    char INFOR[ 120 ], N[ 80 ], B[ 20 ], B_OW[ 20 ];
		    ConvPoint( SUB, B ); 
		    ConvPoint( POINT, B_OW );
		    PsCodeName( AtmName( POINT ), N );
		    sprintf( INFOR, FindString( sPOINT_WRONG_STRUCTURE ),
								B, B_OW, N );
/*
		    PushInfoText( INFOR );
		    NextInfoLine( SYSTEM_MSG );
*/
		    printf( "\n%s", INFOR );
		    InternalReact2( POINT, NULLPNTR );
		    VPOS_CONSISTENT = 0;
		    break;		  
		  }
		OW = AtxOwner( SUB );
		if( OW != POINT )
		  {
		    char INFOR[ 120 ], N[ 80 ], B[ 20 ], B_OW[ 20 ];
		    PsCodeName( AtmName( SUB ), N );
		    ConvPoint( SUB, B ); 
		    sprintf( INFOR, FindString( sINCONSISTENCY_POINT ), B, N );
/*
                    PushInfoText( INFOR );
		    NextInfoLine( SYSTEM_MSG );
*/
		    printf( "\n%s", INFOR );
		    PsCodeName( AtmName( POINT ), N );
		    ConvPoint( POINT, B ); 
		    ConvPoint( OW, B_OW );

/*
		    beep();
*/
		    sprintf( INFOR, FindString( sOWNER_SHOULD_BE ),
							B, N, B_OW );
/*
		    PushInfoText( INFOR );
		    NextInfoLine( SYSTEM_MSG );
*/
		    printf( "\n%s", INFOR );
		    InternalReact2( SUB, POINT );
		    VPOS_CONSISTENT = 0;
		    return;
		  }
		CheckPdsPoint( SUB );
	      }
	  }
	if( LOGICAL_B <= TYPE && TYPE <= LOGICAL_E || TYPE == HOLDER )
	  {
	    long POINTED = AtxSubord( POINT );
	    if( POINTED != NULLPNTR )
	      {
		if( PsIsGarbage( POINTED ) )
		  {
		    atomdesc *AD = AtxRead( POINT, 0 );
		        /* Inconsistency: logical pointer leads to garbage */
		    char INFOR[ 120 ], N[ 80 ], B[ 20 ], B_PO[ 20 ];
		    PsCodeName( AtmName( POINT ), N );
		    ConvPoint( POINT, B );
		    ConvPoint( POINTED, B_PO );
/*
		    beep();
*/
		    sprintf( INFOR, FindString( sPOINTER_REPAIRED ),
								B, N, B_PO );
/*
		    PushInfoText( INFOR );
		    NextInfoLine( SYSTEM_MSG );
*/
		    printf( "\n%s", INFOR );
		    ((atom_head *)DmPointer( AD->DMOBJ ))->VALUE = NULLPNTR;
		    AD->FLAGS |= UPDATED;
		    CHANGES_IN_VPOS = 1;
		    VPOS_CONSISTENT = 0;
		  }
		else
		  {		
		    int NOT_FOUND = 1;
		    long BACK = AtxNext( POINTED );
		    if( BACK != NULLPNTR && AtxType( BACK ) == BACKWARD )
		      {
		       /* Check if the logical pointer is among backward pointers*/
		       unsigned short int DM_BACK = AtxRead( BACK, 0 )->DMOBJ;
		       long *BP = (long *)(DmFix( DM_BACK ) + ATOMHEAD );
		       for( ; DmInside( (char *)BP, DM_BACK ); BP++ )
		           if( *BP == POINT )
			     {
			       NOT_FOUND = 0;
			       break;
			     }
		        DmUnfix( DM_BACK );
		      }
		    else
		        BACK = AtxInsert( AtmCreate( BACKWARD, 0xFFF8 ),
							NULLPNTR, POINTED );
		    if( NOT_FOUND )
		      {
			unsigned short int DM_BACK;
			/* Inconsistency: logical pointer leads to garbage or
			it has no twin backward;
			include POINT to backward pointers */
			char INFOR[ 120 ], N[ 80 ], B[ 20 ], B_PO[ 20 ];
			PsCodeName( AtmName( POINT ), N );
			ConvPoint( POINT, B );
			ConvPoint( POINTED, B_PO );
/*
			beep();
*/
			sprintf( INFOR, FindString( sPOINT_REPAIRED ),
								B, N, B_PO );
/*
			PushInfoText( INFOR );
			NextInfoLine( SYSTEM_MSG );
*/
			printf( "\n%s", INFOR );
			DM_BACK = AtxGetValue( BACK );
			DmPush( DM_BACK, (char *)&POINT, 4 );
			AtxPutValue( BACK, DM_BACK );
			VPOS_CONSISTENT = 0;
		      }
		  }
	      }	  
	  }
	if( TYPE == BACKWARD )
	  {
	    /* Check all points contained in the backward atom if they lead
	    to appropriate logical pointers */
	    atomdesc *AD = AtxRead( POINT, 0 );
	    unsigned short int DM_BACK = AD->DMOBJ;
	    long *BP = (long *)( DmFix( DM_BACK ) + ATOMHEAD );
	    for( ; DmInside( (char *)BP, DM_BACK ); BP++ )
		if( *BP != NULLPNTR )
		  {
		    int LPTYPE = PsIsGarbage( *BP ) ? -1 : AtxType( *BP );
		    if( LPTYPE == -1 ||
		        ( ! ( ( LOGICAL_B <= LPTYPE && LPTYPE <= LOGICAL_E ||
			LPTYPE == HOLDER ) &&
			AtxNext( AtxSubord( *BP ) ) == POINT ) ) )
		      {
			char INFOR[ 120 ], B[ 20 ];
/*
			beep();
*/
			ConvPoint( POINT, B );
			sprintf( INFOR, FindString( sBACKWARD_REPAIRED ), B );
/*
			PushInfoText( INFOR );
			NextInfoLine( SYSTEM_MSG );
*/
			printf( "\n%s", INFOR );
			AD->FLAGS |= UPDATED;
			CHANGES_IN_VPOS = 1;
			VPOS_CONSISTENT = 0;
			*BP = NULLPNTR;		      
		      }
		  }
	    DmUnfix( DM_BACK );
	  }
	if( TYPE == TOEND )
	  {
	    /* Check if its pointer really points the last member; if not,
	       repair it by assigning a proper value */
	    int INCON = 0;
	    atomdesc *AD;
	    long P, POWN, *LL;
	    for( P = POINT; AtxNext( P ) != NULLPNTR; P = AtxNext( P ) );
	    POWN = AtxSubord( AtxOwner( P ) );
	    if( POWN != POINT )	/* If TOEND is not the first member */
	      {
		INCON = 1;
		AtxDelete( NULLPNTR, POINT );
	      }
	    AD = AtxRead( POINT, 0 );
	    LL = &( ((atom_head *)DmPointer( AD->DMOBJ ))->VALUE );
	    if( *LL != P )
	      {
		INCON = 1;
		*LL = P;
		AD->FLAGS |= UPDATED;
	      }
	    if( INCON )
	      {
		char INFOR[ 120 ], B[ 20 ];
/*
		beep();
*/
		ConvPoint( POINT, B );
		sprintf( INFOR, FindString( sTOEND_REPAIRED ), B );
/*
		PushInfoText( INFOR );
		NextInfoLine( SYSTEM_MSG );
*/
		printf( "\n%s", INFOR );
		CHANGES_IN_VPOS = 1;
		VPOS_CONSISTENT = 0;
	      }
	  }
      }
  }


/*--------------------------------------------------------------------------*/
/* The procedure makes a list of areas occupied by atoms (sticking areas); it
checks if atoms do not overlap. It also checks if pointers to owners are
properly connected and if each logical pointer has a twin backward pointer
and otherwise. At the end the procedure tests if the current garbage lists are
negations of lists of occupied areas; if there is a conflict, the negation of
a list of occupied areas substitutes a garbage list. The procedure checks if
file sizes written in the file table are correct. In case of inconsistency the
procedure tries to make an appropriate correction of VPOS. */

int CheckWholeVPOS( )
  {
    char INFOR[ 120 ];
    long POINT, NEXTPOINT, _COUNTER, PIECES, T_ACT_FSIZE = 0, T_COUNTER = 0;
    int FNO;
    filedesc *FD;
/*
    DirectMessage( FindString( sDATA_STRUCTURE_CHECKED ),
		23, 1, 25, 80, 4, 15 );
*/
    VPOS_CONSISTENT = 1;	/* Shows consistency of VPOS */
    NBR_OF_ATOMS = 0;
    SIZE_OF_VALUES = 0;

	/* Temporary action: changing a program module into spider */
    for( POINT = AtmSubord( AtmNext( 0L ) ); POINT != NULLPNTR;
							POINT = NEXTPOINT )
      {
	NEXTPOINT = AtmNext( POINT );
	if( AtxType( POINT ) == OLD_PRGELEM )
	    AtmMoveAllMembers( POINT,
	    AtmCreateAndInsert( PRGELEM, AtmName(POINT), NULLPNTR, POINT ) );
      }
	/* End of temporary action */

/* Temporary property:remove all spiders from organizational ring */
for( POINT = AtmSubord( 0L ); POINT != NULLPNTR; )
  {
    unsigned short int NN;
    long P = AtmNext( POINT );
    if( AtmTypeAndName( POINT, &NN ) == 254 || NN == 0xFFF7 )
        AtmDelete( POINT );
    POINT = P;
  }
/* End of temporary property */

/*
    PushInfoTextMsg( sVPOS_STATISTICS );
    NextInfoLine( SYSTEM_MSG );
*/
    OCCUPIED = DmCreate( 0 );
    for( FD = (filedesc *)DmFix( FILETABLE ); 
				DmInside( (char *)FD, FILETABLE ); FD++ )
      {
	unsigned short int DMO = 0;
	if( ! ( FD IS_CANCELLED ) )
	    DMO = DmCreate( 0 );
	DmPush( OCCUPIED, (char *)&DMO, 2 );
      }
    DmUnfix( FILETABLE );
    OCTAB = (unsigned short int *)DmFix( OCCUPIED );

    for( POINT = 0L; POINT != NULLPNTR; POINT = NEXTPOINT )
      {
	NEXTPOINT = AtxNext( POINT ); 
        if( PsIsGarbage( POINT ) || PsWrongStructure( POINT ) )
          {
	    char INFOR[ 120 ], B[ 20 ];
	    ConvPoint( POINT, B );
	    sprintf( INFOR, FindString( sPOINT_ANYTHING ), B );
/*
	    PushInfoText( INFOR );
	    NextInfoLine( SYSTEM_MSG );
*/
	    printf( "\n%s", INFOR );
	    VPOS_CONSISTENT = 0;
	    break;
          }    
	CheckPdsPoint( POINT );
      }

    /* Comparison with existing garbage tables and changing
       them to negations of actual loadings */

    FNO = 0;
    for( FD = (filedesc *)DmFix( FILETABLE );
		DmInside( (char *)OCTAB, OCCUPIED ); OCTAB++, FNO++, FD++ )
      {
	int INCON;
	gardesc *NEW_GAR, *OLD_GAR;
	long *BOC, *OC, OLD_GARSIZE, NEW_GARSIZE;
	long ACT_FSIZE = 0L;		/* Actual file size */
	if( FD IS_CANCELLED )
	    continue;
	_COUNTER = PIECES = 0;
	if( DmSize( *OCTAB ) != 0 )
	  {
	    for( OC = (long *)DmFix( *OCTAB );
				DmInside( (char *)OC, *OCTAB ); OC += 2 )
	      {
		_COUNTER += *( OC + 1 ) - *OC;
		PIECES++;
	      }
	    PIECES--;
	    OC--;
	    ACT_FSIZE = *OC;
	    DmUnfix( *OCTAB );
	  }
	if( ACT_FSIZE != FD->APPASIZE )
	  {
	    sprintf( INFOR, FindString( sFILE_SIZE_IMPROVED ),
						FNO, FD->APPASIZE, ACT_FSIZE );
/*
	    PushInfoText( INFOR );
	    NextInfoLine( SYSTEM_MSG );
*/
	    printf( "\n%s", INFOR );
	    CHANGES_IN_VPOS = 1;
	    FD->APPASIZE = ACT_FSIZE;
	  }
	sprintf( INFOR, FindString( sFILE_SIZE_UTILIZED ),
		FNO, ACT_FSIZE, _COUNTER,
		ACT_FSIZE > 0 ? ( _COUNTER * 100 + 50 ) / ACT_FSIZE : 0L );
/*
	PushInfoText( INFOR );
	NextInfoLine( SYSTEM_MSG );
*/

	T_ACT_FSIZE += ACT_FSIZE;
	T_COUNTER += _COUNTER;

	sprintf( INFOR, FindString( sGARBAGE_CONSISTS ), PIECES );
/*
	PushInfoText( INFOR );
	NextInfoLine( SYSTEM_MSG );
*/


	/* Form an object with the new garbage; same object *OCTAB is used */
	if( DmSize( *OCTAB ) != 0 )
	  {
	    BOC = (long *)DmFix( *OCTAB );
	    for( NEW_GAR = (gardesc *)BOC, OC = BOC + 1;
			DmInside( (char *)OC, *OCTAB ); NEW_GAR++, OC += 2 )
	      {
		long GBEG = *OC, GSIZE = *(OC+1) - *OC;
		PsLongGar( NEW_GAR, GBEG, GSIZE );
	      }
	    DmUnfix( *OCTAB );
	    DmChangeSize( *OCTAB, ( DmSize(*OCTAB) - 8 ) / 8*sizeof(gardesc) );
	  }
	/* Compare existing garbage descriptor with the given in *OCTAB */
	OLD_GARSIZE = NEW_GARSIZE = 0;
	INCON = 0;	/* Indicator "is inconsistent" */
	if( ! FD->GARBAGE )
	    FD->GARBAGE = DmCreate( 0 );
	for( OLD_GAR = (gardesc *)DmFix( FD->GARBAGE ),
		 NEW_GAR = (gardesc *)DmFix( *OCTAB );
		 DmInside( (char *)OLD_GAR, FD->GARBAGE ) ||
		 DmInside( (char *)NEW_GAR, *OCTAB ); OLD_GAR++,NEW_GAR++)
	  {
	    long OBEG = -1, OSIZE = 0, NBEG = -1, NSIZE = 0;	      
	    if( DmInside( (char *)OLD_GAR, FD->GARBAGE ) )
		PsGarLong( OLD_GAR, OBEG, OSIZE );
	    if( DmInside( (char *)NEW_GAR, *OCTAB ) )
		PsGarLong( NEW_GAR, NBEG, NSIZE );
	    OLD_GARSIZE += OSIZE;
	    NEW_GARSIZE += NSIZE;
	    if( OBEG != NBEG || OSIZE != NSIZE )
		INCON = 1;
	  }
	DmUnfix( FD->GARBAGE );
	DmUnfix( *OCTAB );
	if( INCON )
	  {
	    sprintf( INFOR, FindString( sGARBAGE_INADEQUATE ),
					FNO, OLD_GARSIZE, NEW_GARSIZE );
/*
	    PushInfoText( INFOR );
	    NextInfoLine( SYSTEM_MSG );
*/
	    printf( "\n%s", INFOR );
	    if( FD->GARBAGE )
		DmDelete( FD->GARBAGE );
	    FD->GARBAGE = *OCTAB;
	    CHANGES_IN_VPOS = 1;
	  }
	else
	    DmDelete( *OCTAB );
      }

    sprintf( INFOR, FindString( sTOTAL_ATOMS ), NBR_OF_ATOMS );
/*
    PushInfoText( INFOR );
    NextInfoLine( SYSTEM_MSG );
*/

    sprintf( INFOR, FindString( sTOTAL_VALUES ), SIZE_OF_VALUES,
		T_ACT_FSIZE > 0 ? ( SIZE_OF_VALUES*100+50 )/T_ACT_FSIZE : 0L,
		T_COUNTER > 0 ? ( SIZE_OF_VALUES*100+50 )/T_COUNTER : 0L );
/*
    PushInfoText( INFOR );
    NextInfoLine( SYSTEM_MSG );
*/
    DmUnfix( FILETABLE );
    DmUnfixDelete( OCCUPIED );
/*
    DirectMessage( NULL );
*/
    return( VPOS_CONSISTENT );
  }


/* ======================================================================== */
