/*-------------------------------------------------------------------------+
| Virtul persistent object store, internal functions, header (SUN version) |
|                                                                          |
| Date of last updating: 18 November 1991                                  |
|                                                                          |
| Copyright (C) 1991 by Kazimierz Subieta.                                 |
| This software is subject to a licence agreement.                         |
| No restrictions when used for didactic purposes.                         |
+_________________________________________________________________________*/


#define FMSIZE 80 /* For alignment must be dividible by 4 */


typedef struct	/* atomdesc: atom description within inventory */
  {
    unsigned char	FLAGS;	/* Flags associated with the atom:
				b0 - updating flag:	1 means updated
				b1 - cancelling flag:   1 means cancelled
				b2 - head only flag:	1 means head only
				b3 - overflow object:	1 means overflow
					*/
    char ADDITIONAL;	/* K.S. Optimization: to make size 8 */
    unsigned short int	DMOBJ;	/* Name of DM object containing the atom */
    long	POINT;	/* DB pointer of an atom; the key of hash function */
  } atomdesc;

/*...........................................................................*/
typedef struct		/* Stored description of a data file */
  {
    char PATHNODE[ FMSIZE ];	/* Path + file name */
    char EXT[4];
    long SIZE; /* Size of the file. Alignment! FMSIZE % 4 should be 0 */
  } dbfiledesc;

/*...........................................................................*/
typedef struct		/* Description of a data file in opened VPOS */
  {	/* The file number is simultaneously the relative position of the
	   file item within FILETABLE; if the item is cancelled, PATHNODE is
	   an empty string. */
    dbfiledesc	DESC;		/* Stored description of a data file */
    unsigned char	FLAGS;		/* Flags associated with the file
				b0 - updating flag:	1 means updated
				b1 - cancelling flag:   1 means cancelled
				b2 - new file flag: 1 means that the new file
				    has been physically created by fopen.
				b3 - temporality:	1 means temporary file
				b4 - apparency: 1 means that the file exists
				    apparently, without disc counterpart.
					*/
    char        DUMMY; /* K.S. for alignment */
    unsigned short int	GARBAGE;	/* DM object with garbage areas */
	/* Note on the organization of garbage on data files.
	1. Each file has an atom with the description of its garbage areas.
	   The description consists of the following items (gardesc):
	   a) pointer to garbage area (3 bytes)
	   b) size of garbage area (3 bytes).
	2. All these atoms are kept in main file and are part of db
	   description.
	3. After the table of files is read, all garbage description atoms
	   are also read and connected to the table description. */
    long	APPASIZE;	/* Apparent file size */
    FILE	*FPOINTER;	/* File pointer to a data file,
				or NULL if the file is not opened */
  } filedesc;

/*...........................................................................*/
typedef struct		/* An item of garbage description */
  {
    long GARPOINT;
    long GARSIZE;
  } gardesc;

/*...........................................................................*/
typedef struct		/* An item of a spider atom */
  {
    unsigned char     MEMBER_TYPE; /* Introduced for inheritance property */
    char DUMMY; /* K.S.for alignment */
    unsigned short int MEMBER_NAME;
    long     MEMBER;
  } spider_item;

/*---------------------------------------------------------------------------*/

#define UPDATED		0x01
#define TEMPORARY	0x08
#define OVERFLOW	0x08
#define IS_CANCELLED	->FLAGS & 0x02

extern unsigned short int PERMNAMES;
extern unsigned short int  FILETABLE;
extern unsigned short int   AINVMAIN;
extern atomdesc  *BEGAMAIN;	/* Beginning of atom inventory */
extern unsigned short int   AINVOVER;	/* DM object with overflow items of atom inventory */
extern atomdesc  *BEGAOVER;

/*---------------------------------------------------------------------------*/

extern void	 AtxDelete(	/* long, long PREVIOUS, POINT */ );
extern void	 AtxDisconnect( /* long */ );
extern int	 AtxFile(/*long POINT*/);
extern unsigned short int  AtxGetValue(	/* long POINT */ );
extern long	 AtxInsert(	/* unsigned,long,long DMOBJECT,OWNER,AFTER*/);
extern long	 AtxNext(	/* long POINT */ );
extern long	 AtxOwner(	/* long POINT */ );
extern long	 AtxPoint(/*long POINT*/);
extern long	 AtxPutValue(	/* long, unsigned POINT, DMOBJECT */ );
extern atomdesc*  AtxRead(	/* long, int POINT, MODE */ );
extern long	 AtxSize(	/* long POINT */ );
extern long	 AtxSubord(	/* long POINT */ );
extern int	 AtxType( /* long POINT */ );
extern void	 AtxWrite(	/* FILE *, unsigned FPOINTER, DMOBJ */ );
extern void	 DiscFromSpider( /* long, long SPOWNER, MEMBER */ );
#define		     PsGarLong( _GARBAGE, _GARPOINT, _GARSIZE ) \
  {					\
    (_GARPOINT) = (_GARBAGE)->GARPOINT;	\
    (_GARSIZE ) = (_GARBAGE)->GARSIZE;	\
  }
extern void	 PsLongGar(	/* gardesc *, long, long
			GARBAGE, GARPOINT, GARSIZE */ );

extern int    RemNUDmAtom(	/* atomdesc *, int AT, FRQ */ );
#define       SkipHolders(P) while( AtxType( P ) == HOLDER ) P = AtxSubord( P )
extern void DispWholeVPOS( /* FILE * */ );

extern int CheckWholeVPOS( );

/* ======================================================================== */
