/*-------------------------------------------------------------------------+
| Virtul persistent object store, basic functions, header (SUN version)    |
|                                                                          |
| Date of last updating: 18 November 1991                                  |
|                                                                          |
| Copyright (C) 1991 by Kazimierz Subieta.                                 |
| This software is subject to a licence agreement.                         |
| No restrictions when used for didactic purposes.                         |
+_________________________________________________________________________*/

#define ATOMHEAD        sizeof(atom_head)
#define NULLPNTR        -1L
#define VPOS_ENTRY      0L

extern int PS_ERROR;
extern unsigned char CURR_FILE;
extern unsigned char CHANGES_IN_VPOS;

/*........................................................ atom_head ........*/
typedef struct          /* atom_head:  - 13 bytes */
  {
    unsigned char       TYPE;   /* The following types are provided 
			RING       0..5    - ring atoms
			LOGICAL    6-8     - logical pointer atoms
			HOLDER     9       - place holder atom
			TOEND      10      - reference to ring's last atom 
			LONGINT    11..20  - long integer atoms
			BYTES      21..240 - variable size atoms
			SPIDER    241..254 - spider atoms
			BACKWARD   255     - backward pointers atom */
    unsigned char       STATUS; /* The following bytes have a meaning:
			b0 - free,
			b1 = 0 - pointer to next     b1 = 1 - pointer to owner
					*/
    unsigned short int  NAME;
    long        POINT;  /* Data pointer: organization assumed:
			POINT = (B3 B2 B1 B0), where Bi denote a byte.
			B0 - file number,
			B3 B2 B1 - relative position within the file. */
    long        VALUE;  /* May contain size of value or second pointer */

    unsigned char       RESERVED[4];    /* For future use and for alignment */
  } atom_head;
/* Note alignment to even number of bytes */

/* This part of definitions applies concepts of LOQIS programming language */

#define RING            0
#define AGGREGATE       1       /* This constant is also valid for stacks */
#define OLD_PRGELEM     2
#define VPOSLINK        6
#define ANCESTOR        7       /* Inheritance element */
#define VPOSINT         11
#define VPOSDATE        12      /* Integer denoting date in days */
#define VPOSTIME        13      /* Integer denoting time in seconds */
#define VPOSREAL        21
#define VPOSSTRING      22
#define LOQISTEXT       128
#define PRGAGNES        129
#define PRGDUMMY        130
#define PRG_C           132
#define STR_TABLE       239     /* Table of strings used on stacks */
#define FREEDATA        240
#define PRGELEM         253
#define SPIDER          254
/* End of "This part..." */

extern unsigned char RING_B;
extern unsigned char RING_E;
extern unsigned char LOGICAL_B;
extern unsigned char LOGICAL_E;
extern unsigned char HOLDER;
extern unsigned char TOEND;
extern unsigned char LONGINT_B;
extern unsigned char LONGINT_E;
extern unsigned char BYTES_B;
extern unsigned char BYTES_E;
extern unsigned char SPIDER_B;
extern unsigned char SPIDER_E;
extern unsigned char BACKWARD;
extern char OLD_EXT[];

/*...........................................................................*/

extern long      AtmAssign(     /*      long, long */ );
extern long      AtmChangeSize( /*long, long POINT, VALSIZE */ );
extern unsigned short int AtmCreate(    /* int, unsigned TYPE, NAME */ );
extern long      AtmCreateAndInsert( /*int, unsigned, long, long */ );
extern void      AtmDelete(     /* long POINT */ );
extern int       AtmDetPointer( /* unsigned, long DMOBJECT, POINTER */ );
extern int       AtmDetValue(   /* unsigned, char *, long
				 DMOBJECT, VALUE, VALSIZE */ );
extern int       AtmFile(       /* long  POINT */ );
extern long      AtmFind(       /* long, unsigned, int OWNER, NAME, TYPE */ );
extern long      AtmFindNext(   /*long, unsigned, int ENTRY, NAME, TYPE */ );
extern int       AtmFlag(       /*long, int */);
extern void      AtmFlagOn(     /*      long, int */);
extern void      AtmFlagOff(    /* long, int */ );
extern unsigned short int AtmGet( /* long POINT */ );
extern unsigned short int  AtmGetValue( /* long POINT */ );
extern long      AtmInsCpLast(  /* long, long  POINT, OWNER */ );
extern long      AtmInsert(     /* unsigned,long,long DMOBJECT,OWNER,AFTER */);
extern long      AtmInsLast(    /* unsigned, long DMOBJECT, OWNER */ );
extern int       AtmIntAssign(  /*long, long POINT, VALUE */ );
extern long      AtmIntValue(   /* long  POINT */ );
extern void      AtmMove(       /* long, long, long POINT,OWNER,AFTER */ );
extern long      AtmMoveAllMembers( /* long, long OLD_OWNER, NEW_OWNER */ );
extern void      AtmMoveLast(   /* long, long  POINT, OWNER */ );
extern unsigned short int AtmName( /* long POINT */ );
extern long      AtmNext(       /* long POINT */ );
extern long      AtmOwner(      /* long POINT */ );
extern int       AtmPntAssign(  /* long, long, int POINT, POINTER, NEWTYPE */);
extern long      AtmPoint(      /* long POINT */ );
extern long      AtmPutValue(   /* long, unsigned POINT, DMOBJECT */ );
extern unsigned short int AtmReferences( /* long POINT */ );
extern void      AtmRename(     /* long, unsigned POINT, NAME */ );
extern long      AtmSubord(     /* long POINT */ );
extern int       AtmType(       /* long POINT */ );
extern int       AtmTypeAndName(/* long, unsigned * POINT, NAME */ );
extern void      AtmUpdated(    /* long POINT */ );
extern long      AtmValAssign(  /* long, int, char *, long
				   POINT, TYPE, VALUE, VALSIZE */ );
extern long      AtmValSize(    /* long POINT */ );
extern int       CheckWholeVPOS();
extern int       ConvPoint(     /* long, char * */ );
extern void      DispPoint(     /* FILE *, long, int */ );
extern void      DispWholeVPOS(         /* FILE * */ );
extern void      PsAtxDel();
extern int       PsCancUpdates();
extern void      PsCleanDict();
extern int       PsClose();
extern int       PsCloseFile(   /* int FILENO */ );
extern int       PsCodeName(    /* unsigned,   char * CODE, NAME */ );
extern int       PsCreateFile(  /* char *, int FILENAME, TEMPOR */ );
extern int       PsDeleteFile(  /* int FILENO */ );
extern long      PsDelHolder(   /* long POINT */ );
extern int       PsDelName(     /* unsigned, int CODE, MODE */ );
extern void      PsDetTypes(    /* int, int, int, int
			RING_TYPE, LOGIC_TYPE, LONG_TYPE, BYTES_TYPE */ );
extern int       PsInit();
extern unsigned short int PsInsName(    /* char * NAME */ );
extern int       PsIsGarbage(   /* long POINT */ );
extern unsigned  short int PsNameCode(  /* char * NAME */ );
extern int       PsNameSize(    /* unsigned CODE */ );
extern unsigned short int PsNamesVolume( );
extern unsigned short int PsNoHolder( );
extern int       PsOpen(        /* char * MAINFILE */);
extern int       PsRewrite(     /* char *, char *, long */ );
extern void      PsStatistics(  /* unsigned *, unsigned * DMATOMS, FILESTAT*/ );
extern int       PsTemporary(   /* int FILENO */ );
extern int       PsReorganize(  /* char * */ );

/* ======================================================================== */
